    ! Module pikaia - converted further
    module Genetic_Algorithm

    implicit none

    !     Common block to make iseed visible to rninit (and to save
    !     it between calls)
    integer, save  :: iseed

    CONTAINS

    subroutine pikaia(ff,n,ctrl,x,f,status)

    ! Code converted using TO_F90 by Alan Miller
    ! Date: 2001-07-09  Time: 15:54:13

    !====================================================================
    !  Optimization (maximization) of user-supplied "fitness" function ff
    !  over n-dimensional parameter space  x  using a basic genetic algorithm
    !  method.

    !  Paul Charbonneau & Barry Knapp
    !  High Altitude Observatory
    !  National Center for Atmospheric Research
    !  Boulder CO 80307-3000
    !  USA
    !  <paulchar@hao.ucar.edu>
    !  <knapp@hao.ucar.edu>

    !  Web site:
    !  http://www.hao.ucar.edu/public/research/si/pikaia/pikaia.html

    !  Version 1.0   [ 1995 December 01 ]      adjusted to version 1.2 [2011 Feb 28]
    !                                               by david de la croix

    !  Genetic algorithms are heuristic search techniques that incorporate in a
    !  computational setting, the biological notion of evolution by means of
    !  natural selection.  This subroutine implements the three basic operations
    !  of selection, crossover, and mutation, operating on "genotypes" encoded as
    !  strings.

    !  References:

    !     Charbonneau, Paul.  "Genetic Algorithms in Astronomy and Astrophysics."
    !        Astrophysical J. (Supplement), vol 101, in press (December 1995).

    !     Goldberg, David E.  Genetic Algorithms in Search, Optimization,
    !        & Machine Learning.  Addison-Wesley, 1989.

    !     Davis, Lawrence, ed.  Handbook of Genetic Algorithms.
    !        Van Nostrand Reinhold, 1991.
    !====================================================================
    !  USES: ff, urand, setctl, report, rnkpop, select, encode, decode,
    !        cross, mutate, genrep, stdrep, newpop, adjmut

    integer, intent(in)   :: n
    real, intent(inout)  :: ctrl(12)
    real*8, intent(out)     :: x(n)
    real*8, intent(out)     :: f
    integer, intent(out)  :: status

    ! this command replaces EXTERNAL FF of the old code
    interface
    function ff(n, x) RESULT(fn_val)
    implicit none
    integer, intent(in)  :: n
    real*8, intent(in)     :: x(n)
    real*8                 :: fn_val
    end function ff
    end interface

    !  Input:
    !  o Integer  n  is the parameter space dimension, i.e., the number
    !    of adjustable parameters.

    !  o Function  ff  is a user-supplied scalar function of n variables, which
    !    must have the calling sequence f = ff(n,x), where x is a real parameter
    !    array of length n.  This function must be written so as to bound all
    !    parameters to the interval [0,1]; that is, the user must determine
    !    a priori bounds for the parameter space, and ff must use these bounds
    !    to perform the appropriate scalings to recover true parameter values in
    !    the a priori ranges.

    !    By convention, ff should return higher values for more optimal
    !    parameter values (i.e., individuals which are more "fit").
    !    For example, in fitting a function through data points, ff
    !    could return the inverse of chi**2.

    !    In most cases initialization code will have to be written
    !    (either in a driver or in a separate subroutine) which loads
    !    in data values and communicates with ff via one or more labeled
    !    common blocks.  An example exercise driver and fitness function
    !    are provided in the accompanying file, xpkaia.f.


    !  Input/Output:


    !  o Array  ctrl  is an array of control flags and parameters, to
    !    control the genetic behavior of the algorithm, and also printed
    !    output.  A default value will be used for any control variable
    !    which is supplied with a value less than zero.  On exit, ctrl
    !    contains the actual values used as control variables.  The
    !    elements of ctrl and their defaults are:

    !       ctrl( 1) - number of individuals in a population (default
    !                  is 100)
    !       ctrl( 2) - number of generations over which solution is
    !                  to evolve (default is 500)
    !       ctrl( 3) - number of significant digits (i.e., number of
    !                  genes) retained in chromosomal encoding (default
    !                  is 6)  (Note: This number is limited by the
    !                  machine floating point precision.  Most 32-bit
    !                  floating point representations have only 6 full
    !                  digits of precision.  To achieve greater preci-
    !                  sion this routine could be converted to double
    !                  precision, but note that this would also require
    !                  a double precision random number generator, which
    !                  likely would not have more than 9 digits of
    !                  precision if it used 4-byte integers internally.)
    !       ctrl( 4) - crossover probability; must be  <= 1.0 (default is 0.85)
    !       ctrl( 5) - mutation mode; 1/2=steady/variable (default is 2)
    !       ctrl( 6) - initial mutation rate; should be small (default is 0.005)
    !                  (Note: the mutation rate is the probability that any one
    !                  gene locus will mutate in any one generation.)
    !       ctrl( 7) - minimum mutation rate; must be >= 0.0 (default is 0.0005)
    !       ctrl( 8) - maximum mutation rate; must be <= 1.0 (default is 0.25)
    !       ctrl( 9) - relative fitness differential; range from 0
    !                  (none) to 1 (maximum).  (default is 1.)
    !       ctrl(10) - reproduction plan; 1/2/3=Full generational
    !                  replacement/Steady-state-replace-random/Steady-
    !                  state-replace-worst (default is 3)
    !       ctrl(11) - elitism flag; 0/1=off/on (default is 0)
    !                  (Applies only to reproduction plans 1 and 2)
    !       ctrl(12) - printed output 0/1/2=None/Minimal/Verbose (default is 0)


    ! Output:


    !  o Array  x(1:n)  is the "fittest" (optimal) solution found,
    !     i.e., the solution which maximizes fitness function ff

    !  o Scalar  f  is the value of the fitness function at x

    !  o Integer  status  is an indicator of the success or failure
    !     of the call to pikaia (0=success; non-zero=failure)


    ! Constants

    integer, parameter :: nmax = 32, pmax = 500, dmax = 6

    !  o NMAX is the maximum number of adjustable parameters (n <= NMAX)

    !  o PMAX is the maximum population (ctrl(1) <= PMAX)

    !  o DMAX is the maximum number of Genes (digits) per Chromosome
    !        segement (parameter) (ctrl(3) <= DMAX)


    !     Local variables
    integer :: np, nd, ngen, imut, irep, ielite, ivrb, k, ip, ig, ip1,  &
        ip2, NEW, newtot
    real*8 :: pcross, pmut, pmutmn, pmutmx, fdif

    real*8 :: ph(nmax,2), newph(nmax,pmax)
    real*8 :: oldph(nmax,pmax)

    integer :: gn1(nmax*dmax), gn2(nmax*dmax)
    integer :: ifit(pmax), jfit(pmax)
    real*8 :: fitns(pmax)

    !     User-supplied uniform random number generator
    ! real :: urand
    ! EXTERNAL urand

    ! Function urand should not take any arguments.  If the user wishes to be able
    ! to initialize urand, so that the same sequence of random numbers can be
    ! repeated, this capability could be implemented with a separate subroutine,
    ! and called from the user's driver program.  An example urand function
    ! (and initialization subroutine) which uses the function ran0 (the "minimal
    ! standard" random number generator of Park and Miller [Comm. ACM 31, 1192-
    ! 1201, Oct 1988; Comm. ACM 36 No. 7, 105-110, July 1993]) is provided.


    !     Set control variables from input and defaults
    call setctl(ctrl, n, np, ngen, nd, pcross, pmutmn, pmutmx, pmut, imut, fdif, &
        irep, ielite, ivrb, status)
    if (status /= 0) then
        write (*, *) ' Control vector (ctrl) argument(s) invalid'
        return
    end if

    !     Make sure locally-dimensioned arrays are big enough
    if (n > nmax .OR. np > pmax .OR. nd > dmax) then
        write (*, *) ' Number of parameters, population, or genes too large'
        status = -1
        return
    end if

    !     Compute initial (random but bounded) phenotypes
    do  ip = 1, np
        do  k = 1, n
            oldph(k,ip) = urand()
        end do
        fitns(ip) = ff(n, oldph(:,ip))
    end do

    !     Rank initial population by fitness order
    call rnkpop(np,fitns,ifit,jfit)

    !     Main Generation Loop
    do  ig = 1, ngen

        !        Main Population Loop
        newtot = 0
        do  ip = 1, np / 2

            !           1. pick two parents
            call select(np,jfit,fdif,ip1)
30          call select(np,jfit,fdif,ip2)
            if (ip1 == ip2) go to 30

            !           2. encode parent phenotypes
            call encode(n,nd,oldph(1,ip1),gn1)
            call encode(n,nd,oldph(1,ip2),gn2)

            !           3. breed
            call cross(n,nd,pcross,gn1,gn2)
            call mutate(n,nd,pmut,gn1,imut)
            call mutate(n,nd,pmut,gn2,imut)

            !           4. decode offspring genotypes
            call decode(n,nd,gn1,ph(1,1))
            call decode(n,nd,gn2,ph(1,2))

            !           5. insert into population
            if (irep == 1) then
                call genrep(nmax,n,np,ip,ph,newph)
            else
                call stdrep(ff,nmax,n,np,irep,ielite,ph,oldph,fitns,ifit, jfit,NEW)
                newtot = newtot + NEW
            end if

            !        End of Main Population Loop
        end do

        !        if running full generational replacement: swap populations
        if (irep == 1) call newpop(ff,ielite,nmax,n,np,oldph,newph,ifit,  &
            jfit,fitns,newtot)

        !        adjust mutation rate?
        if (imut.eq.2 .or. imut.eq.3 .or. imut.eq.5 .or. imut.eq.6) &
            call adjmut(NMAX,n,np,oldph,fitns,ifit,pmutmn,pmutmx,pmut,imut)

        !        print generation report to standard output?
        if (ivrb > 0) call report(ivrb,nmax,n,np,nd,oldph,fitns,ifit,pmut,ig,newtot)

        !     End of Main Generation Loop
    end do

    !     Return best phenotype and its fitness
    do  k = 1, n
        x(k) = oldph(k,ifit(np))
    end do
    f = fitns(ifit(np))

    end subroutine pikaia

    !********************************************************************

    subroutine setctl(ctrl,n,np,ngen,nd,pcross,pmutmn,pmutmx,pmut,  &
        imut,fdif,irep,ielite,ivrb,status)
    !===================================================================
    !     Set control variables and flags from input and defaults
    !===================================================================

    real, intent(inout)  :: ctrl(12)
    integer, intent(in)   :: n

    !     Output
    integer, intent(out)  :: np
    integer, intent(out)  :: ngen
    integer, intent(out)  :: nd
    real*8, intent(out)     :: pcross
    real*8, intent(out)     :: pmutmn
    real*8, intent(out)     :: pmutmx
    real*8, intent(out)     :: pmut
    integer, intent(out)  :: imut
    real*8, intent(out)     :: fdif
    integer, intent(out)  :: irep
    integer, intent(out)  :: ielite
    integer, intent(out)  :: ivrb
    integer, intent(out)  :: status


    !     Local
    integer :: i
    real*8, save  :: dfault(12) = (/ 100., 500., 5., .85, 2., .005, .0005, .25,  &
        1., 1., 1., 0. /)

    do  i = 1, 12
        if (ctrl(i) < 0.) ctrl(i) = dfault(i)
    end do

    np = ctrl(1)
    ngen = ctrl(2)
    nd = ctrl(3)
    pcross = ctrl(4)
    imut = ctrl(5)
    pmut = ctrl(6)
    pmutmn = ctrl(7)
    pmutmx = ctrl(8)
    fdif = ctrl(9)
    irep = ctrl(10)
    ielite = ctrl(11)
    ivrb = ctrl(12)
    status = 0

    !     Print a header
    if (ivrb.gt.0) then

        write(*,2) ngen,np,n,nd,pcross,pmut,pmutmn,pmutmx,fdif
2       format(/1x,60('*'),/,                                      &
            ' *',13x,'PIKAIA Genetic Algorithm Report ',13x,'*',/,   &
            1x,60('*'),//,                                      &
            '   Number of Generations evolving: ',i4,/,              &
            '       Individuals per generation: ',i4,/,              &
            '    Number of Chromosome segments: ',i4,/,              &
            '    Length of Chromosome segments: ',i4,/,              &
            '            Crossover probability: ',f9.4,/,            &
            '            Initial mutation rate: ',f9.4,/,           &
            '            Minimum mutation rate: ',f9.4,/,           &
            '            Maximum mutation rate: ',f9.4,/,           &
            '    Relative fitness differential: ',f9.4)
        if (imut.eq.1) write(*,3) 'Uniform, Constant Rate'
        if (imut.eq.2) write(*,3) 'Uniform, Variable Rate (F)'
        if (imut.eq.3) write(*,3) 'Uniform, Variable Rate (D)'
        if (imut.eq.4) write(*,3) 'Uniform+Creep, Constant Rate'
        if (imut.eq.5) write(*,3) 'Uniform+Creep, Variable Rate (F)'
        if (imut.eq.6) write(*,3) 'Uniform+Creep, Variable Rate (D)'
3       format('                    Mutation Mode: ',A)
        if (irep.eq.1) write(*,4) 'Full generational replacement'
        if (irep.eq.2) write(*,4) 'Steady-state-replace-random'
        if (irep.eq.3) write(*,4) 'Steady-state-replace-worst'
4       format('                Reproduction Plan: ',A)
    endif

    !     Check some control values
    if (imut.ne.1 .and. imut.ne.2 .and. imut.ne.3 .and. imut.ne.4  &
        .and. imut.ne.5 .and. imut.ne.6) then
    write(*,10)
    status = 5
    endif
10  format(' ERROR: illegal value for imut (ctrl(5))')

    if (fdif.gt.1.) then
        write(*,11)
        status = 9
    endif
11  format(' ERROR: illegal value for fdif (ctrl(9))')

    if (irep.ne.1 .and. irep.ne.2 .and. irep.ne.3) then
        write(*,12)
        status = 10
    endif
12  format(' ERROR: illegal value for irep (ctrl(10))')

    if (pcross.gt.1.0 .or. pcross.lt.0.) then
        write(*,13)
        status = 4
    endif
13  format(' ERROR: illegal value for pcross (ctrl(4))')

    if (ielite.ne.0 .and. ielite.ne.1) then
        write(*,14)
        status = 11
    endif
14  format(' ERROR: illegal value for ielite (ctrl(11))')

    if (irep.eq.1 .and. imut.eq.1 .and. pmut.gt.0.5 .and.ielite.eq.0) then
        write(*,15)
    endif
15  format(' WARNING: dangerously high value for pmut (ctrl(6));',  &
        /' (Should enforce elitism with ctrl(11)=1.)')

    if (irep.eq.1 .and. imut.eq.2 .and. pmutmx.gt.0.5 .and.     &
        ielite.eq.0) then
    write(*,16)
    endif
16  format(' WARNING: dangerously high value for pmutmx (ctrl(8));', &
        /' (Should enforce elitism with ctrl(11)=1.)')

    if (fdif.lt.0.33 .and. irep.ne.3) then
        write(*,17)
    endif
17  format(' WARNING: dangerously low value of fdif (ctrl(9))')

    if (mod(np,2).gt.0) then
        np=np-1
        write(*,18) np
    endif
18  format(' WARNING: decreasing population size (ctrl(1)) to np=',i4)

    end subroutine setctl

    !********************************************************************

    subroutine report(ivrb, ndim, n, np, nd, oldph, fitns, ifit, pmut, ig, nnew)

    !     Write generation report to standard output

    integer,  intent(in)  :: ivrb
    integer,  intent(in)  :: ndim
    integer,  intent(in)  :: n
    integer,  intent(in)  :: np
    integer,  intent(in)  :: nd
    real*8, intent(in)      :: oldph(ndim, np)
    real*8, intent(in)      :: fitns(np)
    integer, intent(in)   :: ifit(np)
    real*8, intent(in)      :: pmut
    integer, intent(in)   :: ig
    integer, intent(in)   :: nnew

    !     Local
    real*8, save  :: bestft = 0.0, pmutpv = 0.0
    integer  :: ndpwr, k
    LOGICAL  :: rpt

    rpt = .false.

    if (pmut /= pmutpv) then
        pmutpv = pmut
        rpt = .true.
    end if

    if (fitns(ifit(np)) /= bestft) then
        bestft = fitns(ifit(np))
        rpt = .true.
    end if

    if (rpt .OR. ivrb >= 2) then

        !        Power of 10 to make integer genotypes for display
        ndpwr = nint(10.**nd)

        write (*, '(/i6, i6, f10.6, 4f10.6)') ig, nnew, pmut,  &
            fitns(ifit(np)), fitns(ifit(np-1)), fitns(ifit(np/2))
        do  k = 1, n
            write (*, '(22x, 3i10)') nint(ndpwr*oldph(k, ifit(np))),  &
                nint(ndpwr*oldph(k, ifit(np-1))), nint(ndpwr*oldph(k, ifit(np/2)))
        end do

    end if
    end subroutine report

    !**********************************************************************
    !                         GENETICS module
    !**********************************************************************

    !     ENCODE:    encodes phenotype into genotype
    !                called by: PIKAIA

    !     DECODE:    decodes genotype into phenotype
    !                called by: PIKAIA

    !     CROSS:     Breeds two offspring from two parents
    !                called by: PIKAIA

    !     MUTATE:    Introduces random mutation in a genotype
    !                called by: PIKAIA

    !     ADJMUT:    Implements variable mutation rate
    !                called by: PIKAIA

    !**********************************************************************

    subroutine encode(n, nd, ph, gn)
    !======================================================================
    !     encode phenotype parameters into integer genotype
    !     ph(k) are x, y coordinates [ 0 < x, y < 1 ]
    !======================================================================


    integer, intent(in)   :: n
    integer, intent(in)   :: nd
    real*8, intent(inout)  :: ph(n)
    integer, intent(out)  :: gn(n*nd)

    !     Local:
    integer :: ip, i, j, ii
    real*8 :: z

    z = 10. ** nd
    ii = 0
    do  i = 1, n
        ip = INT(ph(i)*z)
        do  j = nd, 1, -1
            gn(ii+j) = MOD(ip, 10)
            ip = ip / 10
        end do
        ii = ii + nd
    end do

    end subroutine encode

    !**********************************************************************

    subroutine decode(n, nd, gn, ph)
    !======================================================================
    !     decode genotype into phenotype parameters
    !     ph(k) are x, y coordinates [ 0 < x, y < 1 ]
    !======================================================================

    integer, intent(in)  :: n
    integer, intent(in)  :: nd
    integer, intent(in)  :: gn(n*nd)
    real*8, intent(out)    :: ph(n)

    !     Local:
    integer :: ip, i, j, ii
    real*8 :: z

    z = 10. ** (-nd)
    ii = 0
    do  i = 1, n
        ip = 0
        do  j = 1, nd
            ip = 10 * ip + gn(ii+j)
        end do
        ph(i) = ip * z
        ii = ii + nd
    end do

    end subroutine decode

    !**********************************************************************

    subroutine cross(n, nd, pcross, gn1, gn2)
    !======================================================================
    !     breeds two parent chromosomes into two offspring chromosomes
    !     breeding occurs through crossover. If the crossover probability
    !     test yields true (crossover taking place), either one-point or
    !     two-point crossover is used, with equal probabilities.
    !
    !     Compatibility with version 1.0: To enforce 100% use of one-point
    !     crossover, un-comment appropriate line in source code below
    !======================================================================
    !     USES: urand

    integer, intent(in)      :: n
    integer, intent(in)      :: nd
    real*8, intent(in)         :: pcross

    integer, intent(inout)  :: gn1(n*nd)
    integer, intent(inout)  :: gn2(n*nd)

    !     Local:
    integer :: i, ispl,ispl2,itmp, t

    !     Function
    ! real :: urand
    ! EXTERNAL urand

    !     Use crossover probability to decide whether a crossover occurs
    if (urand() < pcross) then

        !        Compute crossover point
        ispl = INT(urand()*n*nd) + 1

        !        Now choose between one-point and two-point crossover
        if (urand().lt.0.5) then
            ispl2=n*nd
        else
            ispl2=int(urand()*n*nd)+1
            !           Un-comment following line to enforce one-point crossover
            !           ispl2=n*nd
            if (ispl2.lt.ispl) then
                itmp=ispl2
                ispl2=ispl
                ispl=itmp
            endif
        endif

        !        Swap genes from ispl to ispl2
        do 10 i=ispl,ispl2
            t=gn2(i)
            gn2(i)=gn1(i)
            gn1(i)=t
10      continue
    endif

    end subroutine cross

    !**********************************************************************

    subroutine mutate(n, nd, pmut, gn,imut)
    !======================================================================
    !     Mutations occur at rate pmut at all gene loci
    !======================================================================
    !     Mutations occur at rate pmut at all gene loci
    !        imut=1    Uniform mutation, constant rate
    !        imut=2    Uniform mutation, variable rate based on fitness
    !        imut=3    Uniform mutation, variable rate based on distance
    !        imut=4    Uniform or creep mutation, constant rate
    !        imut=5    Uniform or creep mutation, variable rate based on
    !                  fitness
    !        imut=6    Uniform or creep mutation, variable rate based on
    !                  distance
    !======================================================================
    !     USES: urand

    integer, intent(in)      :: n
    integer, intent(in)      :: nd
    integer, intent(in)      :: imut
    real*8, intent(in)         :: pmut

    integer, intent(inout)  :: gn(n*nd)

    !     Local:
    integer :: i,j,k,l,ist,inc,loc,kk

    !     Decide which type of mutation is to occur
    if(imut.ge.4.and.urand().le.0.5)then

        !     CREEP MUTATION OPERATOR
        !     Subject each locus to random +/- 1 increment at the rate pmut
        do 1 i=1,n
            do 2 j=1,nd
                if (urand().lt.pmut) then
                    !     Construct integer
                    loc=(i-1)*nd+j
                    inc=nint ( urand() )*2-1
                    ist=(i-1)*nd+1
                    gn(loc)=gn(loc)+inc
                    !                 write(*,*) ist,loc,inc
                    !     This is where we carry over the one (up to two digits)
                    !     first take care of decrement below 0 case
                    if(inc.lt.0 .and. gn(loc).lt.0)then
                        if(j.eq.1)then
                            gn(loc)=0
                        else
                            do 3 k=loc,ist+1,-1
                                gn(k)=9
                                gn(k-1)=gn(k-1)-1
                                if( gn(k-1).ge.0 )goto 4
3                           continue
                            !    we popped under 0.00000 lower bound; fix it up
                            if( gn(ist).lt.0.)then
                                do 5 l=ist,loc
                                    gn(l)=0
5                               continue
                            endif
4                           continue
                        endif
                    endif
                    if(inc.gt.0 .and. gn(loc).gt.9)then
                        if(j.eq.1)then
                            gn(loc)=9
                        else
                            do 6 k=loc,ist+1,-1
                                gn(k)=0
                                gn(k-1)=gn(k-1)+1
                                if( gn(k-1).le.9 )goto 7
6                           continue
                            !                       we popped over 9.99999 upper bound; fix it up
                            if( gn(ist).gt.9 )then
                                do 8 l=ist,loc
                                    gn(l)=9
8                               continue
                            endif
7                           continue
                        endif
                    endif
                endif
2           continue
1       continue

    else

        !     UNIFORM MUTATION OPERATOR
        !     Subject each locus to random mutation at the rate pmut
        do 10 i=1,n*nd
            if (urand().lt.pmut) then
                gn(i)=int(urand()*10.)
            endif
10      continue
    endif


    end subroutine mutate

    !**********************************************************************

    subroutine adjmut(ndim,n,np,oldph,fitns,ifit,pmutmn,pmutmx,pmut,imut)
    !======================================================================
    !     dynamical adjustment of mutation rate;
    !        imut=2 or imut=5 : adjustment based on fitness differential
    !                           between best and median individuals
    !        imut=3 or imut=6 : adjustment based on metric distance
    !                           between best and median individuals
    !======================================================================
    !     Input:
    integer, intent(in)   :: n,np,ndim,imut
    real*8, intent(in)      :: fitns(:)
    integer, intent(in)   :: ifit(:)
    real*8, intent(in)      :: pmutmn
    real*8, intent(in)      :: pmutmx,oldph(ndim,np)

    !     Input/Output:
    real*8, intent(inout)  :: pmut

    !     Local:
    integer :: i
    real*8  :: rdif
    real*8, parameter  :: rdiflo = 0.05, rdifhi = 0.25, delta = 1.5

    if(imut.eq.2.or.imut.eq.5)then
        !     Adjustment based on fitness differential
        rdif=abs(fitns(ifit(np))-fitns(ifit(np/2)))/(fitns(ifit(np))+fitns(ifit(np/2)))
    else if(imut.eq.3.or.imut.eq.6)then
        !     Adjustment based on normalized metric distance
        rdif=0.
        do 1 i=1,n
            rdif=rdif+( oldph(i,ifit(np))-oldph(i,ifit(np/2)) )**2
1       continue
        rdif=sqrt( rdif ) / float(n)
    endif

    if(rdif.le.rdiflo)then
        pmut=min(pmutmx,pmut*delta)
    else if(rdif.ge.rdifhi)then
        pmut=max(pmutmn,pmut/delta)
    endif

    end subroutine adjmut

    !**********************************************************************
    !                       REPRODUCTION module
    !**********************************************************************

    !  SELECT:   Parent selection by roulette wheel algorithm
    !            called by: PIKAIA

    !  RNKPOP:   Ranks initial population
    !            called by: PIKAIA, NEWPOP

    !  GENREP:   Inserts offspring into population, for full
    !            generational replacement
    !            called by: PIKAIA

    !  STDREP:   Inserts offspring into population, for steady-state
    !            reproduction
    !            called by: PIKAIA
    !            calls:     FF

    !  NEWPOP:   Replaces old generation with new generation
    !            called by: PIKAIA
    !            calls:     FF, RNKPOP

    !**********************************************************************

    subroutine select(np, jfit, fdif, idad)
    !======================================================================
    !     Selects a parent from the population, using roulette wheel
    !     algorithm with the relative fitnesses of the phenotypes as
    !     the "hit" probabilities [see Davis 1991, chap. 1].
    !======================================================================
    !     USES: urand

    !     Input:
    integer, intent(in)   :: np
    integer, intent(in)   :: jfit(np)
    real*8, intent(in)      :: fdif

    !     Output:
    integer, intent(out)  :: idad

    !     Local:
    integer :: np1, i
    real*8 :: dice, rtfit

    np1 = np + 1
    dice = urand() * np * np1
    rtfit = 0.
    do  i = 1, np
        rtfit = rtfit + np1 + fdif * (np1-2*jfit(i))
        if (rtfit >= dice) then
            idad = i
            go to 20
        end if
    end do
    !     Assert: loop will never exit by falling through

20  return
    end subroutine select

    !**********************************************************************

    subroutine rnkpop(n, arrin, indx, rank)
    !======================================================================
    !     Calls external sort routine to produce key index and rank order
    !     of input array arrin (which is not altered).
    !======================================================================
    !     USES: rqsort

    !     Input
    integer, intent(in)   :: n
    real*8, intent(in)      :: arrin(:)

    !     Output
    integer, intent(out)  :: indx(:)
    integer, intent(out)  :: rank(:)

    !     Local
    integer :: i

    !     Compute the key index
    call rqsort(n, arrin, indx)

    !     ...and the rank order
    do  i = 1, n
        rank(indx(i)) = n - i + 1
    end do

    end subroutine rnkpop

    !***********************************************************************

    subroutine genrep(ndim, n, np, ip, ph, newph)
    !=======================================================================
    !     full generational replacement: accumulate offspring into new
    !     population array
    !=======================================================================

    !     Input:
    integer, intent(in)  :: ndim
    integer, intent(in)  :: n
    integer, intent(in)  :: np
    integer, intent(in)  :: ip
    real*8, intent(in)     :: ph(ndim, 2)

    !     Output:
    real*8, intent(out)    :: newph(ndim, np)

    !     Local:
    integer :: i1, i2, k

    !     Insert one offspring pair into new population
    i1 = 2 * ip - 1
    i2 = i1 + 1
    do  k = 1, n
        newph(k, i1) = ph(k, 1)
        newph(k, i2) = ph(k, 2)
    end do

    end subroutine genrep

    !**********************************************************************

    subroutine stdrep(ff, ndim, n, np, irep, ielite, ph, oldph, fitns, ifit, jfit, nnew)
    !======================================================================
    !     steady-state reproduction: insert offspring pair into population
    !     only if they are fit enough (replace-random if irep=2 or
    !     replace-worst if irep=3).
    !======================================================================
    !     USES: ff, urand

    !     Input:
    integer, intent(in)      :: ndim
    integer, intent(in)      :: n
    integer, intent(in)      :: np
    integer, intent(in)      :: irep
    integer, intent(in)      :: ielite
    real*8, intent(in)         :: ph(ndim, 2)

    !     Input/Output:
    real*8, intent(inout)     :: oldph(ndim, np)
    real*8, intent(inout)     :: fitns(np)
    integer, intent(inout)  :: ifit(np)
    integer, intent(inout)  :: jfit(np)

    !     Output:
    integer, intent(out)     :: nnew

    interface
    function ff(n, x) RESULT(fn_val)
    implicit none
    integer, intent(in)  :: n
    real*8, intent(in)     :: x(n)
    real*8                 :: fn_val
    end function ff
    end interface

    ! EXTERNAL ff

    !     Local:
    integer :: i, j, k, i1, if1
    real*8 :: fit

    !     External function
    ! real :: urand
    ! EXTERNAL urand


    nnew = 0
    loop70:  do  j = 1, 2

        !        1. compute offspring fitness (with caller's fitness function)
        fit = ff(n, ph(:, j))

        !        2. if fit enough, insert in population
        do  i = np, 1, -1
            if (fit > fitns(ifit(i))) then

                !              make sure the phenotype is not already in the population
                if (i < np) then
                    do  k = 1, n
                        if (oldph(k, ifit(i+1)) /= ph(k, j)) go to 20
                    end do
                    CYCLE loop70
                end if

                !              offspring is fit enough for insertion, and is unique

                !              (i) insert phenotype at appropriate place in population
20              if (irep == 3) then
                    i1 = 1
                else if (ielite == 0 .OR. i == np) then
                    i1 = INT(urand()*np) + 1
                else
                    i1 = INT(urand()*(np-1)) + 1
                end if
                if1 = ifit(i1)
                fitns(if1) = fit
                do  k = 1, n
                    oldph(k, if1) = ph(k, j)
                end do

                !              (ii) shift and update ranking arrays
                if (i < i1) then

                    !                 shift up
                    jfit(if1) = np - i
                    do  k = i1 - 1, i + 1, -1
                        jfit(ifit(k)) = jfit(ifit(k)) - 1
                        ifit(k+1) = ifit(k)
                    end do
                    ifit(i+1) = if1
                else

                    !                 shift down
                    jfit(if1) = np - i + 1
                    do  k = i1 + 1, i
                        jfit(ifit(k)) = jfit(ifit(k)) + 1
                        ifit(k-1) = ifit(k)
                    end do
                    ifit(i) = if1
                end if
                nnew = nnew + 1
                CYCLE loop70
            end if
        end do

    end do loop70

    end subroutine stdrep

    !**********************************************************************

    subroutine newpop(ff, ielite, ndim, n, np, oldph, newph, ifit, jfit, fitns, nnew)
    !======================================================================
    !     replaces old population by new; recomputes fitnesses & ranks
    !======================================================================
    !     USES: ff, rnkpop

    !     Input:
    integer, intent(in)   :: ielite
    integer, intent(in)   :: ndim
    integer, intent(in)   :: n
    integer, intent(in)   :: np

    !     Input/Output:
    real*8, intent(inout)  :: oldph(ndim, np)
    real*8, intent(inout)  :: newph(ndim, np)

    !     Output:
    integer, intent(out)  :: ifit(np)
    integer, intent(out)  :: jfit(np)
    real*8, intent(out)     :: fitns(np)
    integer, intent(out)  :: nnew

    interface
    function ff(n, x) RESULT(fn_val)
    implicit none
    integer, intent(in)  :: n
    real*8, intent(in)     :: x(n)
    real*8                 :: fn_val
    end function ff
    end interface

    !     Local:
    integer :: i, k

    nnew = np

    !     if using elitism, introduce in new population fittest of old
    !     population (if greater than fitness of the individual it is
    !     to replace)
    if (ielite == 1 .AND. ff(n, newph(:, 1)) < fitns(ifit(np))) then
        do  k = 1, n
            newph(k, 1) = oldph(k, ifit(np))
        end do
        nnew = nnew - 1
    end if

    !     replace population
    do  i = 1, np
        do  k = 1, n
            oldph(k, i) = newph(k, i)
        end do

        !        get fitness using caller's fitness function
        fitns(i) = ff(n, oldph(:, i))
    end do

    !     compute new population fitness rank order
    call rnkpop(np, fitns, ifit, jfit)

    end subroutine newpop

    !*********************************************************************

    function urand() RESULT(fn_val)
    !=====================================================================
    !  Return the next pseudo-random deviate from a sequence which is
    !  uniformly distributed in the interval [0, 1]

    !  Uses the function ran0, the "minimal standard" random number
    !  generator of Park and Miller (Comm. ACM 31, 1192-1201, Oct 1988;
    !  Comm. ACM 36 No. 7, 105-110, July 1993).
    !=====================================================================

    !     Input - none

    !     Output
    real*8  :: fn_val

    !     Local
    ! integer :: iseed
    ! real :: ran0
    ! EXTERNAL ran0

    fn_val = ran0()

    end function urand

    !*********************************************************************

    subroutine rninit(seed)
    !=====================================================================
    !     Initialize random number generator urand with given seed
    !=====================================================================

    !     Input
    integer, intent(in)  :: seed

    !     Output - none

    !     Set the seed value
    iseed = seed
    if (iseed <= 0) iseed = 123456

    end subroutine rninit

    !*********************************************************************

    function ran0() RESULT(fn_val)
    !=====================================================================
    !  "Minimal standard" pseudo-random number generator of Park and Miller.
    !  Returns a uniform random deviate r s.t. 0 < r < 1.0.
    !  Set seed to any non-zero integer value to initialize a sequence, then do
    !  not change seed between calls for successive deviates in the sequence.

    !  References:
    !     Park, S. and Miller, K., "Random Number Generators: Good Ones
    !        are Hard to Find", Comm. ACM 31, 1192-1201 (Oct. 1988)
    !     Park, S. and Miller, K., in "Remarks on Choosing and Implementing
    !        Random Number Generators", Comm. ACM 36 No. 7, 105-110 (July 1993)
    !=====================================================================
    ! *** Declaration section ***

    !     Output:
    real*8  :: fn_val

    !     Constants:

    integer, parameter  :: a = 48271, m = 2147483647, q = 44488, r = 3399

    real*8, parameter :: scale = 1./m, eps = 1.2E-7, rnmx = 1. - eps

    !     Local:
    integer  :: j

    ! *** Executable section ***
    j = iseed / q
    iseed = a * (iseed - j*q) - r * j
    if (iseed < 0) iseed = iseed + m
    fn_val = MIN(iseed*scale, rnmx)

    end function ran0

    !**********************************************************************

    subroutine rqsort(n, a, p)
    !======================================================================
    !  Return integer array p which indexes array a in increasing order.
    !  Array a is not disturbed.  The Quicksort algorithm is used.

    !  B. G. Knapp, 86/12/23

    !  Reference: N. Wirth, Algorithms and Data Structures/
    !  Prentice-Hall, 1986
    !======================================================================

    integer, intent(in)   :: n
    real*8, intent(in)      :: a(:)
    integer, intent(out)  :: p(:)

    !     Constants

    integer, parameter :: lgn = 32, q = 11
    !        (LGN = log base 2 of maximum n;
    !         Q = smallest subfile to use quicksort on)

    !     Local:
    real*8 :: x
    integer :: stackl(lgn), stackr(lgn), s, t, l, m, r, i, j

    !     Initialize the stack
    stackl(1) = 1
    stackr(1) = n
    s = 1

    !     Initialize the pointer array
    do  i = 1, n
        p(i) = i
    end do

20  if (s > 0) then
        l = stackl(s)
        r = stackr(s)
        s = s - 1

30      if ((r-l) < q) then

            !           Use straight insertion
            do  i = l + 1, r
                t = p(i)
                x = a(t)
                do  j = i - 1, l, -1
                    if (a(p(j)) <= x) go to 50
                    p(j+1) = p(j)
                end do
                j = l - 1
50              p(j+1) = t
            end do
        else

            !           Use quicksort, with pivot as median of a(l), a(m), a(r)
            m = (l+r) / 2
            t = p(m)
            if (a(t) < a(p(l))) then
                p(m) = p(l)
                p(l) = t
                t = p(m)
            end if
            if (a(t) > a(p(r))) then
                p(m) = p(r)
                p(r) = t
                t = p(m)
                if (a(t) < a(p(l))) then
                    p(m) = p(l)
                    p(l) = t
                    t = p(m)
                end if
            end if

            !           Partition
            x = a(t)
            i = l + 1
            j = r - 1
70          if (i <= j) then
80              if (a(p(i)) < x) then
                    i = i + 1
                    go to 80
                end if
90              if (x < a(p(j))) then
                    j = j - 1
                    go to 90
                end if
                if (i <= j) then
                    t = p(i)
                    p(i) = p(j)
                    p(j) = t
                    i = i + 1
                    j = j - 1
                end if
                go to 70
            end if

            !           Stack the larger subfile
            s = s + 1
            if (j-l > r-i) then
                stackl(s) = l
                stackr(s) = j
                l = i
            else
                stackl(s) = i
                stackr(s) = r
                r = j
            end if
            go to 30
        end if
        go to 20
    end if

    end subroutine rqsort

    end module Genetic_Algorithm
    !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    MODULE Powell_Optimize





    PRIVATE
    PUBLIC  :: newuoa

    CONTAINS


    SUBROUTINE NEWUOA (N,NPT,X,RHOBEG,RHOEND,IPRINT,MAXFUN,W)
    IMPLICIT REAL*8 (A-H,O-Z)
    DIMENSION X(*),W(*)

    !     This subroutine seeks the least value of a function of many variables,
    !     by a trust region method that forms quadratic models by interpolation.
    !     There can be some freedom in the interpolation conditions, which is
    !     taken up by minimizing the Frobenius norm of the change to the second
    !     derivative of the quadratic model, beginning with a zero matrix. The
    !     arguments of the subroutine are as follows.

    !     N must be set to the number of variables and must be at least two.
    !     NPT is the number of interpolation conditions. Its value must be in the
    !       interval [N+2,(N+1)(N+2)/2].
    !     Initial values of the variables must be set in X(1),X(2),...,X(N). They
    !       will be changed to the values that give the least calculated F.
    !     RHOBEG and RHOEND must be set to the initial and final values of a trust
    !       region radius, so both must be positive with RHOEND<=RHOBEG. Typically
    !       RHOBEG should be about one tenth of the greatest expected change to a
    !       variable, and RHOEND should indicate the accuracy that is required in
    !       the final values of the variables.
    !     The value of IPRINT should be set to 0, 1, 2 or 3, which controls the
    !       amount of printing. Specifically, there is no output if IPRINT=0 and
    !       there is output only at the return if IPRINT=1. Otherwise, each new
    !       value of RHO is printed, with the best vector of variables so far and
    !       the corresponding value of the objective function. Further, each new
    !       value of F with its variables are output if IPRINT=3.
    !     MAXFUN must be set to an upper bound on the number of calls of CALFUN.
    !     The array W will be used for working space. Its length must be at least
    !     (NPT+13)*(NPT+N)+3*N*(N+3)/2.

    !     SUBROUTINE CALFUN (N,X,F) must be provided by the user. It must set F to
    !     the value of the objective function for the variables X(1),X(2),...,X(N).

    !     Partition the working space array, so that different parts of it can be
    !     treated separately by the subroutine that performs the main calculation.

    NP=N+1
    NPTM=NPT-NP
    IF (NPT .LT. N+2 .OR. NPT .GT. ((N+2)*NP)/2) THEN
        PRINT 10
10      FORMAT (/4X,'Return from NEWUOA because NPT is not in',' the required interval')
        GO TO 20
    END IF
    NDIM=NPT+N
    IXB=1
    IXO=IXB+N
    IXN=IXO+N
    IXP=IXN+N
    IFV=IXP+N*NPT
    IGQ=IFV+NPT
    IHQ=IGQ+N
    IPQ=IHQ+(N*NP)/2
    IBMAT=IPQ+NPT
    IZMAT=IBMAT+NDIM*N
    ID=IZMAT+NPT*NPTM
    IVL=ID+N
    IW=IVL+NDIM

    !     The above settings provide a partition of W for subroutine NEWUOB.
    !     The partition requires the first NPT*(NPT+N)+5*N*(N+3)/2 elements of
    !     W plus the space that is needed by the last array of NEWUOB.

    CALL NEWUOB (N,NPT,X,RHOBEG,RHOEND,IPRINT,MAXFUN,W(IXB), &
        W(IXO),W(IXN),W(IXP),W(IFV),W(IGQ),W(IHQ),W(IPQ),W(IBMAT), &
        W(IZMAT),NDIM,W(ID),W(IVL),W(IW))
20  RETURN
    END   SUBROUTINE NEWUOA



    SUBROUTINE NEWUOB (N,NPT,X,RHOBEG,RHOEND,IPRINT,MAXFUN,XBASE, &
        XOPT,XNEW,XPT,FVAL,GQ,HQ,PQ,BMAT,ZMAT,NDIM,D,VLAG,W)
    IMPLICIT REAL*8 (A-H,O-Z)
    DIMENSION X(*),XBASE(*),XOPT(*),XNEW(*),XPT(NPT,*),FVAL(*), &
        GQ(*),HQ(*),PQ(*),BMAT(NDIM,*),ZMAT(NPT,*),D(*),VLAG(*),W(*)

    !     The arguments N, NPT, X, RHOBEG, RHOEND, IPRINT and MAXFUN are identical
    !       to the corresponding arguments in SUBROUTINE NEWUOA.
    !     XBASE will hold a shift of origin that should reduce the contributions
    !       from rounding errors to values of the model and Lagrange functions.
    !     XOPT will be set to the displacement from XBASE of the vector of
    !       variables that provides the least calculated F so far.
    !     XNEW will be set to the displacement from XBASE of the vector of
    !       variables for the current calculation of F.
    !     XPT will contain the interpolation point coordinates relative to XBASE.
    !     FVAL will hold the values of F at the interpolation points.
    !     GQ will hold the gradient of the quadratic model at XBASE.
    !     HQ will hold the explicit second derivatives of the quadratic model.
    !     PQ will contain the parameters of the implicit second derivatives of
    !       the quadratic model.
    !     BMAT will hold the last N columns of H.
    !     ZMAT will hold the factorization of the leading NPT by NPT submatrix of
    !       H, this factorization being ZMAT times Diag(DZ) times ZMAT^T, where
    !       the elements of DZ are plus or minus one, as specified by IDZ.
    !     NDIM is the first dimension of BMAT and has the value NPT+N.
    !     D is reserved for trial steps from XOPT.
    !     VLAG will contain the values of the Lagrange functions at a new point X.
    !       They are part of a product that requires VLAG to be of length NDIM.
    !     The array W will be used for working space. Its length must be at least
    !       10*NDIM = 10*(NPT+N).

    !     Set some constants.

    HALF=0.5D0
    ONE=1.0D0
    TENTH=0.1D0
    ZERO=0.0D0
    NP=N+1
    NH=(N*NP)/2
    NPTM=NPT-NP
    NFTEST=MAX0(MAXFUN,1)

    !     Set the initial elements of XPT, BMAT, HQ, PQ and ZMAT to zero.

    DO 20 J=1,N
        XBASE(J)=X(J)
        DO 10 K=1,NPT
10      XPT(K,J)=ZERO
        DO 20 I=1,NDIM
20  BMAT(I,J)=ZERO
    DO 30 IH=1,NH
30  HQ(IH)=ZERO
    DO 40 K=1,NPT
        PQ(K)=ZERO
        DO 40 J=1,NPTM
40  ZMAT(K,J)=ZERO

    !     Begin the initialization procedure. NF becomes one more than the number
    !     of function values so far. The coordinates of the displacement of the
    !     next initial interpolation point from XBASE are set in XPT(NF,.).

    RHOSQ=RHOBEG*RHOBEG
    RECIP=ONE/RHOSQ
    RECIQ=DSQRT(HALF)/RHOSQ
    NF=0
50  NFM=NF
    NFMM=NF-N
    NF=NF+1
    IF (NFM .LE. 2*N) THEN
        IF (NFM .GE. 1 .AND. NFM .LE. N) THEN
            XPT(NF,NFM)=RHOBEG
        ELSE IF (NFM .GT. N) THEN
            XPT(NF,NFMM)=-RHOBEG
        END IF
    ELSE
        ITEMP=(NFMM-1)/N
        JPT=NFM-ITEMP*N-N
        IPT=JPT+ITEMP
        IF (IPT .GT. N) THEN
            ITEMP=JPT
            JPT=IPT-N
            IPT=ITEMP
        END IF
        XIPT=RHOBEG
        IF (FVAL(IPT+NP) .LT. FVAL(IPT+1)) XIPT=-XIPT
        XJPT=RHOBEG
        IF (FVAL(JPT+NP) .LT. FVAL(JPT+1)) XJPT=-XJPT
        XPT(NF,IPT)=XIPT
        XPT(NF,JPT)=XJPT
    END IF

    !     Calculate the next value of F, label 70 being reached immediately
    !     after this calculation. The least function value so far and its index
    !     are required.

    DO 60 J=1,N
60  X(J)=XPT(NF,J)+XBASE(J)
    GOTO 310
70  FVAL(NF)=F
    IF (NF .EQ. 1) THEN
        FBEG=F
        FOPT=F
        KOPT=1
    ELSE IF (F .LT. FOPT) THEN
        FOPT=F
        KOPT=NF
    END IF

    !     Set the nonzero initial elements of BMAT and the quadratic model in
    !     the cases when NF is at most 2*N+1.

    IF (NFM .LE. 2*N) THEN
        IF (NFM .GE. 1 .AND. NFM .LE. N) THEN
            GQ(NFM)=(F-FBEG)/RHOBEG
            IF (NPT .LT. NF+N) THEN
                BMAT(1,NFM)=-ONE/RHOBEG
                BMAT(NF,NFM)=ONE/RHOBEG
                BMAT(NPT+NFM,NFM)=-HALF*RHOSQ
            END IF
        ELSE IF (NFM .GT. N) THEN
            BMAT(NF-N,NFMM)=HALF/RHOBEG
            BMAT(NF,NFMM)=-HALF/RHOBEG
            ZMAT(1,NFMM)=-RECIQ-RECIQ
            ZMAT(NF-N,NFMM)=RECIQ
            ZMAT(NF,NFMM)=RECIQ
            IH=(NFMM*(NFMM+1))/2
            TEMP=(FBEG-F)/RHOBEG
            HQ(IH)=(GQ(NFMM)-TEMP)/RHOBEG
            GQ(NFMM)=HALF*(GQ(NFMM)+TEMP)
        END IF

        !     Set the off-diagonal second derivatives of the Lagrange functions and
        !     the initial quadratic model.

    ELSE
        IH=(IPT*(IPT-1))/2+JPT
        IF (XIPT .LT. ZERO) IPT=IPT+N
        IF (XJPT .LT. ZERO) JPT=JPT+N
        ZMAT(1,NFMM)=RECIP
        ZMAT(NF,NFMM)=RECIP
        ZMAT(IPT+1,NFMM)=-RECIP
        ZMAT(JPT+1,NFMM)=-RECIP
        HQ(IH)=(FBEG-FVAL(IPT+1)-FVAL(JPT+1)+F)/(XIPT*XJPT)
    END IF
    IF (NF .LT. NPT) GOTO 50

    !     Begin the iterative procedure, because the initial model is complete.

    RHO=RHOBEG
    DELTA=RHO
    IDZ=1
    DIFFA=ZERO
    DIFFB=ZERO
    ITEST=0
    XOPTSQ=ZERO
    DO 80 I=1,N
        XOPT(I)=XPT(KOPT,I)
80  XOPTSQ=XOPTSQ+XOPT(I)**2
90  NFSAV=NF

    !     Generate the next trust region step and test its length. Set KNEW
    !     to -1 if the purpose of the next F will be to improve the model.

100 KNEW=0
    CALL TRSAPP (N,NPT,XOPT,XPT,GQ,HQ,PQ,DELTA,D,W,W(NP), &
        W(NP+N),W(NP+2*N),CRVMIN)
    DSQ=ZERO
    DO 110 I=1,N
110 DSQ=DSQ+D(I)**2
    DNORM=DMIN1(DELTA,DSQRT(DSQ))
    IF (DNORM .LT. HALF*RHO) THEN
        KNEW=-1
        DELTA=TENTH*DELTA
        RATIO=-1.0D0
        IF (DELTA .LE. 1.5D0*RHO) DELTA=RHO
        IF (NF .LE. NFSAV+2) GOTO 460
        TEMP=0.125D0*CRVMIN*RHO*RHO
        IF (TEMP .LE. DMAX1(DIFFA,DIFFB,DIFFC)) GOTO 460
        GOTO 490
    END IF

    !     Shift XBASE if XOPT may be too far from XBASE. First make the changes
    !     to BMAT that do not depend on ZMAT.

120 IF (DSQ .LE. 1.0D-3*XOPTSQ) THEN
        TEMPQ=0.25D0*XOPTSQ
        DO 140 K=1,NPT
            SUM=ZERO
            DO 130 I=1,N
130         SUM=SUM+XPT(K,I)*XOPT(I)
            TEMP=PQ(K)*SUM
            SUM=SUM-HALF*XOPTSQ
            W(NPT+K)=SUM
            DO 140 I=1,N
                GQ(I)=GQ(I)+TEMP*XPT(K,I)
                XPT(K,I)=XPT(K,I)-HALF*XOPT(I)
                VLAG(I)=BMAT(K,I)
                W(I)=SUM*XPT(K,I)+TEMPQ*XOPT(I)
                IP=NPT+I
                DO 140 J=1,I
140     BMAT(IP,J)=BMAT(IP,J)+VLAG(I)*W(J)+W(I)*VLAG(J)

        !     Then the revisions of BMAT that depend on ZMAT are calculated.

        DO 180 K=1,NPTM
            SUMZ=ZERO
            DO 150 I=1,NPT
                SUMZ=SUMZ+ZMAT(I,K)
150         W(I)=W(NPT+I)*ZMAT(I,K)
            DO 170 J=1,N
                SUM=TEMPQ*SUMZ*XOPT(J)
                DO 160 I=1,NPT
160             SUM=SUM+W(I)*XPT(I,J)
                VLAG(J)=SUM
                IF (K .LT. IDZ) SUM=-SUM
                DO 170 I=1,NPT
170         BMAT(I,J)=BMAT(I,J)+SUM*ZMAT(I,K)
            DO 180 I=1,N
                IP=I+NPT
                TEMP=VLAG(I)
                IF (K .LT. IDZ) TEMP=-TEMP
                DO 180 J=1,I
180     BMAT(IP,J)=BMAT(IP,J)+TEMP*VLAG(J)

        !     The following instructions complete the shift of XBASE, including
        !     the changes to the parameters of the quadratic model.

        IH=0
        DO 200 J=1,N
            W(J)=ZERO
            DO 190 K=1,NPT
                W(J)=W(J)+PQ(K)*XPT(K,J)
190         XPT(K,J)=XPT(K,J)-HALF*XOPT(J)
            DO 200 I=1,J
                IH=IH+1
                IF (I .LT. J) GQ(J)=GQ(J)+HQ(IH)*XOPT(I)
                GQ(I)=GQ(I)+HQ(IH)*XOPT(J)
                HQ(IH)=HQ(IH)+W(I)*XOPT(J)+XOPT(I)*W(J)
200     BMAT(NPT+I,J)=BMAT(NPT+J,I)
        DO 210 J=1,N
            XBASE(J)=XBASE(J)+XOPT(J)
210     XOPT(J)=ZERO
        XOPTSQ=ZERO
    END IF

    !     Pick the model step if KNEW is positive. A different choice of D
    !     may be made later, if the choice of D by BIGLAG causes substantial
    !     cancellation in DENOM.

    IF (KNEW .GT. 0) THEN
        CALL BIGLAG (N,NPT,XOPT,XPT,BMAT,ZMAT,IDZ,NDIM,KNEW,DSTEP, &
            D,ALPHA,VLAG,VLAG(NPT+1),W,W(NP),W(NP+N))
    END IF

    !     Calculate VLAG and BETA for the current choice of D. The first NPT
    !     components of W_check will be held in W.

    DO 230 K=1,NPT
        SUMA=ZERO
        SUMB=ZERO
        SUM=ZERO
        DO 220 J=1,N
            SUMA=SUMA+XPT(K,J)*D(J)
            SUMB=SUMB+XPT(K,J)*XOPT(J)
220     SUM=SUM+BMAT(K,J)*D(J)
        W(K)=SUMA*(HALF*SUMA+SUMB)
230 VLAG(K)=SUM
    BETA=ZERO
    DO 250 K=1,NPTM
        SUM=ZERO
        DO 240 I=1,NPT
240     SUM=SUM+ZMAT(I,K)*W(I)
        IF (K .LT. IDZ) THEN
            BETA=BETA+SUM*SUM
            SUM=-SUM
        ELSE
            BETA=BETA-SUM*SUM
        END IF
        DO 250 I=1,NPT
250 VLAG(I)=VLAG(I)+SUM*ZMAT(I,K)
    BSUM=ZERO
    DX=ZERO
    DO 280 J=1,N
        SUM=ZERO
        DO 260 I=1,NPT
260     SUM=SUM+W(I)*BMAT(I,J)
        BSUM=BSUM+SUM*D(J)
        JP=NPT+J
        DO 270 K=1,N
270     SUM=SUM+BMAT(JP,K)*D(K)
        VLAG(JP)=SUM
        BSUM=BSUM+SUM*D(J)
280 DX=DX+D(J)*XOPT(J)
    BETA=DX*DX+DSQ*(XOPTSQ+DX+DX+HALF*DSQ)+BETA-BSUM
    VLAG(KOPT)=VLAG(KOPT)+ONE

    !     If KNEW is positive and if the cancellation in DENOM is unacceptable,
    !     then BIGDEN calculates an alternative model step, XNEW being used for
    !     working space.

    IF (KNEW .GT. 0) THEN
        TEMP=ONE+ALPHA*BETA/VLAG(KNEW)**2
        IF (DABS(TEMP) .LE. 0.8D0) THEN
            CALL BIGDEN (N,NPT,XOPT,XPT,BMAT,ZMAT,IDZ,NDIM,KOPT, &
                KNEW,D,W,VLAG,BETA,XNEW,W(NDIM+1),W(6*NDIM+1))
        END IF
    END IF

    !     Calculate the next value of the objective function.

290 DO 300 I=1,N
        XNEW(I)=XOPT(I)+D(I)
300 X(I)=XBASE(I)+XNEW(I)
    NF=NF+1
310 IF (NF .GT. NFTEST) THEN
        NF=NF-1
        IF (IPRINT .GT. 0) PRINT 320
320     FORMAT (/4X,'Return from NEWUOA because CALFUN has been',' called MAXFUN times.')
        GOTO 530
    END IF
    CALL CALFUN (N,X,F)
    IF (IPRINT .EQ. 3) THEN
        PRINT 330, NF,F,(X(I),I=1,N)
330     FORMAT (/4X,'Function number',I6,'    F =',1PD18.10, &
            '    The corresponding X is:'/(2X,5D15.6))
    END IF
    IF (NF .LE. NPT) GOTO 70
    IF (KNEW .EQ. -1) GOTO 530

    !     Use the quadratic model to predict the change in F due to the step D,
    !     and set DIFF to the error of this prediction.

    VQUAD=ZERO
    IH=0
    DO 340 J=1,N
        VQUAD=VQUAD+D(J)*GQ(J)
        DO 340 I=1,J
            IH=IH+1
            TEMP=D(I)*XNEW(J)+D(J)*XOPT(I)
            IF (I .EQ. J) TEMP=HALF*TEMP
340 VQUAD=VQUAD+TEMP*HQ(IH)
    DO 350 K=1,NPT
350 VQUAD=VQUAD+PQ(K)*W(K)
    DIFF=F-FOPT-VQUAD
    DIFFC=DIFFB
    DIFFB=DIFFA
    DIFFA=DABS(DIFF)
    IF (DNORM .GT. RHO) NFSAV=NF

    !     Update FOPT and XOPT if the new F is the least value of the objective
    !     function so far. The branch when KNEW is positive occurs if D is not
    !     a trust region step.

    FSAVE=FOPT
    IF (F .LT. FOPT) THEN
        FOPT=F
        XOPTSQ=ZERO
        DO 360 I=1,N
            XOPT(I)=XNEW(I)
360     XOPTSQ=XOPTSQ+XOPT(I)**2
    END IF
    KSAVE=KNEW
    IF (KNEW .GT. 0) GOTO 410

    !     Pick the next value of DELTA after a trust region step.

    IF (VQUAD .GE. ZERO) THEN
        IF (IPRINT .GT. 0) PRINT 370
370     FORMAT (/4X,'Return from NEWUOA because a trust', &
            ' region step has failed to reduce Q.')
        GOTO 530
    END IF
    RATIO=(F-FSAVE)/VQUAD
    IF (RATIO .LE. TENTH) THEN
        DELTA=HALF*DNORM
    ELSE IF (RATIO .LE. 0.7D0) THEN
        DELTA=DMAX1(HALF*DELTA,DNORM)
    ELSE
        DELTA=DMAX1(HALF*DELTA,DNORM+DNORM)
    END IF
    IF (DELTA .LE. 1.5D0*RHO) DELTA=RHO

    !     Set KNEW to the index of the next interpolation point to be deleted.

    RHOSQ=DMAX1(TENTH*DELTA,RHO)**2
    KTEMP=0
    DETRAT=ZERO
    IF (F .GE. FSAVE) THEN
        KTEMP=KOPT
        DETRAT=ONE
    END IF
    DO 400 K=1,NPT
        HDIAG=ZERO
        DO 380 J=1,NPTM
            TEMP=ONE
            IF (J .LT. IDZ) TEMP=-ONE
380     HDIAG=HDIAG+TEMP*ZMAT(K,J)**2
        TEMP=DABS(BETA*HDIAG+VLAG(K)**2)
        DISTSQ=ZERO
        DO 390 J=1,N
390     DISTSQ=DISTSQ+(XPT(K,J)-XOPT(J))**2
        IF (DISTSQ .GT. RHOSQ) TEMP=TEMP*(DISTSQ/RHOSQ)**3
        IF (TEMP .GT. DETRAT .AND. K .NE. KTEMP) THEN
            DETRAT=TEMP
            KNEW=K
        END IF
400 CONTINUE
    IF (KNEW .EQ. 0) GOTO 460

    !     Update BMAT, ZMAT and IDZ, so that the KNEW-th interpolation point
    !     can be moved. Begin the updating of the quadratic model, starting
    !     with the explicit second derivative term.

410 CALL UPDATE (N,NPT,BMAT,ZMAT,IDZ,NDIM,VLAG,BETA,KNEW,W)
    FVAL(KNEW)=F
    IH=0
    DO 420 I=1,N
        TEMP=PQ(KNEW)*XPT(KNEW,I)
        DO 420 J=1,I
            IH=IH+1
420 HQ(IH)=HQ(IH)+TEMP*XPT(KNEW,J)
    PQ(KNEW)=ZERO

    !     Update the other second derivative parameters, and then the gradient
    !     vector of the model. Also include the new interpolation point.

    DO 440 J=1,NPTM
        TEMP=DIFF*ZMAT(KNEW,J)
        IF (J .LT. IDZ) TEMP=-TEMP
        DO 440 K=1,NPT
440 PQ(K)=PQ(K)+TEMP*ZMAT(K,J)
    GQSQ=ZERO
    DO 450 I=1,N
        GQ(I)=GQ(I)+DIFF*BMAT(KNEW,I)
        GQSQ=GQSQ+GQ(I)**2
450 XPT(KNEW,I)=XNEW(I)

    !     If a trust region step makes a small change to the objective function,
    !     then calculate the gradient of the least Frobenius norm interpolant at
    !     XBASE, and store it in W, using VLAG for a vector of right hand sides.

    IF (KSAVE .EQ. 0 .AND. DELTA .EQ. RHO) THEN
        IF (DABS(RATIO) .GT. 1.0D-2) THEN
            ITEST=0
        ELSE
            DO 700 K=1,NPT
700         VLAG(K)=FVAL(K)-FVAL(KOPT)
            GISQ=ZERO
            DO 720 I=1,N
                SUM=ZERO
                DO 710 K=1,NPT
710             SUM=SUM+BMAT(K,I)*VLAG(K)
                GISQ=GISQ+SUM*SUM
720         W(I)=SUM

            !     Test whether to replace the new quadratic model by the least Frobenius
            !     norm interpolant, making the replacement if the test is satisfied.

            ITEST=ITEST+1
            IF (GQSQ .LT. 1.0D2*GISQ) ITEST=0
            IF (ITEST .GE. 3) THEN
                DO 730 I=1,N
730             GQ(I)=W(I)
                DO 740 IH=1,NH
740             HQ(IH)=ZERO
                DO 760 J=1,NPTM
                    W(J)=ZERO
                    DO 750 K=1,NPT
750                 W(J)=W(J)+VLAG(K)*ZMAT(K,J)
760             IF (J .LT. IDZ) W(J)=-W(J)
                DO 770 K=1,NPT
                    PQ(K)=ZERO
                    DO 770 J=1,NPTM
770             PQ(K)=PQ(K)+ZMAT(K,J)*W(J)
                ITEST=0
            END IF
        END IF
    END IF
    IF (F .LT. FSAVE) KOPT=KNEW

    !     If a trust region step has provided a sufficient decrease in F, then
    !     branch for another trust region calculation. The case KSAVE>0 occurs
    !     when the new function value was calculated by a model step.

    IF (F .LE. FSAVE+TENTH*VQUAD) GOTO 100
    IF (KSAVE .GT. 0) GOTO 100

    !     Alternatively, find out if the interpolation points are close enough
    !     to the best point so far.

    KNEW=0
460 DISTSQ=4.0D0*DELTA*DELTA
    DO 480 K=1,NPT
        SUM=ZERO
        DO 470 J=1,N
470     SUM=SUM+(XPT(K,J)-XOPT(J))**2
        IF (SUM .GT. DISTSQ) THEN
            KNEW=K
            DISTSQ=SUM
        END IF
480 CONTINUE

    !     If KNEW is positive, then set DSTEP, and branch back for the next
    !     iteration, which will generate a "model step".

    IF (KNEW .GT. 0) THEN
        DSTEP=DMAX1(DMIN1(TENTH*DSQRT(DISTSQ),HALF*DELTA),RHO)
        DSQ=DSTEP*DSTEP
        GOTO 120
    END IF
    IF (RATIO .GT. ZERO) GOTO 100
    IF (DMAX1(DELTA,DNORM) .GT. RHO) GOTO 100

    !     The calculations with the current value of RHO are complete. Pick the
    !     next values of RHO and DELTA.

490 IF (RHO .GT. RHOEND) THEN
        DELTA=HALF*RHO
        RATIO=RHO/RHOEND
        IF (RATIO .LE. 16.0D0) THEN
            RHO=RHOEND
        ELSE IF (RATIO .LE. 250.0D0) THEN
            RHO=DSQRT(RATIO)*RHOEND
        ELSE
            RHO=TENTH*RHO
        END IF
        DELTA=DMAX1(DELTA,RHO)
        IF (IPRINT .GE. 2) THEN
            IF (IPRINT .GE. 3) PRINT 500
500         FORMAT (5X)
            PRINT 510, RHO,NF
510         FORMAT (/4X,'New RHO =',1PD11.4,5X,'Number of', &
                ' function values =',I6)
            PRINT 520, FOPT,(XBASE(I)+XOPT(I),I=1,N)
520         FORMAT (4X,'Least value of F =',1PD23.15,9X, &
                'The corresponding X is:'/(2X,5D15.6))
        END IF
        GOTO 90
    END IF

    !     Return from the calculation, after another Newton-Raphson step, if
    !     it is too short to have been tried before.

    IF (KNEW .EQ. -1) GOTO 290
530 IF (FOPT .LE. F) THEN
        DO 540 I=1,N
540     X(I)=XBASE(I)+XOPT(I)
        F=FOPT
    END IF
    IF (IPRINT .GE. 1) THEN
        PRINT 550, NF
550     FORMAT (/4X,'At the return from NEWUOA',5X, &
            'Number of function values =',I6)
        PRINT 520, F,(X(I),I=1,N)
    END IF
    RETURN
    END SUBROUTINE NEWUOB


    SUBROUTINE TRSAPP (N,NPT,XOPT,XPT,GQ,HQ,PQ,DELTA,STEP, &
        D,G,HD,HS,CRVMIN)
    IMPLICIT REAL*8 (A-H,O-Z)
    DIMENSION XOPT(*),XPT(NPT,*),GQ(*),HQ(*),PQ(*),STEP(*), &
        D(*),G(*),HD(*),HS(*)

    !     N is the number of variables of a quadratic objective function, Q say.
    !     The arguments NPT, XOPT, XPT, GQ, HQ and PQ have their usual meanings,
    !       in order to define the current quadratic model Q.
    !     DELTA is the trust region radius, and has to be positive.
    !     STEP will be set to the calculated trial step.
    !     The arrays D, G, HD and HS will be used for working space.
    !     CRVMIN will be set to the least curvature of H along the conjugate
    !       directions that occur, except that it is set to zero if STEP goes
    !       all the way to the trust region boundary.

    !     The calculation of STEP begins with the truncated conjugate gradient
    !     method. If the boundary of the trust region is reached, then further
    !     changes to STEP may be made, each one being in the 2D space spanned
    !     by the current STEP and the corresponding gradient of Q. Thus STEP
    !     should provide a substantial reduction to Q within the trust region.

    !     Initialization, which includes setting HD to H times XOPT.

    HALF=0.5D0
    ZERO=0.0D0
    TWOPI=8.0D0*DATAN(1.0D0)
    DELSQ=DELTA*DELTA
    ITERC=0
    ITERMAX=N
    ITERSW=ITERMAX
    DO 10 I=1,N
10  D(I)=XOPT(I)
    GOTO 170

    !     Prepare for the first line search.

20  QRED=ZERO
    DD=ZERO
    DO 30 I=1,N
        STEP(I)=ZERO
        HS(I)=ZERO
        G(I)=GQ(I)+HD(I)
        D(I)=-G(I)
30  DD=DD+D(I)**2
    CRVMIN=ZERO
    IF (DD .EQ. ZERO) GOTO 160
    DS=ZERO
    SS=ZERO
    GG=DD
    GGBEG=GG

    !     Calculate the step to the trust region boundary and the product HD.

40  ITERC=ITERC+1
    TEMP=DELSQ-SS
    BSTEP=TEMP/(DS+DSQRT(DS*DS+DD*TEMP))
    GOTO 170
50  DHD=ZERO
    DO 60 J=1,N
60  DHD=DHD+D(J)*HD(J)

    !     Update CRVMIN and set the step-length ALPHA.

    ALPHA=BSTEP
    IF (DHD .GT. ZERO) THEN
        TEMP=DHD/DD
        IF (ITERC .EQ. 1) CRVMIN=TEMP
        CRVMIN=DMIN1(CRVMIN,TEMP)
        ALPHA=DMIN1(ALPHA,GG/DHD)
    END IF
    QADD=ALPHA*(GG-HALF*ALPHA*DHD)
    QRED=QRED+QADD

    !     Update STEP and HS.

    GGSAV=GG
    GG=ZERO
    DO 70 I=1,N
        STEP(I)=STEP(I)+ALPHA*D(I)
        HS(I)=HS(I)+ALPHA*HD(I)
70  GG=GG+(G(I)+HS(I))**2

    !     Begin another conjugate direction iteration if required.

    IF (ALPHA .LT. BSTEP) THEN
        IF (QADD .LE. 0.01D0*QRED) GOTO 160
        IF (GG .LE. 1.0D-4*GGBEG) GOTO 160
        IF (ITERC .EQ. ITERMAX) GOTO 160
        TEMP=GG/GGSAV
        DD=ZERO
        DS=ZERO
        SS=ZERO
        DO 80 I=1,N
            D(I)=TEMP*D(I)-G(I)-HS(I)
            DD=DD+D(I)**2
            DS=DS+D(I)*STEP(I)
80      SS=SS+STEP(I)**2
        IF (DS .LE. ZERO) GOTO 160
        IF (SS .LT. DELSQ) GOTO 40
    END IF
    CRVMIN=ZERO
    ITERSW=ITERC

    !     Test whether an alternative iteration is required.

90  IF (GG .LE. 1.0D-4*GGBEG) GOTO 160
    SG=ZERO
    SHS=ZERO
    DO 100 I=1,N
        SG=SG+STEP(I)*G(I)
100 SHS=SHS+STEP(I)*HS(I)
    SGK=SG+SHS
    ANGTEST=SGK/DSQRT(GG*DELSQ)
    IF (ANGTEST .LE. -0.99D0) GOTO 160

    !     Begin the alternative iteration by calculating D and HD and some
    !     scalar products.

    ITERC=ITERC+1
    TEMP=DSQRT(DELSQ*GG-SGK*SGK)
    TEMPA=DELSQ/TEMP
    TEMPB=SGK/TEMP
    DO 110 I=1,N
110 D(I)=TEMPA*(G(I)+HS(I))-TEMPB*STEP(I)
    GOTO 170
120 DG=ZERO
    DHD=ZERO
    DHS=ZERO
    DO 130 I=1,N
        DG=DG+D(I)*G(I)
        DHD=DHD+HD(I)*D(I)
130 DHS=DHS+HD(I)*STEP(I)

    !     Seek the value of the angle that minimizes Q.

    CF=HALF*(SHS-DHD)
    QBEG=SG+CF
    QSAV=QBEG
    QMIN=QBEG
    ISAVE=0
    IU=49
    TEMP=TWOPI/DFLOAT(IU+1)
    DO 140 I=1,IU
        ANGLE=DFLOAT(I)*TEMP
        CTH=DCOS(ANGLE)
        STH=DSIN(ANGLE)
        QNEW=(SG+CF*CTH)*CTH+(DG+DHS*CTH)*STH
        IF (QNEW .LT. QMIN) THEN
            QMIN=QNEW
            ISAVE=I
            TEMPA=QSAV
        ELSE IF (I .EQ. ISAVE+1) THEN
            TEMPB=QNEW
        END IF
140 QSAV=QNEW
    IF (ISAVE .EQ. ZERO) TEMPA=QNEW
    IF (ISAVE .EQ. IU) TEMPB=QBEG
    ANGLE=ZERO
    IF (TEMPA .NE. TEMPB) THEN
        TEMPA=TEMPA-QMIN
        TEMPB=TEMPB-QMIN
        ANGLE=HALF*(TEMPA-TEMPB)/(TEMPA+TEMPB)
    END IF
    ANGLE=TEMP*(DFLOAT(ISAVE)+ANGLE)

    !     Calculate the new STEP and HS. Then test for convergence.

    CTH=DCOS(ANGLE)
    STH=DSIN(ANGLE)
    REDUC=QBEG-(SG+CF*CTH)*CTH-(DG+DHS*CTH)*STH
    GG=ZERO
    DO 150 I=1,N
        STEP(I)=CTH*STEP(I)+STH*D(I)
        HS(I)=CTH*HS(I)+STH*HD(I)
150 GG=GG+(G(I)+HS(I))**2
    QRED=QRED+REDUC
    RATIO=REDUC/QRED
    IF (ITERC .LT. ITERMAX .AND. RATIO .GT. 0.01D0) GOTO 90
160 RETURN

    !     The following instructions act as a subroutine for setting the vector
    !     HD to the vector D multiplied by the second derivative matrix of Q.
    !     They are called from three different places, which are distinguished
    !     by the value of ITERC.

170 DO 180 I=1,N
180 HD(I)=ZERO
    DO 200 K=1,NPT
        TEMP=ZERO
        DO 190 J=1,N
190     TEMP=TEMP+XPT(K,J)*D(J)
        TEMP=TEMP*PQ(K)
        DO 200 I=1,N
200 HD(I)=HD(I)+TEMP*XPT(K,I)
    IH=0
    DO 210 J=1,N
        DO 210 I=1,J
            IH=IH+1
            IF (I .LT. J) HD(J)=HD(J)+HQ(IH)*D(I)
210 HD(I)=HD(I)+HQ(IH)*D(J)
    IF (ITERC .EQ. 0) GOTO 20
    IF (ITERC .LE. ITERSW) GOTO 50
    GOTO 120
    END SUBROUTINE TRSAPP

    SUBROUTINE BIGLAG (N,NPT,XOPT,XPT,BMAT,ZMAT,IDZ,NDIM,KNEW, &
        DELTA,D,ALPHA,HCOL,GC,GD,S,W)
    IMPLICIT REAL*8 (A-H,O-Z)
    DIMENSION XOPT(*),XPT(NPT,*),BMAT(NDIM,*),ZMAT(NPT,*),D(*), &
        HCOL(*),GC(*),GD(*),S(*),W(*)

    !     N is the number of variables.
    !     NPT is the number of interpolation equations.
    !     XOPT is the best interpolation point so far.
    !     XPT contains the coordinates of the current interpolation points.
    !     BMAT provides the last N columns of H.
    !     ZMAT and IDZ give a factorization of the first NPT by NPT submatrix of H.
    !     NDIM is the first dimension of BMAT and has the value NPT+N.
    !     KNEW is the index of the interpolation point that is going to be moved.
    !     DELTA is the current trust region bound.
    !     D will be set to the step from XOPT to the new point.
    !     ALPHA will be set to the KNEW-th diagonal element of the H matrix.
    !     HCOL, GC, GD, S and W will be used for working space.

    !     The step D is calculated in a way that attempts to maximize the modulus
    !     of LFUNC(XOPT+D), subject to the bound ||D|| .LE. DELTA, where LFUNC is
    !     the KNEW-th Lagrange function.

    !     Set some constants.

    HALF=0.5D0
    ONE=1.0D0
    ZERO=0.0D0
    TWOPI=8.0D0*DATAN(ONE)
    DELSQ=DELTA*DELTA
    NPTM=NPT-N-1

    !     Set the first NPT components of HCOL to the leading elements of the
    !     KNEW-th column of H.

    ITERC=0
    DO 10 K=1,NPT
10  HCOL(K)=ZERO
    DO 20 J=1,NPTM
        TEMP=ZMAT(KNEW,J)
        IF (J .LT. IDZ) TEMP=-TEMP
        DO 20 K=1,NPT
20  HCOL(K)=HCOL(K)+TEMP*ZMAT(K,J)
    ALPHA=HCOL(KNEW)

    !     Set the unscaled initial direction D. Form the gradient of LFUNC at
    !     XOPT, and multiply D by the second derivative matrix of LFUNC.

    DD=ZERO
    DO 30 I=1,N
        D(I)=XPT(KNEW,I)-XOPT(I)
        GC(I)=BMAT(KNEW,I)
        GD(I)=ZERO
30  DD=DD+D(I)**2
    DO 50 K=1,NPT
        TEMP=ZERO
        SUM=ZERO
        DO 40 J=1,N
            TEMP=TEMP+XPT(K,J)*XOPT(J)
40      SUM=SUM+XPT(K,J)*D(J)
        TEMP=HCOL(K)*TEMP
        SUM=HCOL(K)*SUM
        DO 50 I=1,N
            GC(I)=GC(I)+TEMP*XPT(K,I)
50  GD(I)=GD(I)+SUM*XPT(K,I)

    !     Scale D and GD, with a sign change if required. Set S to another
    !     vector in the initial two dimensional subspace.

    GG=ZERO
    SP=ZERO
    DHD=ZERO
    DO 60 I=1,N
        GG=GG+GC(I)**2
        SP=SP+D(I)*GC(I)
60  DHD=DHD+D(I)*GD(I)
    SCALE=DELTA/DSQRT(DD)
    IF (SP*DHD .LT. ZERO) SCALE=-SCALE
    TEMP=ZERO
    IF (SP*SP .GT. 0.99D0*DD*GG) TEMP=ONE
    TAU=SCALE*(DABS(SP)+HALF*SCALE*DABS(DHD))
    IF (GG*DELSQ .LT. 0.01D0*TAU*TAU) TEMP=ONE
    DO 70 I=1,N
        D(I)=SCALE*D(I)
        GD(I)=SCALE*GD(I)
70  S(I)=GC(I)+TEMP*GD(I)

    !     Begin the iteration by overwriting S with a vector that has the
    !     required length and direction, except that termination occurs if
    !     the given D and S are nearly parallel.

80  ITERC=ITERC+1
    DD=ZERO
    SP=ZERO
    SS=ZERO
    DO 90 I=1,N
        DD=DD+D(I)**2
        SP=SP+D(I)*S(I)
90  SS=SS+S(I)**2
    TEMP=DD*SS-SP*SP
    IF (TEMP .LE. 1.0D-8*DD*SS) GOTO 160
    DENOM=DSQRT(TEMP)
    DO 100 I=1,N
        S(I)=(DD*S(I)-SP*D(I))/DENOM
100 W(I)=ZERO

    !     Calculate the coefficients of the objective function on the circle,
    !     beginning with the multiplication of S by the second derivative matrix.

    DO 120 K=1,NPT
        SUM=ZERO
        DO 110 J=1,N
110     SUM=SUM+XPT(K,J)*S(J)
        SUM=HCOL(K)*SUM
        DO 120 I=1,N
120 W(I)=W(I)+SUM*XPT(K,I)
    CF1=ZERO
    CF2=ZERO
    CF3=ZERO
    CF4=ZERO
    CF5=ZERO
    DO 130 I=1,N
        CF1=CF1+S(I)*W(I)
        CF2=CF2+D(I)*GC(I)
        CF3=CF3+S(I)*GC(I)
        CF4=CF4+D(I)*GD(I)
130 CF5=CF5+S(I)*GD(I)
    CF1=HALF*CF1
    CF4=HALF*CF4-CF1

    !     Seek the value of the angle that maximizes the modulus of TAU.

    TAUBEG=CF1+CF2+CF4
    TAUMAX=TAUBEG
    TAUOLD=TAUBEG
    ISAVE=0
    IU=49
    TEMP=TWOPI/DFLOAT(IU+1)
    DO 140 I=1,IU
        ANGLE=DFLOAT(I)*TEMP
        CTH=DCOS(ANGLE)
        STH=DSIN(ANGLE)
        TAU=CF1+(CF2+CF4*CTH)*CTH+(CF3+CF5*CTH)*STH
        IF (DABS(TAU) .GT. DABS(TAUMAX)) THEN
            TAUMAX=TAU
            ISAVE=I
            TEMPA=TAUOLD
        ELSE IF (I .EQ. ISAVE+1) THEN
            TEMPB=TAU
        END IF
140 TAUOLD=TAU
    IF (ISAVE .EQ. 0) TEMPA=TAU
    IF (ISAVE .EQ. IU) TEMPB=TAUBEG
    STEP=ZERO
    IF (TEMPA .NE. TEMPB) THEN
        TEMPA=TEMPA-TAUMAX
        TEMPB=TEMPB-TAUMAX
        STEP=HALF*(TEMPA-TEMPB)/(TEMPA+TEMPB)
    END IF
    ANGLE=TEMP*(DFLOAT(ISAVE)+STEP)

    !     Calculate the new D and GD. Then test for convergence.

    CTH=DCOS(ANGLE)
    STH=DSIN(ANGLE)
    TAU=CF1+(CF2+CF4*CTH)*CTH+(CF3+CF5*CTH)*STH
    DO 150 I=1,N
        D(I)=CTH*D(I)+STH*S(I)
        GD(I)=CTH*GD(I)+STH*W(I)
150 S(I)=GC(I)+GD(I)
    IF (DABS(TAU) .LE. 1.1D0*DABS(TAUBEG)) GOTO 160
    IF (ITERC .LT. N) GOTO 80
160 RETURN
    END SUBROUTINE BIGLAG


    SUBROUTINE BIGDEN (N,NPT,XOPT,XPT,BMAT,ZMAT,IDZ,NDIM,KOPT, &
        KNEW,D,W,VLAG,BETA,S,WVEC,PROD)
    IMPLICIT REAL*8 (A-H,O-Z)
    DIMENSION XOPT(*),XPT(NPT,*),BMAT(NDIM,*),ZMAT(NPT,*),D(*), &
        W(*),VLAG(*),S(*),WVEC(NDIM,*),PROD(NDIM,*)
    DIMENSION DEN(9),DENEX(9),PAR(9)

    !     N is the number of variables.
    !     NPT is the number of interpolation equations.
    !     XOPT is the best interpolation point so far.
    !     XPT contains the coordinates of the current interpolation points.
    !     BMAT provides the last N columns of H.
    !     ZMAT and IDZ give a factorization of the first NPT by NPT submatrix of H.
    !     NDIM is the first dimension of BMAT and has the value NPT+N.
    !     KOPT is the index of the optimal interpolation point.
    !     KNEW is the index of the interpolation point that is going to be moved.
    !     D will be set to the step from XOPT to the new point, and on entry it
    !       should be the D that was calculated by the last call of BIGLAG. The
    !       length of the initial D provides a trust region bound on the final D.
    !     W will be set to Wcheck for the final choice of D.
    !     VLAG will be set to Theta*Wcheck+e_b for the final choice of D.
    !     BETA will be set to the value that will occur in the updating formula
    !       when the KNEW-th interpolation point is moved to its new position.
    !     S, WVEC, PROD and the private arrays DEN, DENEX and PAR will be used
    !       for working space.

    !     D is calculated in a way that should provide a denominator with a large
    !     modulus in the updating formula when the KNEW-th interpolation point is
    !     shifted to the new position XOPT+D.

    !     Set some constants.

    HALF=0.5D0
    ONE=1.0D0
    QUART=0.25D0
    TWO=2.0D0
    ZERO=0.0D0
    TWOPI=8.0D0*DATAN(ONE)
    NPTM=NPT-N-1

    !     Store the first NPT elements of the KNEW-th column of H in W(N+1)
    !     to W(N+NPT).

    DO 10 K=1,NPT
10  W(N+K)=ZERO
    DO 20 J=1,NPTM
        TEMP=ZMAT(KNEW,J)
        IF (J .LT. IDZ) TEMP=-TEMP
        DO 20 K=1,NPT
20  W(N+K)=W(N+K)+TEMP*ZMAT(K,J)
    ALPHA=W(N+KNEW)

    !     The initial search direction D is taken from the last call of BIGLAG,
    !     and the initial S is set below, usually to the direction from X_OPT
    !     to X_KNEW, but a different direction to an interpolation point may
    !     be chosen, in order to prevent S from being nearly parallel to D.

    DD=ZERO
    DS=ZERO
    SS=ZERO
    XOPTSQ=ZERO
    DO 30 I=1,N
        DD=DD+D(I)**2
        S(I)=XPT(KNEW,I)-XOPT(I)
        DS=DS+D(I)*S(I)
        SS=SS+S(I)**2
30  XOPTSQ=XOPTSQ+XOPT(I)**2
    IF (DS*DS .GT. 0.99D0*DD*SS) THEN
        KSAV=KNEW
        DTEST=DS*DS/SS
        DO 50 K=1,NPT
            IF (K .NE. KOPT) THEN
                DSTEMP=ZERO
                SSTEMP=ZERO
                DO 40 I=1,N
                    DIFF=XPT(K,I)-XOPT(I)
                    DSTEMP=DSTEMP+D(I)*DIFF
40              SSTEMP=SSTEMP+DIFF*DIFF
                IF (DSTEMP*DSTEMP/SSTEMP .LT. DTEST) THEN
                    KSAV=K
                    DTEST=DSTEMP*DSTEMP/SSTEMP
                    DS=DSTEMP
                    SS=SSTEMP
                END IF
            END IF
50      CONTINUE
        DO 60 I=1,N
60      S(I)=XPT(KSAV,I)-XOPT(I)
    END IF
    SSDEN=DD*SS-DS*DS
    ITERC=0
    DENSAV=ZERO

    !     Begin the iteration by overwriting S with a vector that has the
    !     required length and direction.

70  ITERC=ITERC+1
    TEMP=ONE/DSQRT(SSDEN)
    XOPTD=ZERO
    XOPTS=ZERO
    DO 80 I=1,N
        S(I)=TEMP*(DD*S(I)-DS*D(I))
        XOPTD=XOPTD+XOPT(I)*D(I)
80  XOPTS=XOPTS+XOPT(I)*S(I)

    !     Set the coefficients of the first two terms of BETA.

    TEMPA=HALF*XOPTD*XOPTD
    TEMPB=HALF*XOPTS*XOPTS
    DEN(1)=DD*(XOPTSQ+HALF*DD)+TEMPA+TEMPB
    DEN(2)=TWO*XOPTD*DD
    DEN(3)=TWO*XOPTS*DD
    DEN(4)=TEMPA-TEMPB
    DEN(5)=XOPTD*XOPTS
    DO 90 I=6,9
90  DEN(I)=ZERO

    !     Put the coefficients of Wcheck in WVEC.

    DO 110 K=1,NPT
        TEMPA=ZERO
        TEMPB=ZERO
        TEMPC=ZERO
        DO 100 I=1,N
            TEMPA=TEMPA+XPT(K,I)*D(I)
            TEMPB=TEMPB+XPT(K,I)*S(I)
100     TEMPC=TEMPC+XPT(K,I)*XOPT(I)
        WVEC(K,1)=QUART*(TEMPA*TEMPA+TEMPB*TEMPB)
        WVEC(K,2)=TEMPA*TEMPC
        WVEC(K,3)=TEMPB*TEMPC
        WVEC(K,4)=QUART*(TEMPA*TEMPA-TEMPB*TEMPB)
110 WVEC(K,5)=HALF*TEMPA*TEMPB
    DO 120 I=1,N
        IP=I+NPT
        WVEC(IP,1)=ZERO
        WVEC(IP,2)=D(I)
        WVEC(IP,3)=S(I)
        WVEC(IP,4)=ZERO
120 WVEC(IP,5)=ZERO

    !     Put the coefficents of THETA*Wcheck in PROD.

    DO 190 JC=1,5
        NW=NPT
        IF (JC .EQ. 2 .OR. JC .EQ. 3) NW=NDIM
        DO 130 K=1,NPT
130     PROD(K,JC)=ZERO
        DO 150 J=1,NPTM
            SUM=ZERO
            DO 140 K=1,NPT
140         SUM=SUM+ZMAT(K,J)*WVEC(K,JC)
            IF (J .LT. IDZ) SUM=-SUM
            DO 150 K=1,NPT
150     PROD(K,JC)=PROD(K,JC)+SUM*ZMAT(K,J)
        IF (NW .EQ. NDIM) THEN
            DO 170 K=1,NPT
                SUM=ZERO
                DO 160 J=1,N
160             SUM=SUM+BMAT(K,J)*WVEC(NPT+J,JC)
170         PROD(K,JC)=PROD(K,JC)+SUM
        END IF
        DO 190 J=1,N
            SUM=ZERO
            DO 180 I=1,NW
180         SUM=SUM+BMAT(I,J)*WVEC(I,JC)
190 PROD(NPT+J,JC)=SUM

    !     Include in DEN the part of BETA that depends on THETA.

    DO 210 K=1,NDIM
        SUM=ZERO
        DO 200 I=1,5
            PAR(I)=HALF*PROD(K,I)*WVEC(K,I)
200     SUM=SUM+PAR(I)
        DEN(1)=DEN(1)-PAR(1)-SUM
        TEMPA=PROD(K,1)*WVEC(K,2)+PROD(K,2)*WVEC(K,1)
        TEMPB=PROD(K,2)*WVEC(K,4)+PROD(K,4)*WVEC(K,2)
        TEMPC=PROD(K,3)*WVEC(K,5)+PROD(K,5)*WVEC(K,3)
        DEN(2)=DEN(2)-TEMPA-HALF*(TEMPB+TEMPC)
        DEN(6)=DEN(6)-HALF*(TEMPB-TEMPC)
        TEMPA=PROD(K,1)*WVEC(K,3)+PROD(K,3)*WVEC(K,1)
        TEMPB=PROD(K,2)*WVEC(K,5)+PROD(K,5)*WVEC(K,2)
        TEMPC=PROD(K,3)*WVEC(K,4)+PROD(K,4)*WVEC(K,3)
        DEN(3)=DEN(3)-TEMPA-HALF*(TEMPB-TEMPC)
        DEN(7)=DEN(7)-HALF*(TEMPB+TEMPC)
        TEMPA=PROD(K,1)*WVEC(K,4)+PROD(K,4)*WVEC(K,1)
        DEN(4)=DEN(4)-TEMPA-PAR(2)+PAR(3)
        TEMPA=PROD(K,1)*WVEC(K,5)+PROD(K,5)*WVEC(K,1)
        TEMPB=PROD(K,2)*WVEC(K,3)+PROD(K,3)*WVEC(K,2)
        DEN(5)=DEN(5)-TEMPA-HALF*TEMPB
        DEN(8)=DEN(8)-PAR(4)+PAR(5)
        TEMPA=PROD(K,4)*WVEC(K,5)+PROD(K,5)*WVEC(K,4)
210 DEN(9)=DEN(9)-HALF*TEMPA

    !     Extend DEN so that it holds all the coefficients of DENOM.

    SUM=ZERO
    DO 220 I=1,5
        PAR(I)=HALF*PROD(KNEW,I)**2
220 SUM=SUM+PAR(I)
    DENEX(1)=ALPHA*DEN(1)+PAR(1)+SUM
    TEMPA=TWO*PROD(KNEW,1)*PROD(KNEW,2)
    TEMPB=PROD(KNEW,2)*PROD(KNEW,4)
    TEMPC=PROD(KNEW,3)*PROD(KNEW,5)
    DENEX(2)=ALPHA*DEN(2)+TEMPA+TEMPB+TEMPC
    DENEX(6)=ALPHA*DEN(6)+TEMPB-TEMPC
    TEMPA=TWO*PROD(KNEW,1)*PROD(KNEW,3)
    TEMPB=PROD(KNEW,2)*PROD(KNEW,5)
    TEMPC=PROD(KNEW,3)*PROD(KNEW,4)
    DENEX(3)=ALPHA*DEN(3)+TEMPA+TEMPB-TEMPC
    DENEX(7)=ALPHA*DEN(7)+TEMPB+TEMPC
    TEMPA=TWO*PROD(KNEW,1)*PROD(KNEW,4)
    DENEX(4)=ALPHA*DEN(4)+TEMPA+PAR(2)-PAR(3)
    TEMPA=TWO*PROD(KNEW,1)*PROD(KNEW,5)
    DENEX(5)=ALPHA*DEN(5)+TEMPA+PROD(KNEW,2)*PROD(KNEW,3)
    DENEX(8)=ALPHA*DEN(8)+PAR(4)-PAR(5)
    DENEX(9)=ALPHA*DEN(9)+PROD(KNEW,4)*PROD(KNEW,5)

    !     Seek the value of the angle that maximizes the modulus of DENOM.

    SUM=DENEX(1)+DENEX(2)+DENEX(4)+DENEX(6)+DENEX(8)
    DENOLD=SUM
    DENMAX=SUM
    ISAVE=0
    IU=49
    TEMP=TWOPI/DFLOAT(IU+1)
    PAR(1)=ONE
    DO 250 I=1,IU
        ANGLE=DFLOAT(I)*TEMP
        PAR(2)=DCOS(ANGLE)
        PAR(3)=DSIN(ANGLE)
        DO 230 J=4,8,2
            PAR(J)=PAR(2)*PAR(J-2)-PAR(3)*PAR(J-1)
230     PAR(J+1)=PAR(2)*PAR(J-1)+PAR(3)*PAR(J-2)
        SUMOLD=SUM
        SUM=ZERO
        DO 240 J=1,9
240     SUM=SUM+DENEX(J)*PAR(J)
        IF (DABS(SUM) .GT. DABS(DENMAX)) THEN
            DENMAX=SUM
            ISAVE=I
            TEMPA=SUMOLD
        ELSE IF (I .EQ. ISAVE+1) THEN
            TEMPB=SUM
        END IF
250 CONTINUE
    IF (ISAVE .EQ. 0) TEMPA=SUM
    IF (ISAVE .EQ. IU) TEMPB=DENOLD
    STEP=ZERO
    IF (TEMPA .NE. TEMPB) THEN
        TEMPA=TEMPA-DENMAX
        TEMPB=TEMPB-DENMAX
        STEP=HALF*(TEMPA-TEMPB)/(TEMPA+TEMPB)
    END IF
    ANGLE=TEMP*(DFLOAT(ISAVE)+STEP)

    !     Calculate the new parameters of the denominator, the new VLAG vector
    !     and the new D. Then test for convergence.

    PAR(2)=DCOS(ANGLE)
    PAR(3)=DSIN(ANGLE)
    DO 260 J=4,8,2
        PAR(J)=PAR(2)*PAR(J-2)-PAR(3)*PAR(J-1)
260 PAR(J+1)=PAR(2)*PAR(J-1)+PAR(3)*PAR(J-2)
    BETA=ZERO
    DENMAX=ZERO
    DO 270 J=1,9
        BETA=BETA+DEN(J)*PAR(J)
270 DENMAX=DENMAX+DENEX(J)*PAR(J)
    DO 280 K=1,NDIM
        VLAG(K)=ZERO
        DO 280 J=1,5
280 VLAG(K)=VLAG(K)+PROD(K,J)*PAR(J)
    TAU=VLAG(KNEW)
    DD=ZERO
    TEMPA=ZERO
    TEMPB=ZERO
    DO 290 I=1,N
        D(I)=PAR(2)*D(I)+PAR(3)*S(I)
        W(I)=XOPT(I)+D(I)
        DD=DD+D(I)**2
        TEMPA=TEMPA+D(I)*W(I)
290 TEMPB=TEMPB+W(I)*W(I)
    IF (ITERC .GE. N) GOTO 340
    IF (ITERC .GT. 1) DENSAV=DMAX1(DENSAV,DENOLD)
    IF (DABS(DENMAX) .LE. 1.1D0*DABS(DENSAV)) GOTO 340
    DENSAV=DENMAX

    !     Set S to half the gradient of the denominator with respect to D.
    !     Then branch for the next iteration.

    DO 300 I=1,N
        TEMP=TEMPA*XOPT(I)+TEMPB*D(I)-VLAG(NPT+I)
300 S(I)=TAU*BMAT(KNEW,I)+ALPHA*TEMP
    DO 320 K=1,NPT
        SUM=ZERO
        DO 310 J=1,N
310     SUM=SUM+XPT(K,J)*W(J)
        TEMP=(TAU*W(N+K)-ALPHA*VLAG(K))*SUM
        DO 320 I=1,N
320 S(I)=S(I)+TEMP*XPT(K,I)
    SS=ZERO
    DS=ZERO
    DO 330 I=1,N
        SS=SS+S(I)**2
330 DS=DS+D(I)*S(I)
    SSDEN=DD*SS-DS*DS
    IF (SSDEN .GE. 1.0D-8*DD*SS) GOTO 70

    !     Set the vector W before the RETURN from the subroutine.

340 DO 350 K=1,NDIM
        W(K)=ZERO
        DO 350 J=1,5
350 W(K)=W(K)+WVEC(K,J)*PAR(J)
    VLAG(KOPT)=VLAG(KOPT)+ONE
    RETURN
    END      SUBROUTINE BIGDEN


    SUBROUTINE UPDATE (N,NPT,BMAT,ZMAT,IDZ,NDIM,VLAG,BETA,KNEW,W)
    IMPLICIT REAL*8 (A-H,O-Z)
    DIMENSION BMAT(NDIM,*),ZMAT(NPT,*),VLAG(*),W(*)

    !     The arrays BMAT and ZMAT with IDZ are updated, in order to shift the
    !     interpolation point that has index KNEW. On entry, VLAG contains the
    !     components of the vector Theta*Wcheck+e_b of the updating formula
    !     (6.11), and BETA holds the value of the parameter that has this name.
    !     The vector W is used for working space.

    !     Set some constants.

    ONE=1.0D0
    ZERO=0.0D0
    NPTM=NPT-N-1

    !     Apply the rotations that put zeros in the KNEW-th row of ZMAT.

    JL=1
    DO 20 J=2,NPTM
        IF (J .EQ. IDZ) THEN
            JL=IDZ
        ELSE IF (ZMAT(KNEW,J) .NE. ZERO) THEN
            TEMP=DSQRT(ZMAT(KNEW,JL)**2+ZMAT(KNEW,J)**2)
            TEMPA=ZMAT(KNEW,JL)/TEMP
            TEMPB=ZMAT(KNEW,J)/TEMP
            DO 10 I=1,NPT
                TEMP=TEMPA*ZMAT(I,JL)+TEMPB*ZMAT(I,J)
                ZMAT(I,J)=TEMPA*ZMAT(I,J)-TEMPB*ZMAT(I,JL)
10          ZMAT(I,JL)=TEMP
            ZMAT(KNEW,J)=ZERO
        END IF
20  CONTINUE

    !     Put the first NPT components of the KNEW-th column of HLAG into W,
    !     and calculate the parameters of the updating formula.

    TEMPA=ZMAT(KNEW,1)
    IF (IDZ .GE. 2) TEMPA=-TEMPA
    IF (JL .GT. 1) TEMPB=ZMAT(KNEW,JL)
    DO 30 I=1,NPT
        W(I)=TEMPA*ZMAT(I,1)
        IF (JL .GT. 1) W(I)=W(I)+TEMPB*ZMAT(I,JL)
30  CONTINUE
    ALPHA=W(KNEW)
    TAU=VLAG(KNEW)
    TAUSQ=TAU*TAU
    DENOM=ALPHA*BETA+TAUSQ
    VLAG(KNEW)=VLAG(KNEW)-ONE

    !     Complete the updating of ZMAT when there is only one nonzero element
    !     in the KNEW-th row of the new matrix ZMAT, but, if IFLAG is set to one,
    !     then the first column of ZMAT will be exchanged with another one later.

    IFLAG=0
    IF (JL .EQ. 1) THEN
        TEMP=DSQRT(DABS(DENOM))
        TEMPB=TEMPA/TEMP
        TEMPA=TAU/TEMP
        DO 40 I=1,NPT
40      ZMAT(I,1)=TEMPA*ZMAT(I,1)-TEMPB*VLAG(I)
        IF (IDZ .EQ. 1 .AND. TEMP .LT. ZERO) IDZ=2
        IF (IDZ .GE. 2 .AND. TEMP .GE. ZERO) IFLAG=1
    ELSE

        !     Complete the updating of ZMAT in the alternative case.

        JA=1
        IF (BETA .GE. ZERO) JA=JL
        JB=JL+1-JA
        TEMP=ZMAT(KNEW,JB)/DENOM
        TEMPA=TEMP*BETA
        TEMPB=TEMP*TAU
        TEMP=ZMAT(KNEW,JA)
        SCALA=ONE/DSQRT(DABS(BETA)*TEMP*TEMP+TAUSQ)
        SCALB=SCALA*DSQRT(DABS(DENOM))
        DO 50 I=1,NPT
            ZMAT(I,JA)=SCALA*(TAU*ZMAT(I,JA)-TEMP*VLAG(I))
50      ZMAT(I,JB)=SCALB*(ZMAT(I,JB)-TEMPA*W(I)-TEMPB*VLAG(I))
        IF (DENOM .LE. ZERO) THEN
            IF (BETA .LT. ZERO) IDZ=IDZ+1
            IF (BETA .GE. ZERO) IFLAG=1
        END IF
    END IF

    !     IDZ is reduced in the following case, and usually the first column
    !     of ZMAT is exchanged with a later one.

    IF (IFLAG .EQ. 1) THEN
        IDZ=IDZ-1
        DO 60 I=1,NPT
            TEMP=ZMAT(I,1)
            ZMAT(I,1)=ZMAT(I,IDZ)
60      ZMAT(I,IDZ)=TEMP
    END IF

    !     Finally, update the matrix BMAT.

    DO 70 J=1,N
        JP=NPT+J
        W(JP)=BMAT(KNEW,J)
        TEMPA=(ALPHA*VLAG(JP)-TAU*W(JP))/DENOM
        TEMPB=(-BETA*W(JP)-TAU*VLAG(JP))/DENOM
        DO 70 I=1,JP
            BMAT(I,J)=BMAT(I,J)+TEMPA*VLAG(I)+TEMPB*W(I)
            IF (I .GT. NPT) BMAT(JP,I-NPT)=BMAT(I,J)
70  CONTINUE
    RETURN
    END  SUBROUTINE UPDATE


    END MODULE Powell_Optimize
    !==============================================================================================================
    !==============================================================================================================
    !==============================================================================================================
    !==============================================================================================================

    program dinks_international    ! debut du programme

    use Genetic_Algorithm
    use Powell_Optimize

    implicit none

    interface
    function func(npar,p) result(fn_val)
    implicit none
    integer, intent(in)  :: npar
    real*8, intent(in)     :: p(npar)
    real*8                 :: fn_val
    end function func
    end interface

    INTERFACE
    SUBROUTINE calfun(n, x, fct)
    IMPLICIT NONE
    INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)
    INTEGER, INTENT(IN)    :: n
    REAL (dp), INTENT(IN)  :: x(n)
    REAL (dp), INTENT(OUT) :: fct
    END SUBROUTINE calfun
    END INTERFACE

    INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)
    REAL (dp)  :: rhobeg, rhoend
    INTEGER    ::  iprint, maxfun

    integer :: choix,choix2      ! choice

    integer :: ncountry
    parameter (ncountry=38) ! number of countries

    integer :: npar, i,k0,k1,k4,status, seed
    parameter (npar=11) ! number of parameters
    real (dp):: p(npar), pboot(npar) ! vector of parameters in (0,1) and vector of parameters
    ! (phi0,mean_exp,nu,cmin,mu,alpha,deltam,deltaf,theta_inf,var_log,lovem)

    real::  ctrl(12)
    real*8:: f ! pikaia stuff

    real:: ggc(19,19)
    integer,parameter::nobs=250000  ! this is the maximum number of households
    integer :: cat_husband(19,nobs)
    integer :: nobseff,nobseff1,nobseff2,nobseff3 ! this is the effective number of households, depending on the computation that is demanded
    ! typically, pikaia => 2*1000, uobyka => 2*15625, just compute equilibrium =>2*125000
    real:: tpdraw1(nobs),tpdraw2(nobs)

    character (LEN=10)::fname,fname_out,fname_prm  ! country name
    character (LEN=2)::str

    character (len=21)::bfilename1,bfilename2,bfilename3  !filenames related to the bootstrap thing
    character (len=2)::bnum


    real :: simul_marriage_female(19),simul_marriage_male(19) ! simulated marriage rates
    real:: simul_childless_sg(19),simul_childless_ma(19)
    ! simulated childlessness rate of singles and married per wage category
    real :: simul_n_sg(19),simul_n_ma(19)      ! simlated  fertility of single and married mothers per wage category


    real :: data_sh_female_bis(19),data_marriage_female_bis(19) ! alternative global weights for single only

    real :: data_sh_female(19),data_sh_male(19),temp1(19),temp2(19),temp3(19)          ! data on shares per wage category
    real :: data_childless_sg(19),data_childless_ma(19),data_se_childless_sg(19),data_se_childless_ma(19) ! data on childlessness rate of singles and married per wage category
    real :: data_n_sg(19),data_n_ma(19),data_se_n_sg(19),data_se_n_ma(19),data_se_marriage(19)   ! data on fertility of single and married mothers per wage category
    real :: data_marriage_female(19),data_marriage_male(19),data_wages_female(19)      ! data on marriage rate
    integer:: data_childless_sg_p(19),data_childless_ma_p(19), & ! 1/0  whether the moment is precise or not
    data_n_sg_p(19),data_n_ma_p(19),data_marriage_female_p(19),data_marriage_male_p(19)
    real :: data_n_sg_tot,data_n_ma_tot,data_childless_sg_tot,data_childless_ma_tot,data_marriage_female_tot,data_marriage_male_tot
    real :: simul_n_sg_tot,simul_n_ma_tot,simul_childless_sg_tot,simul_childless_ma_tot,simul_marriage_female_tot, &
        simul_marriage_male_tot

    real :: data_mortal(19),data_unwanted(19),gamma,rhodat,lambda,temp(19),tmp ! data read from files and transferred to model
    integer :: rhotype,macho, assort,unwanted3 ! switches

    real:: benchf,bench_childless_sg_tot,bench_marriage_female_tot,bench_childless_ma_tot,simaf,simafp,simbf,simbfp,simcf,simcfp,simdf,simdfp
    real:: global_benchf,global_simdf,global_simdfp,global_simbf,global_simbfp,global_simaf,global_simafp,global_simcf,global_simcfp
    real:: dhs_benchf,dhs_simdf,dhs_simdfp,dhs_simbf,dhs_simbfp,dhs_simaf,dhs_simafp,dhs_simcf,dhs_simcfp

    real:: chi_D,chi_T,chi_V,chi_S,chi_M,chi_N  ! childlessness types

    real:: rescale ! scale factor for marriage rates male and female to be equal
    integer:: err ! to control for input/output error

    !grid parameters
    integer :: j ! steps
    real (dp):: target     ! to store the best target reached and the current target

    integer :: prt  ! whether to print the equilibrium
    ! 1 - only print output to file .out
    ! 2 - in addition print parameters to *.prm
    ! 3 - in addition print parameters to *.prm and print childlessness decomposition to agg file

    ! for random number generator
    integer,allocatable,dimension(:)::k3
    integer::m

    real::share_ctry,share_ctrym,share_ctry_bis,share_ctry_dhs  ! to compute aggregate simulated moments
    real :: global_n_sg(19), global_n_ma(19),global_marriage_female(19),global_childless_sg(19), &
        global_childless_ma(19),global_marriage_male(19),global_sh_female(19),global_sh_male(19)
    real :: globals_n_sg(19), globals_n_ma(19),globals_marriage_female(19),globals_childless_sg(19), &
        globals_childless_ma(19),globals_marriage_male(19)

    real :: pwe(19),qwe(19) ! weights to compute matching

    real::dhs_chi_T,dhs_chi_D,dhs_chi_V,dhs_chi_S,dhs_chi_M,dhs_chi_N
    real::globals_chi_T,globals_chi_D,globals_chi_V,globals_chi_S,globals_chi_M,globals_chi_N

    REAL*8:: W
    dimension W(10000)
    integer npt



    ! common blocks to pass more information to calfun()
    common /draws/ cat_husband,tpdraw1,tpdraw2

    common /observ/ data_sh_male,data_sh_female,data_n_sg,data_n_ma,data_childless_sg, &
        data_childless_ma,data_marriage_female,data_marriage_male, &
        data_n_sg_tot,data_n_ma_tot,data_childless_sg_tot,data_childless_ma_tot,data_marriage_female_tot, &
        data_marriage_male_tot,data_mortal,data_unwanted,gamma,rhodat,lambda, &
        data_childless_sg_p,data_childless_ma_p,data_n_sg_p,data_n_ma_p,data_marriage_female_p,data_marriage_male_p, &
        simul_marriage_female,simul_marriage_male,simul_childless_sg,simul_childless_ma,simul_n_sg,simul_n_ma, &
        simul_n_sg_tot,simul_n_ma_tot,simul_childless_sg_tot,simul_childless_ma_tot,simul_marriage_female_tot, &
        simul_marriage_male_tot

    common /entiers/ prt,rhotype,nobseff,macho

    common /noms/ fname

    common /chi/ chi_D,chi_T,chi_V,chi_S,chi_M,chi_N

    !initialize the random number generator
    call random_seed(size=m)
    Allocate(k3(m))
    call random_seed(get=k3(1:m)) !this gets the current integers used as a starting value


    !     Set control variables for uobyka
    maxfun=5000
    rhobeg=.1     ! this is the trust region.
    !RHOBEG should be about one tenth of the greatest expected change to a variable
    rhoend=1.0D-6  ! usual value: 1.0D-8, here we do not need that level of precision

    !     Set control variables for pikaia
    do i=1,12      !(use defaults)
        ctrl(i) = -1
    end do
    npt=2*npar+1

    ctrl(5)=2   !see PK3 p38 of userguide.ps
    ctrl(10)=1
    ctrl(11)=1

    ctrl(5)=5 ! ceci semble ameliorer les choses
    ctrl(1)=100  ! nombre d'individus
    !ctrl(2)=800 ! ceci ne sert a rien non plus
    ctrl(12)=1 !set verbose level to 1
    ctrl(2)=300 ! with 100 individuals, after 300 generations, it seems we are almost there

    rhotype=1 ! benchmark for rho (=0.05 for all countries)
    macho=0 ! marriage requires agreement of both
    assort=1 ! assortative matching switch
    unwanted3=0 ! benchamrk measure of unwantedness

    do i=1,nobs  ! draw theta
        tpdraw1(i)=random_normal()
        tpdraw2(i)=random_normal()
    enddo


1   print *, '   '
    print *,'DINKS, DEWKS & Co International'
    print *,'Thomas Baudin, David de la Croix & Paula Gobbi'
    print *, '   '
    print *,'================================================================='
    print *, 'ESTIMATION                                                        '
    print *, '1 = check the input files'
    print *, '2 = find best parameters for global - pikaia algorithm            '
    print *, '3 = find best parameters for global - newuoa algorithm            '
    print *, '4 = compute the global equilibrium with the current parameters    '
    print *, '5 = find best parameters for all countries - pikaia  algorithm    '
    print *, '6 = find best parameters for all countries - newuoa  algorithm    '
    print *, '7 = compute the equilibrium with the best parameters so far  '
    print *, '           + decomposition of childlessness     '
    print *, '8 = compute standard errors for country '
    print *, ' '
    print *, 'POLICY      '
    print *, '9 = simulate  with no mortality '
    print *, '10= simulate with full control of fertility'
    print *, '11= simulate with gender equality on the labor market'
    print *, '12= simulate with gender equality in the couple'
    print *, '13= simulate with everyone having done primary educ'
    print *, '14= draw the conclusion (=13,9,10,11)'
    print *, ' '
    print *, 'IDENTIFICATION                                                    '
    print *, '15= identification: each parameter is shifted separately            '
    print *, '     '
    print *, 'ROBUSTNESS  '
    print *, '16= switch to rho obtained from Patrinos'
    print *, '17= switch to marriage requiring man agreement only'
    print *, '18= switch off assortative matching'
    print *, '19= switch to alternative mesure of unwantedness'
    print *, ' '
    print *, '20= stop and go have a beer'

    print *,'================================================================='
    read *, choix
    print *, '   '
    print *, '   '
    print *, '   '

    open(unit=19,file="sh-educ-bis.dat")
    read (19,*) data_sh_female_bis
    close (unit=19)
    open(unit=19,file="marriage-female-educ-bis.dat")
    read (19,*) data_marriage_female_bis
    close (unit=19)

    nobseff1=1000 !number of households for pikaia, uobyka, simulation
    nobseff2=25000
    nobseff3=250000

    select case (choix)
    case (1)
        global_n_sg=0.
        global_n_ma=0.
        global_marriage_female=0.
        global_childless_sg=0.
        global_childless_ma=0.
        global_marriage_male=0.
        global_sh_female=0.
        global_sh_male=0.
        call open_data
        print *,'CHECK SPECIAL PAULAs FILE'
        print *, data_sh_female_bis
        print *,data_marriage_female_bis
        print *,' '
        print *,'CHECK THE INPUT FILES'
        do k0=1,ncountry ! iterate over countries
            call read_data
            call totals_and_draw(rescale)
            print *,fname,' ',fname_out,' ',fname_prm
            print *,data_mortal
            print *,data_unwanted
            print *,'gamma=',gamma
            print *,'rho=',rhodat
            print *,'lambda=',lambda
            print *,' weigths p=',pwe
            print *,' weigths q=',qwe
            print *,'share=',share_ctry,share_ctrym,share_ctry_bis
            if (k0>2) then  ! the first two "countries" are aggregates
                global_n_sg=global_n_sg+data_n_sg* data_sh_female*(1-data_marriage_female)*(1-data_childless_sg)*share_ctry_bis
                global_childless_sg=global_childless_sg+data_childless_sg*data_sh_female*(1-data_marriage_female)*share_ctry_bis
                global_n_ma=global_n_ma+data_n_ma* data_sh_female*data_marriage_female*(1-data_childless_ma)*share_ctry
                global_marriage_female=global_marriage_female+data_marriage_female* data_sh_female*share_ctry
                global_childless_ma=global_childless_ma+data_childless_ma*data_sh_female*data_marriage_female*share_ctry
                global_marriage_male=global_marriage_male+data_marriage_male* data_sh_male*share_ctrym/rescale
                global_sh_female=global_sh_female+data_sh_female*share_ctry
                global_sh_male=global_sh_male+data_sh_male*share_ctrym
            endif
        enddo
        call close_data
        print *,'  '
        print *,'CHECK CONSISTENCY OF THE DATA'
        call open_data ! reopening the  file to get the  moments for country 1 (global)
        call read_data
        call totals_and_draw(rescale)
        call close_data

        global_n_sg=global_n_sg/(data_sh_female_bis*(1-data_marriage_female_bis)*(1-data_childless_sg))
        global_childless_sg=global_childless_sg/(data_sh_female_bis*(1-data_marriage_female_bis))
        global_n_ma=global_n_ma/(data_sh_female*data_marriage_female*(1-data_childless_ma))
        global_marriage_female=global_marriage_female/data_sh_female
        global_childless_ma=global_childless_ma/(data_sh_female*data_marriage_female)
        global_marriage_male=global_marriage_male*rescale/data_sh_male

        print *,'Consistency of the shares: '
        print *,'females'
        do j=1,19
            print *,global_sh_female(j),data_sh_female(j)
        enddo
        print *,'males'
        do j=1,19
            print *,global_sh_male(j),data_sh_male(j)
        enddo
        print *,'Consistency of the aggregation: '
        open(unit=13,file='agg-results-check.dat',status='replace')
        write (13,*) 'j ','data_agreg ','data '
        do j=1,19
            write (13,*)j,global_n_sg(j),data_n_sg(j)
        enddo
        do j=1,19
            write (13,*)j,global_n_ma(j),data_n_ma(j)
        enddo
        do j=1,19
            write (13,*)j,global_marriage_female(j),data_marriage_female(j)
        enddo
        do j=1,19
            write (13,*)j,global_childless_sg(j),data_childless_sg(j)
        enddo
        do j=1,19
            write (13,*)j,global_childless_ma(j),data_childless_ma(j)
        enddo
        do j=1,19
            write (13,*)j,global_marriage_male(j),data_marriage_male(j)
        enddo
        close (unit=13)

    case (2)
        nobseff=nobseff1
        call open_data
        call read_data
        call totals_and_draw(rescale)
        prt=0 ! do not print the equilibrium at each computation
        seed=123.
        call rninit(seed)
        ctrl(12)=1 !set verbose level to 1
        call pikaia(func,npar,ctrl,p,f,status)
        prt=2
        open(unit=3,file="pik-"// fname_out,status='replace')
        open(unit=8,file=fname_prm,status='replace')
        call calfun(npar,p,target)
        close(unit=8)
        close(unit=3)
        call close_data

    case (3)
        nobseff=nobseff2
        call open_data
        call read_data
        call totals_and_draw(rescale)
        call read_param(err)
        if (err.ne.0) goto 1
        prt=0
        iprint=3
        call newuoa (npar, npt,p, rhobeg, rhoend, iprint, maxfun,W)
        prt=2
        open(unit=3,file="uob-"// fname_out,status='replace')
        open(unit=8,file=fname_prm,status='replace')
        call calfun(npar,p,target)
        close(unit=8)
        close(unit=3)
        call close_data


    case(4)    ! here we just compute the equilibrium for given parameters
        nobseff=nobseff3
        call open_data
        call read_data
        call totals_and_draw(rescale)
        call read_param(err)
        if (err.ne.0) goto 1
        prt=1
        open(unit=3,file="sim-"// fname_out,status='replace')
        call calfun(npar,p,target)
        close(unit=3)
        call close_data

        close(unit=19)


    case (5)
        nobseff=nobseff1
        call open_data
        do k0=1,ncountry ! iterate over countries
            call read_data
            call totals_and_draw(rescale)
            prt=0 ! do not print the equilibrium at each computation
            seed=123.
            call rninit(seed)
            call pikaia(func,npar,ctrl,p,f,status)
            prt=2
            open(unit=3,file="pik-"// fname_out,status='replace')
            open(unit=8,file=fname_prm,status='replace')
            call calfun(npar,p,target)
            close(unit=8)
            close(unit=3)
        enddo
        call close_data

    case (6)
        nobseff=nobseff2
        call open_data
        do k0=1,ncountry ! iterate over countries
            prt=0 ! do not print the equilibrium at each computation
            call read_data
            call totals_and_draw(rescale)
            call read_param(err)
            if (err.ne.0) goto 1
            prt=0
            iprint=3
            call newuoa (npar, npt,p, rhobeg, rhoend, iprint, maxfun,W)
            prt=2
            open(unit=3,file="uob-"// fname_out,status='replace')
            open(unit=8,file=fname_prm,status='replace')
            call calfun(npar,p,target)
            close(unit=8)
            close(unit=3)
        enddo
        call close_data


    case(7)    ! here we just compute the equilibrium for given parameters
        nobseff=nobseff3
        call open_data
        open(unit=9,file='estimation.dat',status='replace')
        open(unit=6,file='decomposition.dat',status='replace')
        globals_n_sg=0.
        globals_n_ma=0.
        globals_marriage_female=0.
        globals_childless_sg=0.
        globals_childless_ma=0.
        globals_marriage_male=0.
        globals_chi_T=0.
        globals_chi_D=0.
        globals_chi_V=0.
        globals_chi_S=0.
        globals_chi_M=0.
        globals_chi_N=0.
        dhs_chi_T=0.
        dhs_chi_D=0.
        dhs_chi_V=0.
        dhs_chi_S=0.
        dhs_chi_M=0.
        dhs_chi_N=0.
        do k0=1,ncountry ! iterate over countries
            call read_data
            call totals_and_draw(rescale)
            call read_param(err)
            if (err.ne.0) goto 1
            prt=3
            open(unit=3,file="sim-"// fname_out,status='replace')
            call calfun(npar,p,target)
            close(unit=3)
            write (6,'(6f7.1)') chi_T,chi_V,chi_S,chi_M,chi_N,chi_D
            if (k0>2) then
                globals_n_sg=globals_n_sg+simul_n_sg* data_sh_female*(1-data_marriage_female)*(1-data_childless_sg)*share_ctry_bis
                globals_childless_sg=globals_childless_sg+simul_childless_sg*data_sh_female*(1-data_marriage_female)*share_ctry_bis
                globals_n_ma=globals_n_ma+simul_n_ma* data_sh_female*data_marriage_female*(1-data_childless_ma)*share_ctry
                globals_marriage_female=globals_marriage_female+simul_marriage_female* data_sh_female*share_ctry
                globals_childless_ma=globals_childless_ma+simul_childless_ma*data_sh_female*data_marriage_female*share_ctry
                globals_marriage_male=globals_marriage_male+simul_marriage_male* data_sh_male*share_ctrym/rescale
                globals_chi_T=globals_chi_T+chi_T*share_ctry
                globals_chi_D=globals_chi_D+chi_D*share_ctry
                globals_chi_V=globals_chi_V+chi_V*share_ctry
                globals_chi_S=globals_chi_S+chi_S*share_ctry
                globals_chi_M=globals_chi_M+chi_M*share_ctry
                globals_chi_N=globals_chi_N+chi_N*share_ctry
                dhs_chi_T=dhs_chi_T+chi_T*share_ctry_dhs
                dhs_chi_D=dhs_chi_D+chi_D*share_ctry_dhs
                dhs_chi_V=dhs_chi_V+chi_V*share_ctry_dhs
                dhs_chi_S=dhs_chi_S+chi_S*share_ctry_dhs
                dhs_chi_M=dhs_chi_M+chi_M*share_ctry_dhs
                dhs_chi_N=dhs_chi_N+chi_N*share_ctry_dhs
            endif
        enddo
        write (6,'(6f7.1)'),dhs_chi_T,dhs_chi_V,dhs_chi_S,dhs_chi_M,dhs_chi_N,dhs_chi_D
        write (6,'(6f7.1)'),globals_chi_T,globals_chi_V,globals_chi_S,globals_chi_M,globals_chi_N,globals_chi_D
        close(unit=6)
        close(unit=9)
        call close_data

        call open_data ! reopening the  file to get the  moments for country 1 (global)
        call read_data
        call totals_and_draw(rescale)
        call read_param(err)
        if (err.ne.0) goto 1
        prt=0
        call calfun(npar,p,target)
        call close_data

        globals_marriage_female=globals_marriage_female/data_sh_female
        globals_n_sg=globals_n_sg/(data_sh_female_bis*(1-data_marriage_female_bis)*(1-data_childless_sg))
        globals_childless_sg=globals_childless_sg/(data_sh_female_bis*(1-data_marriage_female_bis))
        globals_n_ma=globals_n_ma/(data_sh_female*data_marriage_female*(1-data_childless_ma))
        globals_childless_ma=globals_childless_ma/(data_sh_female*data_marriage_female)
        globals_marriage_male=globals_marriage_male*rescale/data_sh_male

        open(unit=13,file='agg-results-sim.dat',status='replace')
        write (13,*) 'j ','fit_global ','fit_agreg ','data '
        do j=1,19
            write(13,*) j,simul_n_sg(j),globals_n_sg(j),data_n_sg(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg ','data '
        do j=1,19
            write(13,*) j,simul_n_ma(j),globals_n_ma(j),data_n_ma(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg ','data '
        do j=1,19
            write(13,*) j,simul_marriage_female(j),globals_marriage_female(j),data_marriage_female(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg ','data '
        do j=1,19
            write(13,*) j,simul_childless_sg(j),globals_childless_sg(j),data_childless_sg(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg ','data '
        do j=1,19
            write(13,*) j,simul_childless_ma(j),globals_childless_ma(j),data_childless_ma(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg ','data '
        do j=1,19
            write(13,*) j,simul_marriage_male(j),globals_marriage_male(j),data_marriage_male(j)
        enddo
        close(unit=13)

    case (8)
        nobseff=nobseff2
        call open_data
        prt=0 ! do not print the equilibrium at each computation

        print *,'Give the number of the country for which we compute SE'
        read *, choix

        if ((choix<3) .or. (choix>38)) then
            print *,'wrong country number'
            goto 1
        endif

        do j=1,choix
            read (1,*) fname
            read (2,*) fname_out
            read (7,*) data_mortal
            read (11,*) data_unwanted
            read (4,*) fname_prm
            read (12,*) gamma,rhodat
            read (5,*) lambda,pwe,qwe
            read (15,*) share_ctry
            read (16,*) share_ctrym
            read (17,*) share_ctry_bis
            read (19,*) share_ctry_dhs
        enddo

        open(unit=10,file=fname)
        do j=1 , 19
            read (10,*) data_sh_male(j), data_sh_female(j),data_n_sg_p(j),data_n_sg(j),data_n_ma_p(j),data_n_ma(j),&
                data_childless_sg_p(j),data_childless_sg(j),data_childless_ma_p(j),data_childless_ma(j),&
                data_marriage_female_p(j),data_marriage_female(j),data_marriage_male_p(j),data_marriage_male(j)
        enddo
        close (unit=10)

        choix2=choix-2
        write (bnum,'(I0)')choix2
        bfilename1='mean_bootstrap_'//trim(bnum)//'.txt'  ! we open the two files with the bootstrap data
        open(unit=9,file=bfilename1)
        bfilename2='nobs_bootstrap_'//trim(bnum)//'.txt'
        open(unit=20,file=bfilename2)
        bfilename3='parm_bootstrap_'//trim(bnum)//'.txt'
        open(unit=3,file=bfilename3)

        do j=1, 100 ! bootstrap
            !    we need to update moments
            read (9,*) data_n_sg,data_n_ma,data_childless_sg,data_childless_ma,data_marriage_female,data_marriage_male
            read (20,*) data_n_sg_p,data_n_ma_p,data_childless_sg_p,data_childless_ma_p,data_marriage_female_p,data_marriage_male_p
            call totals_and_draw(rescale)
            call read_param(err)
            if (err.ne.0) goto 1
            prt=0
            iprint=3
            call newuoa (npar, npt,p, rhobeg, rhoend, iprint, maxfun,W)
            pboot = (/ 0.07+p(1)*.13,.1+.7*p(2),4.+7.*p(3),.1+.4*p(4),.1+.4*p(5),0.5+.5*p(6),-0.05+.4*p(7),-0.05+.4*p(8),p(9),.2+p(10)/2.,p(11)/5.  /)
            write(3,'(11f9.5)') pboot
        enddo
        close(unit=9)
        close(unit=20)
        close(unit=3)
        call close_data


    case(9)    ! here we just compute the equilibrium for given parameters
        nobseff=nobseff3
        globals_n_sg=0.
        globals_n_ma=0.
        globals_marriage_female=0.
        globals_childless_sg=0.
        globals_childless_ma=0.
        globals_marriage_male=0.
        globals_chi_T=0.
        globals_chi_D=0.
        globals_chi_V=0.
        globals_chi_S=0.
        globals_chi_M=0.
        globals_chi_N=0.
        dhs_chi_T=0.
        dhs_chi_D=0.
        dhs_chi_V=0.
        dhs_chi_S=0.
        dhs_chi_M=0.
        dhs_chi_N=0.
        call open_data
        open(unit=9,file='estimation-a.dat',status='replace')
        open(unit=6,file='decomposition-a.dat',status='replace')

        do k0=1,ncountry ! iterate over countries
            call read_data
            call totals_and_draw(rescale)
            call read_param(err)
            if (err.ne.0) goto 1
            data_mortal=(/1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1./)  ! no mortality
            prt=3
            open(unit=3,file="sima-"// fname_out,status='replace')
            call calfun(npar,p,target)
            close(unit=3)
            write (6,'(6f7.1)') chi_T,chi_V,chi_S,chi_M,chi_N,chi_D
            if (k0>2) then
                globals_n_sg=globals_n_sg+simul_n_sg* data_sh_female*(1-data_marriage_female)*(1-data_childless_sg)*share_ctry_bis
                globals_childless_sg=globals_childless_sg+simul_childless_sg*data_sh_female*(1-data_marriage_female)*share_ctry_bis
                globals_n_ma=globals_n_ma+simul_n_ma* data_sh_female*data_marriage_female*(1-data_childless_ma)*share_ctry
                globals_marriage_female=globals_marriage_female+simul_marriage_female* data_sh_female*share_ctry
                globals_childless_ma=globals_childless_ma+simul_childless_ma*data_sh_female*data_marriage_female*share_ctry
                globals_marriage_male=globals_marriage_male+simul_marriage_male* data_sh_male*share_ctrym/rescale
                globals_chi_T=globals_chi_T+chi_T*share_ctry
                globals_chi_D=globals_chi_D+chi_D*share_ctry
                globals_chi_V=globals_chi_V+chi_V*share_ctry
                globals_chi_S=globals_chi_S+chi_S*share_ctry
                globals_chi_M=globals_chi_M+chi_M*share_ctry
                globals_chi_N=globals_chi_N+chi_N*share_ctry
                dhs_chi_T=dhs_chi_T+chi_T*share_ctry_dhs
                dhs_chi_D=dhs_chi_D+chi_D*share_ctry_dhs
                dhs_chi_V=dhs_chi_V+chi_V*share_ctry_dhs
                dhs_chi_S=dhs_chi_S+chi_S*share_ctry_dhs
                dhs_chi_M=dhs_chi_M+chi_M*share_ctry_dhs
                dhs_chi_N=dhs_chi_N+chi_N*share_ctry_dhs
            endif
        enddo
        write (6,'(6f7.1)'),dhs_chi_T,dhs_chi_V,dhs_chi_S,dhs_chi_M,dhs_chi_N,dhs_chi_D
        write (6,'(6f7.1)'),globals_chi_T,globals_chi_V,globals_chi_S,globals_chi_M,globals_chi_N,globals_chi_D
        close(unit=6)
        close(unit=9)
        call close_data

        call open_data ! reopening the  file to get the  moments for country 1 (global)
        call read_data
        call totals_and_draw(rescale)
        call read_param(err)
        if (err.ne.0) goto 1
        prt=0
        data_mortal=(/1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1./)  ! no mortality
        call calfun(npar,p,target)
        call close_data

        globals_marriage_female=globals_marriage_female/data_sh_female
        globals_n_sg=globals_n_sg/(data_sh_female_bis*(1-data_marriage_female_bis)*(1-data_childless_sg))
        globals_childless_sg=globals_childless_sg/(data_sh_female_bis*(1-data_marriage_female_bis))
        globals_n_ma=globals_n_ma/(data_sh_female*data_marriage_female*(1-data_childless_ma))
        globals_childless_ma=globals_childless_ma/(data_sh_female*data_marriage_female)
        globals_marriage_male=globals_marriage_male*rescale/data_sh_male

        open(unit=13,file='agg-results-sim-a.dat',status='replace')
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_n_sg(j),globals_n_sg(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_n_ma(j),globals_n_ma(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_marriage_female(j),globals_marriage_female(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_childless_sg(j),globals_childless_sg(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_childless_ma(j),globals_childless_ma(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_marriage_male(j),globals_marriage_male(j)
        enddo
        close(unit=13)


    case(10)    ! here we just compute the equilibrium for given parameters
        nobseff=nobseff3
        globals_n_sg=0.
        globals_n_ma=0.
        globals_marriage_female=0.
        globals_childless_sg=0.
        globals_childless_ma=0.
        globals_marriage_male=0.
        globals_chi_T=0.
        globals_chi_D=0.
        globals_chi_V=0.
        globals_chi_S=0.
        globals_chi_M=0.
        globals_chi_N=0.
        dhs_chi_T=0.
        dhs_chi_D=0.
        dhs_chi_V=0.
        dhs_chi_S=0.
        dhs_chi_M=0.
        dhs_chi_N=0.
        call open_data
        open(unit=9,file='estimation-b.dat',status='replace')
        open(unit=6,file='decomposition-b.dat',status='replace')

        do k0=1,ncountry ! iterate over countries
            call read_data
            call totals_and_draw(rescale)
            call read_param(err)
            if (err.ne.0) goto 1
            data_unwanted=(/0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0./)  ! full control of fertility
            prt=3
            open(unit=3,file="simb-"// fname_out,status='replace')
            call calfun(npar,p,target)
            if (k0>2) then
                globals_n_sg=globals_n_sg+simul_n_sg* data_sh_female*(1-data_marriage_female)*(1-data_childless_sg)*share_ctry_bis
                globals_childless_sg=globals_childless_sg+simul_childless_sg*data_sh_female*(1-data_marriage_female)*share_ctry_bis
                globals_n_ma=globals_n_ma+simul_n_ma* data_sh_female*data_marriage_female*(1-data_childless_ma)*share_ctry
                globals_marriage_female=globals_marriage_female+simul_marriage_female* data_sh_female*share_ctry
                globals_childless_ma=globals_childless_ma+simul_childless_ma*data_sh_female*data_marriage_female*share_ctry
                globals_marriage_male=globals_marriage_male+simul_marriage_male* data_sh_male*share_ctrym/rescale
                globals_chi_T=globals_chi_T+chi_T*share_ctry
                globals_chi_D=globals_chi_D+chi_D*share_ctry
                globals_chi_V=globals_chi_V+chi_V*share_ctry
                globals_chi_S=globals_chi_S+chi_S*share_ctry
                globals_chi_M=globals_chi_M+chi_M*share_ctry
                globals_chi_N=globals_chi_N+chi_N*share_ctry
                dhs_chi_T=dhs_chi_T+chi_T*share_ctry_dhs
                dhs_chi_D=dhs_chi_D+chi_D*share_ctry_dhs
                dhs_chi_V=dhs_chi_V+chi_V*share_ctry_dhs
                dhs_chi_S=dhs_chi_S+chi_S*share_ctry_dhs
                dhs_chi_M=dhs_chi_M+chi_M*share_ctry_dhs
                dhs_chi_N=dhs_chi_N+chi_N*share_ctry_dhs


            endif
            close(unit=3)
            write (6,'(6f7.1)') chi_T,chi_V,chi_S,chi_M,chi_N,chi_D
        enddo
        write (6,'(6f7.1)'),dhs_chi_T,dhs_chi_V,dhs_chi_S,dhs_chi_M,dhs_chi_N,dhs_chi_D
        write (6,'(6f7.1)'),globals_chi_T,globals_chi_V,globals_chi_S,globals_chi_M,globals_chi_N,globals_chi_D
        close(unit=6)
        close(unit=9)

        call close_data
        call close_data


        call open_data ! reopening the  file to get the  moments for country 1 (global)
        call read_data
        call totals_and_draw(rescale)
        call read_param(err)
        if (err.ne.0) goto 1
        prt=0
        data_unwanted=(/0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0./)  ! full control of fertility
        call calfun(npar,p,target)
        call close_data

        globals_marriage_female=globals_marriage_female/data_sh_female
        globals_n_sg=globals_n_sg/(data_sh_female_bis*(1-data_marriage_female_bis)*(1-data_childless_sg))
        globals_childless_sg=globals_childless_sg/(data_sh_female_bis*(1-data_marriage_female_bis))
        globals_n_ma=globals_n_ma/(data_sh_female*data_marriage_female*(1-data_childless_ma))
        globals_childless_ma=globals_childless_ma/(data_sh_female*data_marriage_female)
        globals_marriage_male=globals_marriage_male*rescale/data_sh_male

        open(unit=13,file='agg-results-sim-b.dat',status='replace')
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_n_sg(j),globals_n_sg(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_n_ma(j),globals_n_ma(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_marriage_female(j),globals_marriage_female(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_childless_sg(j),globals_childless_sg(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_childless_ma(j),globals_childless_ma(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_marriage_male(j),globals_marriage_male(j)
        enddo
        close(unit=13)

    case(11)    ! here we just compute the equilibrium for given parameters
        nobseff=nobseff3
        globals_n_sg=0.
        globals_n_ma=0.
        globals_marriage_female=0.
        globals_childless_sg=0.
        globals_childless_ma=0.
        globals_marriage_male=0.
        globals_chi_T=0.
        globals_chi_D=0.
        globals_chi_V=0.
        globals_chi_S=0.
        globals_chi_M=0.
        globals_chi_N=0.
        dhs_chi_T=0.
        dhs_chi_D=0.
        dhs_chi_V=0.
        dhs_chi_S=0.
        dhs_chi_M=0.
        dhs_chi_N=0.
        call open_data
        open(unit=9,file='estimation-c.dat',status='replace')
        open(unit=6,file='decomposition-c.dat',status='replace')

        do k0=1,ncountry ! iterate over countries
            call read_data
            call totals_and_draw(rescale)
            call read_param(err)
            if (err.ne.0) goto 1
            gamma=1.  ! gender equality
            prt=3
            open(unit=3,file="simc-"// fname_out,status='replace')
            call calfun(npar,p,target)
            if (k0>2) then

                globals_n_sg=globals_n_sg+simul_n_sg* data_sh_female*(1-data_marriage_female)*(1-data_childless_sg)*share_ctry_bis
                globals_childless_sg=globals_childless_sg+simul_childless_sg*data_sh_female*(1-data_marriage_female)*share_ctry_bis
                globals_n_ma=globals_n_ma+simul_n_ma* data_sh_female*data_marriage_female*(1-data_childless_ma)*share_ctry
                globals_marriage_female=globals_marriage_female+simul_marriage_female* data_sh_female*share_ctry
                globals_childless_ma=globals_childless_ma+simul_childless_ma*data_sh_female*data_marriage_female*share_ctry
                globals_marriage_male=globals_marriage_male+simul_marriage_male* data_sh_male*share_ctrym/rescale
                globals_chi_T=globals_chi_T+chi_T*share_ctry
                globals_chi_D=globals_chi_D+chi_D*share_ctry
                globals_chi_V=globals_chi_V+chi_V*share_ctry
                globals_chi_S=globals_chi_S+chi_S*share_ctry
                globals_chi_M=globals_chi_M+chi_M*share_ctry
                globals_chi_N=globals_chi_N+chi_N*share_ctry
                dhs_chi_T=dhs_chi_T+chi_T*share_ctry_dhs
                dhs_chi_D=dhs_chi_D+chi_D*share_ctry_dhs
                dhs_chi_V=dhs_chi_V+chi_V*share_ctry_dhs
                dhs_chi_S=dhs_chi_S+chi_S*share_ctry_dhs
                dhs_chi_M=dhs_chi_M+chi_M*share_ctry_dhs
                dhs_chi_N=dhs_chi_N+chi_N*share_ctry_dhs


            endif
            close(unit=3)
            write (6,'(6f7.1)') chi_T,chi_V,chi_S,chi_M,chi_N,chi_D
        enddo
        write (6,'(6f7.1)'),dhs_chi_T,dhs_chi_V,dhs_chi_S,dhs_chi_M,dhs_chi_N,dhs_chi_D
        write (6,'(6f7.1)'),globals_chi_T,globals_chi_V,globals_chi_S,globals_chi_M,globals_chi_N,globals_chi_D
        close(unit=6)
        close(unit=9)

        call close_data


        call open_data ! reopening the  file to get the  moments for country 1 (global)
        call read_data
        call totals_and_draw(rescale)
        call read_param(err)
        if (err.ne.0) goto 1
        prt=0
        gamma=1.  ! gender equality
        call calfun(npar,p,target)
        call close_data


        globals_marriage_female=globals_marriage_female/data_sh_female
        globals_n_sg=globals_n_sg/(data_sh_female_bis*(1-data_marriage_female_bis)*(1-data_childless_sg))
        globals_childless_sg=globals_childless_sg/(data_sh_female_bis*(1-data_marriage_female_bis))
        globals_n_ma=globals_n_ma/(data_sh_female*data_marriage_female*(1-data_childless_ma))
        globals_childless_ma=globals_childless_ma/(data_sh_female*data_marriage_female)
        globals_marriage_male=globals_marriage_male*rescale/data_sh_male

        open(unit=13,file='agg-results-sim-c.dat',status='replace')
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_n_sg(j),globals_n_sg(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_n_ma(j),globals_n_ma(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_marriage_female(j),globals_marriage_female(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_childless_sg(j),globals_childless_sg(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_childless_ma(j),globals_childless_ma(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_marriage_male(j),globals_marriage_male(j)
        enddo
        close(unit=13)


    case(12)    ! here we just compute the equilibrium for given parameters
        nobseff=nobseff3
        globals_n_sg=0.
        globals_n_ma=0.
        globals_marriage_female=0.
        globals_childless_sg=0.
        globals_childless_ma=0.
        globals_marriage_male=0.
        globals_chi_T=0.
        globals_chi_D=0.
        globals_chi_V=0.
        globals_chi_S=0.
        globals_chi_M=0.
        globals_chi_N=0.
        dhs_chi_T=0.
        dhs_chi_D=0.
        dhs_chi_V=0.
        dhs_chi_S=0.
        dhs_chi_M=0.
        dhs_chi_N=0.
        call open_data
        open(unit=9,file='estimation-d.dat',status='replace')
        open(unit=6,file='decomposition-d.dat',status='replace')

        do k0=1,ncountry ! iterate over countries
            call read_data
            call totals_and_draw(rescale)
            call read_param(err)
            if (err.ne.0) goto 1
            p(6)=0.  ! gender equality (alpha=0)
            prt=3
            open(unit=3,file="simd-"// fname_out,status='replace')
            call calfun(npar,p,target)
            if (k0>2) then

                globals_n_sg=globals_n_sg+simul_n_sg* data_sh_female*(1-data_marriage_female)*(1-data_childless_sg)*share_ctry_bis
                globals_childless_sg=globals_childless_sg+simul_childless_sg*data_sh_female*(1-data_marriage_female)*share_ctry_bis
                globals_n_ma=globals_n_ma+simul_n_ma* data_sh_female*data_marriage_female*(1-data_childless_ma)*share_ctry
                globals_marriage_female=globals_marriage_female+simul_marriage_female* data_sh_female*share_ctry
                globals_childless_ma=globals_childless_ma+simul_childless_ma*data_sh_female*data_marriage_female*share_ctry
                globals_marriage_male=globals_marriage_male+simul_marriage_male* data_sh_male*share_ctrym/rescale
                globals_chi_T=globals_chi_T+chi_T*share_ctry
                globals_chi_D=globals_chi_D+chi_D*share_ctry
                globals_chi_V=globals_chi_V+chi_V*share_ctry
                globals_chi_S=globals_chi_S+chi_S*share_ctry
                globals_chi_M=globals_chi_M+chi_M*share_ctry
                globals_chi_N=globals_chi_N+chi_N*share_ctry
                dhs_chi_T=dhs_chi_T+chi_T*share_ctry_dhs
                dhs_chi_D=dhs_chi_D+chi_D*share_ctry_dhs
                dhs_chi_V=dhs_chi_V+chi_V*share_ctry_dhs
                dhs_chi_S=dhs_chi_S+chi_S*share_ctry_dhs
                dhs_chi_M=dhs_chi_M+chi_M*share_ctry_dhs
                dhs_chi_N=dhs_chi_N+chi_N*share_ctry_dhs


            endif
            close(unit=3)
            write (6,'(6f7.1)') chi_T,chi_V,chi_S,chi_M,chi_N,chi_D
        enddo
        write (6,'(6f7.1)'),dhs_chi_T,dhs_chi_V,dhs_chi_S,dhs_chi_M,dhs_chi_N,dhs_chi_D
        write (6,'(6f7.1)'),globals_chi_T,globals_chi_V,globals_chi_S,globals_chi_M,globals_chi_N,globals_chi_D
        close(unit=6)
        close(unit=9)

        call close_data

        call open_data ! reopening the  file to get the  moments for country 1 (global)
        call read_data
        call totals_and_draw(rescale)
        call read_param(err)
        if (err.ne.0) goto 1
        prt=0
        p(6)=0.  ! gender equality
        call calfun(npar,p,target)
        call close_data


        globals_marriage_female=globals_marriage_female/data_sh_female
        globals_n_sg=globals_n_sg/(data_sh_female_bis*(1-data_marriage_female_bis)*(1-data_childless_sg))
        globals_childless_sg=globals_childless_sg/(data_sh_female_bis*(1-data_marriage_female_bis))
        globals_n_ma=globals_n_ma/(data_sh_female*data_marriage_female*(1-data_childless_ma))
        globals_childless_ma=globals_childless_ma/(data_sh_female*data_marriage_female)
        globals_marriage_male=globals_marriage_male*rescale/data_sh_male

        open(unit=13,file='agg-results-sim-d.dat',status='replace')
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_n_sg(j),globals_n_sg(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_n_ma(j),globals_n_ma(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_marriage_female(j),globals_marriage_female(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_childless_sg(j),globals_childless_sg(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_childless_ma(j),globals_childless_ma(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_marriage_male(j),globals_marriage_male(j)
        enddo
        close(unit=13)



    case(13)    ! everyone goes to primary school
        nobseff=nobseff3
        globals_n_sg=0.
        globals_n_ma=0.
        globals_marriage_female=0.
        globals_childless_sg=0.
        globals_childless_ma=0.
        globals_marriage_male=0.
        globals_chi_T=0.
        globals_chi_D=0.
        globals_chi_V=0.
        globals_chi_S=0.
        globals_chi_M=0.
        globals_chi_N=0.
        dhs_chi_T=0.
        dhs_chi_D=0.
        dhs_chi_V=0.
        dhs_chi_S=0.
        dhs_chi_M=0.
        dhs_chi_N=0.
        call open_data
        open(unit=9,file='estimation-e.dat',status='replace')
        open(unit=6,file='decomposition-e.dat',status='replace')

        do k0=1,ncountry ! iterate over countries
            call read_data
            call read_param(err)
            if (err.ne.0) goto 1

            data_sh_male(7)=data_sh_male(1)+data_sh_male(2)+data_sh_male(3)+data_sh_male(4)+data_sh_male(5)+data_sh_male(6)+data_sh_male(7)
            data_sh_female(7)=data_sh_female(1)+data_sh_female(2)+data_sh_female(3)+data_sh_female(4) &
                +data_sh_female(5)+data_sh_female(6)+data_sh_female(7)
            data_sh_female_bis(7)=data_sh_female_bis(1)+data_sh_female_bis(2)+data_sh_female_bis(3)+data_sh_female_bis(4) &
                +data_sh_female_bis(5)+data_sh_female_bis(6)+data_sh_female_bis(7)
            do j=1,6
                data_sh_male(j)=0
                data_sh_female(j)=0
                data_sh_female_bis(j)=0
            enddo
            call totals_and_draw(rescale)
            prt=3
            open(unit=3,file="sime-"// fname_out,status='replace')
            call calfun(npar,p,target)
            if (k0>2) then

                globals_n_sg=globals_n_sg+simul_n_sg* data_sh_female*(1-data_marriage_female)*(1-data_childless_sg)*share_ctry_bis
                globals_childless_sg=globals_childless_sg+simul_childless_sg*data_sh_female*(1-data_marriage_female)*share_ctry_bis
                globals_n_ma=globals_n_ma+simul_n_ma* data_sh_female*data_marriage_female*(1-data_childless_ma)*share_ctry
                globals_marriage_female=globals_marriage_female+simul_marriage_female* data_sh_female*share_ctry
                globals_childless_ma=globals_childless_ma+simul_childless_ma*data_sh_female*data_marriage_female*share_ctry
                globals_marriage_male=globals_marriage_male+simul_marriage_male* data_sh_male*share_ctrym/rescale
                globals_chi_T=globals_chi_T+chi_T*share_ctry
                globals_chi_D=globals_chi_D+chi_D*share_ctry
                globals_chi_V=globals_chi_V+chi_V*share_ctry
                globals_chi_S=globals_chi_S+chi_S*share_ctry
                globals_chi_M=globals_chi_M+chi_M*share_ctry
                globals_chi_N=globals_chi_N+chi_N*share_ctry
                dhs_chi_T=dhs_chi_T+chi_T*share_ctry_dhs
                dhs_chi_D=dhs_chi_D+chi_D*share_ctry_dhs
                dhs_chi_V=dhs_chi_V+chi_V*share_ctry_dhs
                dhs_chi_S=dhs_chi_S+chi_S*share_ctry_dhs
                dhs_chi_M=dhs_chi_M+chi_M*share_ctry_dhs
                dhs_chi_N=dhs_chi_N+chi_N*share_ctry_dhs


            endif
            close(unit=3)
            write (6,'(6f7.1)') chi_T,chi_V,chi_S,chi_M,chi_N,chi_D
        enddo
        write (6,'(6f7.1)'),dhs_chi_T,dhs_chi_V,dhs_chi_S,dhs_chi_M,dhs_chi_N,dhs_chi_D
        write (6,'(6f7.1)'),globals_chi_T,globals_chi_V,globals_chi_S,globals_chi_M,globals_chi_N,globals_chi_D
        close(unit=6)
        close(unit=9)

        call close_data


        call open_data ! reopening the  file to get the  moments for country 1 (global)
        call read_data
        call read_param(err)
        if (err.ne.0) goto 1
        prt=0
        data_sh_male(7)=data_sh_male(1)+data_sh_male(2)+data_sh_male(3)+data_sh_male(4)+data_sh_male(5)+data_sh_male(6)+data_sh_male(7)
        data_sh_female(7)=data_sh_female(1)+data_sh_female(2)+data_sh_female(3)+data_sh_female(4) &
            +data_sh_female(5)+data_sh_female(6)+data_sh_female(7)
        data_sh_female_bis(7)=data_sh_female_bis(1)+data_sh_female_bis(2)+data_sh_female_bis(3)+data_sh_female_bis(4) &
            +data_sh_female_bis(5)+data_sh_female_bis(6)+data_sh_female_bis(7)
        do j=1,6
            data_sh_male(j)=0
            data_sh_female(j)=0
            data_sh_female_bis(j)=0
        enddo
        call totals_and_draw(rescale)
        call calfun(npar,p,target)
        call close_data

        globals_marriage_female=globals_marriage_female/data_sh_female
        globals_n_sg=globals_n_sg/(data_sh_female_bis*(1-data_marriage_female_bis)*(1-data_childless_sg))
        globals_childless_sg=globals_childless_sg/(data_sh_female_bis*(1-data_marriage_female_bis))
        globals_n_ma=globals_n_ma/(data_sh_female*data_marriage_female*(1-data_childless_ma))
        globals_childless_ma=globals_childless_ma/(data_sh_female*data_marriage_female)
        globals_marriage_male=globals_marriage_male*rescale/data_sh_male

        open(unit=13,file='agg-results-sim-e.dat',status='replace')
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_n_sg(j),globals_n_sg(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_n_ma(j),globals_n_ma(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_marriage_female(j),globals_marriage_female(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_childless_sg(j),globals_childless_sg(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_childless_ma(j),globals_childless_ma(j)
        enddo
        write (13,*) 'j ','fit_global ','fit_agreg '
        do j=1,19
            write(13,*) j,simul_marriage_male(j),globals_marriage_male(j)
        enddo
        close(unit=13)

    case(14)

        global_benchf=0
        global_simdf=0
        global_simdfp=0
        global_simbf=0
        global_simbfp=0
        global_simaf=0
        global_simafp=0
        global_simcf=0
        global_simcfp=0

        dhs_benchf=0
        dhs_simdf=0
        dhs_simdfp=0
        dhs_simbf=0
        dhs_simbfp=0
        dhs_simaf=0
        dhs_simafp=0
        dhs_simcf=0
        dhs_simcfp=0

        nobseff=nobseff3
        prt=0
        open(unit=13,file='conclusion.dat',status='replace')
        call open_data

        do k0=1,ncountry ! iterate over countries
            call read_data
            call totals_and_draw(rescale)
            call read_param(err)
            call calfun(npar,p,target)
            benchf=simul_n_sg_tot*(1-simul_marriage_female_tot)*(1-simul_childless_sg_tot)+ &
                simul_n_ma_tot* simul_marriage_female_tot*(1-simul_childless_ma_tot)
            bench_childless_sg_tot= simul_childless_sg_tot
            bench_marriage_female_tot= simul_marriage_female_tot
            bench_childless_ma_tot=simul_childless_ma_tot

            temp=data_mortal
            data_mortal=(/1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1./)  ! no mortality
            call calfun(npar,p,target)
            data_mortal=temp

            simaf=simul_n_sg_tot*(1-simul_marriage_female_tot)*(1-simul_childless_sg_tot)+ &
                simul_n_ma_tot* simul_marriage_female_tot*(1-simul_childless_ma_tot)
            simafp=simul_n_sg_tot*(1-bench_marriage_female_tot)*(1-bench_childless_sg_tot)+ &
                simul_n_ma_tot* bench_marriage_female_tot*(1-bench_childless_ma_tot)

            temp=data_unwanted
            data_unwanted=(/0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0./)  ! full control of fertility
            call calfun(npar,p,target)
            data_unwanted=temp

            simbf=simul_n_sg_tot*(1-simul_marriage_female_tot)*(1-simul_childless_sg_tot)+ &
                simul_n_ma_tot* simul_marriage_female_tot*(1-simul_childless_ma_tot)
            simbfp=simul_n_sg_tot*(1-bench_marriage_female_tot)*(1-bench_childless_sg_tot)+ &
                simul_n_ma_tot* bench_marriage_female_tot*(1-bench_childless_ma_tot)


            tmp=gamma
            gamma=1.  ! gender equality
            call calfun(npar,p,target)
            gamma=tmp
            simcf=simul_n_sg_tot*(1-simul_marriage_female_tot)*(1-simul_childless_sg_tot)+ &
                simul_n_ma_tot* simul_marriage_female_tot*(1-simul_childless_ma_tot)
            simcfp=simul_n_sg_tot*(1-bench_marriage_female_tot)*(1-bench_childless_sg_tot)+ &
                simul_n_ma_tot* bench_marriage_female_tot*(1-bench_childless_ma_tot)

            temp1=data_sh_male
            temp2=data_sh_female
            temp3=data_sh_female_bis
            data_sh_male(7)=data_sh_male(1)+data_sh_male(2)+data_sh_male(3)+data_sh_male(4)+data_sh_male(5)+data_sh_male(6)+data_sh_male(7)
            data_sh_female(7)=data_sh_female(1)+data_sh_female(2)+data_sh_female(3)+data_sh_female(4) &
                +data_sh_female(5)+data_sh_female(6)+data_sh_female(7)
            data_sh_female_bis(7)=data_sh_female_bis(1)+data_sh_female_bis(2)+data_sh_female_bis(3)+data_sh_female_bis(4) &
                +data_sh_female_bis(5)+data_sh_female_bis(6)+data_sh_female_bis(7)

            do j=1,6
                data_sh_male(j)=0
                data_sh_female(j)=0
                data_sh_female_bis(j)=0
            enddo
            call totals_and_draw(rescale)
            call calfun(npar,p,target)
            data_sh_male=temp1
            data_sh_female=temp2
            data_sh_female_bis=temp3
            call totals_and_draw(rescale)

            simdf=simul_n_sg_tot*(1-simul_marriage_female_tot)*(1-simul_childless_sg_tot)+ &
                simul_n_ma_tot* simul_marriage_female_tot*(1-simul_childless_ma_tot)
            simdfp=simul_n_sg_tot*(1-bench_marriage_female_tot)*(1-bench_childless_sg_tot)+ &
                simul_n_ma_tot* bench_marriage_female_tot*(1-bench_childless_ma_tot)

            ! check for NaN in results (we cheat a bit and set to zero the NaN, which arise probably because no agents is in this category)
            if (simdf .ne. simdf) then  
                simdf=benchf
            endif
            if (simdfp .ne. simdfp) then  
                simdfp=benchf
            endif
            if (simbf .ne. simbf) then  
                simbf=benchf
            endif
            if (simbfp .ne. simbfp) then  
                simbfp=benchf
            endif
            if (simaf .ne. simaf) then  
                simaf=benchf
            endif
            if (simafp .ne. simafp) then  
                simafp=benchf
            endif
            if (simcf .ne. simcf) then  
                simcf=benchf
            endif
            if (simcfp .ne. simcfp) then  
                simcfp=benchf
            endif
            !note:
            !To determine if a variable is a NaN you can make use of the property
            !that a NaN does not compare equal to anything, including itself.

            if (k0>2) then
                global_benchf=global_benchf+benchf*share_ctry
                global_simdf=global_simdf+simdf*share_ctry
                global_simdfp=global_simdfp+simdfp*share_ctry
                global_simbf=global_simbf+simbf*share_ctry
                global_simbfp=global_simbfp+simbfp*share_ctry
                global_simaf=global_simaf+simaf*share_ctry
                global_simafp=global_simafp+simafp*share_ctry
                global_simcf=global_simcf+simcf*share_ctry
                global_simcfp=global_simcfp+simcfp*share_ctry

                dhs_benchf=dhs_benchf+benchf*share_ctry_dhs
                dhs_simdf=dhs_simdf+simdf*share_ctry_dhs
                dhs_simdfp=dhs_simdfp+simdfp*share_ctry_dhs
                dhs_simbf=dhs_simbf+simbf*share_ctry_dhs
                dhs_simbfp=dhs_simbfp+simbfp*share_ctry_dhs
                dhs_simaf=dhs_simaf+simaf*share_ctry_dhs
                dhs_simafp=dhs_simafp+simafp*share_ctry_dhs
                dhs_simcf=dhs_simcf+simcf*share_ctry_dhs
                dhs_simcfp=dhs_simcfp+simcfp*share_ctry_dhs
            endif


            write(13,'(9f13.3)') benchf, 100*(simdf-benchf)/benchf,100*(simdfp-benchf)/benchf,&
                100*(simbf-benchf)/benchf,100*(simbfp-benchf)/benchf,100*(simaf-benchf)/benchf &
                ,100*(simafp-benchf)/benchf,100*(simcf-benchf)/benchf,100*(simcfp-benchf)/benchf
        enddo

        call close_data

        write(13,'(9f13.3)') dhs_benchf, 100*(dhs_simdf-dhs_benchf)/dhs_benchf,100*(dhs_simdfp-dhs_benchf)/dhs_benchf,&
            100*(dhs_simbf-dhs_benchf)/dhs_benchf,100*(dhs_simbfp-dhs_benchf)/dhs_benchf,100*(dhs_simaf-dhs_benchf)/dhs_benchf &
            ,100*(dhs_simafp-dhs_benchf)/dhs_benchf,100*(dhs_simcf-dhs_benchf)/dhs_benchf,100*(dhs_simcfp-dhs_benchf)/dhs_benchf
        write(13,'(9f13.3)') global_benchf, 100*(global_simdf-global_benchf)/global_benchf,100*(global_simdfp-global_benchf)/global_benchf,&
            100*(global_simbf-global_benchf)/global_benchf,100*(global_simbfp-global_benchf)/global_benchf, &
            100*(global_simaf-global_benchf)/global_benchf &
            ,100*(global_simafp-global_benchf)/global_benchf,100*(global_simcf-global_benchf)/global_benchf,100*(global_simcfp- &
            global_benchf)/global_benchf
        call close_data
        close(13)


    case(15)    ! here we just compute the equilibrium for given parameters
        nobseff=nobseff3
        prt=0
        do k4=0,npar ! loop on number of parameters + 1
            globals_n_sg=0.
            globals_n_ma=0.
            globals_marriage_female=0.
            globals_childless_sg=0.
            globals_childless_ma=0.
            globals_marriage_male=0.
            call open_data
            do k0=1,ncountry ! iterate over countries
                call read_data
                call read_param(err)
                if (err.ne.0) goto 1
                call totals_and_draw(rescale)

                select case (k4) ! comparative statics with respect to parameters
                case (1)
                    p(1)=p(1)+.1
                case (2)
                    p(2)=p(2)+.2
                case (3)
                    p(3)=p(3)+.1
                case (4)
                    p(4)=p(4)+.2
                case (5)
                    p(5)=p(5)+.2
                case (6)
                    p(6)=1.
                case (7)
                    p(7)=p(7)+.3
                case (8)
                    p(8)=p(8)+.3
                case (9)
                    p(9)=1.
                case (10)
                    p(10)=p(10)+.4
                case (11)
                    p(11)=p(11)+.4
                end select
                call calfun(npar,p,target)
                if (k0>2) then
                    globals_n_sg=globals_n_sg+simul_n_sg* data_sh_female*(1-data_marriage_female)*(1-data_childless_sg)*share_ctry_bis
                    globals_childless_sg=globals_childless_sg+simul_childless_sg*data_sh_female*(1-data_marriage_female)*share_ctry_bis
                    globals_n_ma=globals_n_ma+simul_n_ma* data_sh_female*data_marriage_female*(1-data_childless_ma)*share_ctry
                    globals_marriage_female=globals_marriage_female+simul_marriage_female* data_sh_female*share_ctry
                    globals_childless_ma=globals_childless_ma+simul_childless_ma*data_sh_female*data_marriage_female*share_ctry
                    globals_marriage_male=globals_marriage_male+simul_marriage_male* data_sh_male*share_ctrym/rescale

                endif
                print *,"Dealing with parameter ",k4,", country ",k0
            enddo
            call close_data
            call open_data ! reopening the  file to get the  moments for country 1 (global)
            call read_data
            call totals_and_draw(rescale)
            call close_data
            globals_marriage_female=globals_marriage_female/data_sh_female
            globals_n_sg=globals_n_sg/(data_sh_female_bis*(1-data_marriage_female_bis)*(1-data_childless_sg))
            globals_childless_sg=globals_childless_sg/(data_sh_female_bis*(1-data_marriage_female_bis))
            globals_n_ma=globals_n_ma/(data_sh_female*data_marriage_female*(1-data_childless_ma))
            globals_childless_ma=globals_childless_ma/(data_sh_female*data_marriage_female)
            globals_marriage_male=globals_marriage_male*rescale/data_sh_male
            write( str, '(i2)' ) k4 ! writes the integer k4 into the string str, to be able to concatenate in the file name
            open(unit=13,file="ident" // str // ".dat",status='replace')
            write (13,*) 'j ','n_sg '
            do j=1,19
                write(13,*) j,globals_n_sg(j)
            enddo
            write (13,*) 'j ','n_ma '
            do j=1,19
                write(13,*) j,globals_n_ma(j)
            enddo
            write (13,*) 'j ','m_female '
            do j=1,19
                write(13,*) j,globals_marriage_female(j)
            enddo
            write (13,*) 'j ','chi_sg '
            do j=1,19
                write(13,*) j,globals_childless_sg(j)
            enddo
            write (13,*) 'j ','chi_ma '
            do j=1,19
                write(13,*) j,globals_childless_ma(j)
            enddo
            write (13,*) 'j ','m_male '
            do j=1,19
                write(13,*) j,globals_marriage_male(j)
            enddo
            close(unit=13)
        enddo

    case(16)
        rhotype=2
        print *,'done - rho set to value from Patrinos'

    case(17)
        macho=1
        print *,'done - women agreement does not count anymore '

    case(18)
        assort=0
        print *,'done - no assortative matching any more '

    case(19)
        unwanted3=1
        print *,'done - alternative measure of unwantedness '

    case(20)
        stop

    end select

    goto 1

    contains

    !======================================================================================================

    subroutine open_data

    open(unit=1,file='countries-name.dat') ! file with filenames
    open(unit=2,file='countries-out.dat') ! file with filenames  for output
    open(unit=4,file='countries-param.dat') ! file with filenames  for parameter output
    open(unit=5,file='assort.dat') ! data to compute matching
    open(unit=7,file='mortal.dat')
    if (unwanted3==0) then
        open(unit=11,file='unwanted2.dat')
    else
        open(unit=11,file='unwanted3.dat')
    endif
    open(unit=12,file='gamma.dat')
    open(unit=15,file='sh-cntry.dat')
    open(unit=16,file='sh-cntrym.dat')
    open(unit=17,file='sh-cntry-bis.dat')
    open(unit=19,file='sh-cntry-dhs.dat')

    end subroutine open_data

    !======================================================================================================

    subroutine close_data

    close(unit=1)
    close(unit=2)
    close(unit=4)
    close(unit=5)
    close(unit=7)
    close(unit=11)
    close(unit=12)
    close(unit=15)
    close(unit=16)
    close(unit=17)
    close(unit=19)

    end subroutine close_data

    !======================================================================================================

    subroutine read_data


    read (1,*) fname
    read (2,*) fname_out
    read (7,*) data_mortal
    read (11,*) data_unwanted
    read (4,*) fname_prm
    read (12,*) gamma,rhodat
    read (5,*) lambda,pwe,qwe
    open(unit=10,file=fname)
    do j=1 , 19
        read (10,*) data_sh_male(j), data_sh_female(j),data_n_sg_p(j),data_n_sg(j),data_n_ma_p(j),data_n_ma(j),&
            data_childless_sg_p(j),data_childless_sg(j),data_childless_ma_p(j),data_childless_ma(j),&
            data_marriage_female_p(j),data_marriage_female(j),data_marriage_male_p(j),data_marriage_male(j)
    enddo
    close (unit=10)
    read (15,*) share_ctry
    read (16,*) share_ctrym
    read (17,*) share_ctry_bis
    read (19,*) share_ctry_dhs

    end subroutine read_data


    subroutine totals_and_draw(rescale)

    real,intent(out)::rescale
    integer:: i,j,k,nsq,v1,v2,dnsq
    integer :: categories(19)
    real :: gg(19,19)  ! probabilities to meet someone of education
    real:: v2v

    ! compute totals
    data_n_sg_tot=0
    data_n_ma_tot=0
    data_childless_sg_tot=0
    data_childless_ma_tot=0
    data_marriage_female_tot=0
    data_marriage_male_tot=0

    do j=1,19
        data_marriage_female_tot=data_marriage_female_tot+data_marriage_female(j)*data_sh_female(j)
        data_marriage_male_tot=data_marriage_male_tot+data_marriage_male(j)*data_sh_male(j)
        data_childless_sg_tot=data_childless_sg_tot+data_childless_sg(j)*data_sh_female(j)*(1-data_marriage_female(j))
        data_childless_ma_tot=data_childless_ma_tot+data_childless_ma(j)*data_sh_female(j)*data_marriage_female(j)
        data_n_sg_tot=data_n_sg_tot+data_n_sg(j)*data_sh_female(j)*(1-data_childless_sg(j))*(1-data_marriage_female(j))
        data_n_ma_tot=data_n_ma_tot+data_n_ma(j)*data_sh_female(j)*(1-data_childless_ma(j))*data_marriage_female(j)
    end do

    data_childless_sg_tot=data_childless_sg_tot/(1-data_marriage_female_tot)
    data_childless_ma_tot=data_childless_ma_tot/data_marriage_female_tot

    data_n_sg_tot=data_n_sg_tot/(1-data_childless_sg_tot)/(1-data_marriage_female_tot)
    data_n_ma_tot=data_n_ma_tot/(1-data_childless_ma_tot)/data_marriage_female_tot

    if (sum(data_childless_sg_p)==0.) then
        data_childless_sg_tot=0.
        data_n_sg_tot=0.
    endif

    if (data_marriage_female_tot<.999) then
        rescale=data_marriage_female_tot/data_marriage_male_tot    ! rescale marriage rates of males
    else
        rescale=1        ! if we have no single women in data, we consider marriage rates of males as right
    endif

    data_marriage_male_tot=0
    do j=1,19
        data_marriage_male(j)=data_marriage_male(j)*rescale
        data_marriage_male_tot=data_marriage_male_tot+data_marriage_male(j)*data_sh_male(j)
    enddo

    ! draw households
    categories=(/ 1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19 /)

    ggc=0.

    if (assort==0) then
        lambda=0.
        pwe=1.
        qwe=1.
    endif

    do j=1,19 ! type of woman

        do k=1,19
            gg(j,k)=exp(-lambda*abs(j-k+2))*exp(-lambda*abs(j-k-2))*data_sh_male(k)*pwe(j)*qwe(k) ! probability for a woman of type j to draw a man of type k
        enddo
        do k=1,19 ! cumulative probabilities for a woman of type j to draw a man of type k or less

            do i=1,k
                ggc(j,k)=ggc(j,k)+gg(j,i)
            enddo
        enddo
    enddo


    ! alternative
    do j=1,19
        do i=1,nobs  ! women #i
            cat_husband(j,i)=RandomChoicePrg2(19,gg(j,:),categories ) ! we draw the category of the husband
        end do
    end do




    ! open(unit=19,file="ggc.txt",status='replace')
    ! do j=1,19
    !     write(19,'(19f7.3)') ggc(j,:)
    ! enddo
    ! close(unit=19)

    !  open(unit=19,file="gg.txt",status='replace')
    !  do j=1,19
    !      write(19,'(19f7.3)') gg(j,:)
    !  enddo
    !  close(unit=19)

    end subroutine totals_and_draw

    !======================================================================================================

    real function random_normal()     ! draws from a normal (0,1) distribution
    ! Adapted from the following Fortran 77 code
    !      ALGORITHM 712, COLLECTED ALGORITHMS FROM ACM.
    !      THIS WORK PUBLISHED IN TRANSACTIONS ON MATHEMATICAL SOFTWARE,
    !      VOL. 18, NO. 4, DECEMBER, 1992, PP. 434-435.

    !  The function random_normal() returns a normally distributed pseudo-random
    !  number with zero mean and unit variance.

    !  The algorithm uses the ratio of uniforms method of A.J. Kinderman
    !  and J.F. Monahan augmented with quadratic bounding curves.


    !     Local variables
    real     :: s = 0.449871, t = -0.386595, a = 0.19600, b = 0.25472,           &
        r1 = 0.27597, r2 = 0.27846, u, v, x, y, q

    !     Generate P = (u,v) uniform in rectangle enclosing acceptance region

    do
        call random_number(u)
        call random_number(v)
        v = 1.7156 * (v - 0.5)

        !     Evaluate the quadratic form
        x = u - s
        y = ABS(v) - t
        q = x**2 + y*(a*y - b*x)

        !     Accept P if inside inner ellipse
        if (q < r1) exit
        !     Reject P if outside outer ellipse
        if (q > r2) cycle
        !     Reject P if outside acceptance region
        if (v**2 < -4.0*LOG(u)*u**2) exit
    end do

    !     Return ratio of P's coordinates as the normal deviate
    random_normal = v/u

    end function random_normal

    !======================================================================================================

    subroutine read_param(err)

    integer,intent(out):: err

    open(unit=8,file=fname_prm,status='old',iostat=err) ! read the parameters

    if (err==0) then
        do k1=1,npar
            read (8,'(f9.5)') p(k1)
        enddo
    else
        print *,'Problem with the file ',fname_prm
        call close_data
    endif
    close (unit=8)

    end subroutine read_param



    !=========================================================

    integer function RandomChoicePrg2(nvalues,prob,values)
    !same as RandomChoicePrg but for integer values

    implicit none

    integer,intent(in)    :: nvalues                          ! nvalues is the number of categories
    real, dimension(nvalues),intent(in) :: prob        ! prob is a vector of probabilities
    integer, dimension(nvalues),intent(in) ::values          ! values is a vector of values
    integer :: ii, jj
    real::norm,x

    real, dimension(nvalues+1)::seuil
    ! seuil is a vector of transformed proba so that we get thresholds in the interval [0,1]

    ! here we compute the thresholds
    norm=sum(prob)
    seuil(1)=0

    do ii=2,nvalues+1
        seuil(ii)=0
        do jj=1,ii-1
            seuil(ii)=seuil(ii)+prob(jj)/norm
        end do
    end do

    !  we draw a random number in [0,1] and then check in which
    !  interval defined by the threholds it falls.

    call random_number(x)
    do jj=1, nvalues
        if ( (x .lt. seuil(jj+1)) .and. (x .ge. seuil(jj)) ) then
            RandomChoicePrg2=values(jj)
        end if
    end do

    end function RandomChoicePrg2


    !----------------------------------------------------------------------------------

    end ! end of the program



    !==============================================================================================================
    !==============================================================================================================

    !external function needed by pikaia
    function func(npar,p) result(fn_val)

    implicit none

    integer, intent(in) :: npar   ! number of parameters
    real*8, intent(in) :: p(npar)    ! value of the parameters
    real*8 :: score,fn_val

    INTERFACE
    SUBROUTINE comp_mom(nbr,par,fct)
    IMPLICIT NONE
    INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)
    integer,intent(in) :: nbr   ! number of parameters
    real (dp), intent(in) :: par(nbr)    ! value of the parameters
    real (dp), intent(out):: fct                   ! value of the function
    END SUBROUTINE comp_mom
    END INTERFACE

    call comp_mom(npar,p,score)

    fn_val=1./score

    end function func

    !==============================================================================================================

    !external subroutine needed by newuoa
    subroutine calfun(nbr,par,fct)

    implicit none
    INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)
    integer,intent(in) :: nbr   ! number of parameters
    real (dp), intent(in) :: par(nbr)    ! value of the parameters
    real (dp), intent(out):: fct                   ! value of the function

    INTERFACE
    SUBROUTINE comp_mom(nbr,par,fct)
    IMPLICIT NONE
    INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)
    integer,intent(in) :: nbr   ! number of parameters
    real (dp), intent(in) :: par(nbr)    ! value of the parameters
    real (dp), intent(out):: fct                   ! value of the function
    END SUBROUTINE comp_mom
    END INTERFACE

    call comp_mom(nbr,par,fct)

    end subroutine calfun

    !==============================================================================================================

    subroutine comp_mom(npar,p,fn_val)

    implicit none

    INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)
    integer, intent(in) :: npar   ! number of parameters
    real (dp), intent(in) :: p(npar)    ! value of the parameters
    real (dp), intent(out) :: fn_val                   ! value of the function


    real  :: alpha ! fraction of childrearing cost supported by women
    real  :: cmin ! minimum consumption level, required to be able procreate
    real  :: mu ! good cost to be supported by a household
    real :: nu ! preference parameter. The utility of having no children in log(nu)
    ! la fonction log donne bien le logarithme neperien (log10() est le log en base 10)
    real  :: phi0 ! time cost of having children
    real ::rho,gamma,rhodat,lambda               ! mincerian return and gender gap
    real:: deltam,deltaf !loss of time endowment for single males
    real :: theta,theta_inf !bargaining power
    real:: fn_val_prt ! objective function (not taking into account the constraint on marriage rates)
    integer:: enfant
    real:: distf(19,11) ! distribution of fertility
    real::distf_tot(11) ! dsitribution of fertility - aggregated

    integer::prt
    integer:: nobseff

    real :: simul_marriage_female(19),simul_marriage_male(19) ! simulated marriage rates
    real:: simul_childless_sg(19),simul_childless_ma(19)
    ! simulated childlessness rate of singles and married per wage category
    real :: simul_n_sg(19),simul_n_ma(19)      ! simlated  fertility of single and married mothers per wage category
    real:: simul_n_sg_tot,simul_n_ma_tot,simul_childless_sg_tot,simul_childless_ma_tot,simul_marriage_female_tot, &
        simul_marriage_male_tot

    integer :: rhotype,macho

    integer :: j, i,k,v1,v2

    real :: cf_couple,cf_couple_un,cf_single,cf_single_un,cf1_single,cf_cpl
    real, dimension(2) :: pr_unwanted ! proba to have an uncontrolled fertility

    real :: mean_exp ! moments of the distribution of non labor incomes
    real :: data_sh_female(19),data_sh_male(19)          ! data on shares per wage category
    real :: data_childless_sg(19),data_childless_ma(19),data_se_childless_sg(19),data_se_childless_ma(19)   ! data on childlessness rate of singles and married per wage category
    real :: data_n_sg(19),data_n_ma(19),data_se_n_sg(19),data_se_n_ma(19),data_se_marriage(19)      ! data on fertility of single and married mothers per wage category
    integer:: data_childless_sg_p(19),data_childless_ma_p(19), & ! 1/0  whether the moment is precise or not
    data_n_sg_p(19),data_n_ma_p(19),data_marriage_female_p(19),data_marriage_male_p(19)
    real :: data_marriage_female(19),data_marriage_male(19),data_wages_female(19)      ! data on marriage rate
    real :: data_mortal(19),data_unwanted(19)
    real :: simul_sh_male(19) ! simulated distribution of males across wage groups

    real :: data_n_sg_tot,data_n_ma_tot,data_childless_sg_tot,data_childless_ma_tot,data_marriage_female_tot, &
        data_marriage_male_tot

    integer,parameter ::nobs=250000,nmom=114 !114 for six moments, 95 without marriage of males
    real :: simul_fert(nobs)   ! we are going to draw nobs observations for each category
    integer :: marriage(19,nobs)   ! married or not
    real :: omega(nobs),tpdraw1(nobs),tpdraw2(nobs)
    real :: omega_husband(nobs)        ! non labor income
    integer :: cat_husband(19,nobs)

    real:: chi_D,chi_T,chi_V,chi_S,chi_M,chi_N  ! childlessness types

    real::lovem

    real :: moment(nmom)
    real:: dweight(nmom)
    real :: sterilf,sterilm,steril_ma
    real :: util_f,util_m              ! utility of both spouses
    real :: util_f_un,util_m_un,util_f_sing_un               ! utility of both spouses if uncontrolled fertility
    real :: util_f_sing, util_m_sing                ! utility level
    real :: util_f_sterile,util_m_sterile,util_f_sing_sterile
    integer :: fert_couple, birth_couple, birth_couple_un, birth_cpl, fert_single, fert_single_un,birth_single  ! fertility if married  and single
    integer :: fert_couple_un ! fertility if married and single when uncontrolled

    integer :: maxn_ma,maxn_sg ! maximum number of children

    real :: cnttp(19)
    real::theta_min,theta_max! additional parameters for theta distribution

    integer :: regime(19,nobs),hh  !  regimes for 19 educ categories
    real ::tp(19,14),tptot(14),tpsteril(19),tpsteriltot  ! counts women in each regime

    real::mean_log,se_log,var_log ! for the lognormal distribution

    real:: wage_husband,normw,wage_tp
    character (LEN=10)::fname  ! country name
    logical:: disagree

    common /draws/ cat_husband,tpdraw1,tpdraw2

    common /observ/ data_sh_male,data_sh_female,data_n_sg,data_n_ma,data_childless_sg, &
        data_childless_ma,data_marriage_female,data_marriage_male, &
        data_n_sg_tot,data_n_ma_tot,data_childless_sg_tot,data_childless_ma_tot,data_marriage_female_tot, &
        data_marriage_male_tot,data_mortal,data_unwanted,gamma,rhodat,lambda,  &
        data_childless_sg_p,data_childless_ma_p,data_n_sg_p,data_n_ma_p,data_marriage_female_p,data_marriage_male_p, &
        simul_marriage_female,simul_marriage_male,simul_childless_sg,simul_childless_ma,simul_n_sg,simul_n_ma, &
        simul_n_sg_tot,simul_n_ma_tot,simul_childless_sg_tot,simul_childless_ma_tot,simul_marriage_female_tot, &
        simul_marriage_male_tot

    common /entiers/ prt,rhotype,nobseff,macho

    common /noms/ fname

    common /chi/ chi_D,chi_T,chi_V,chi_S,chi_M,chi_N

    sterilf=0.01 !87
    sterilm=0.01 !87
    steril_ma=sterilf+(1-sterilf)*sterilm


    !link between parameters p and parameters of the model

    phi0=0.07+p(1)*.13
    mean_exp=.1+.7*p(2)
    nu=4.+7.*p(3)
    cmin=.1+.4*p(4)
    mu=.1+.4*p(5)
    alpha=0.5+.5*p(6)
    deltam=-0.05+.4*p(7)
    deltaf=-0.05+.4*p(8)
    theta_inf=p(9)
    var_log=.2+p(10)/2.
    lovem=p(11)/5.

    if (rhotype==1) then
        rho=0.05
    else
        rho=rhodat/100.
    endif

    normw=exp(rho*18.)

    if (alpha>1. .or. cmin<0.  .or. theta_inf<0. .or. theta_inf>1.) then    ! puts constraints on some parameters
        fn_val=1000000*nu
        return
    endif

    do j=1,19
        data_wages_female(j)=gamma*exp(rho*(j-1))/normw
    enddo




    ! some constants
    mean_log=log(mean_exp)-var_log/2.
    se_log=sqrt(var_log)


    do i=1,nobs
        omega(i)= exp(tpdraw1(i)*se_log+mean_log)
        omega_husband(i)=exp(tpdraw2(i)*se_log+mean_log)
    end do



    simul_marriage_male=0.                 ! initialize the marriage rate of males
    cnttp=0.
    do k=1,19 ! to compute the share of males of educ k
        do j=1,19 ! females of educ j
            cnttp(k)=cnttp(k)+count(cat_husband(j,1:nobseff)==k)*data_sh_female(j)   !counts the number of males in each wage category
        enddo
    end do
    simul_sh_male=cnttp/nobseff


    distf=0
    distf_tot=0


    !maximum fertility levels
    maxn_sg=int( max(1.-deltaf,0.)/phi0  ) ! on pourrait faire * (25.-max(0,j-10))/25.  pour enlever des annees de procreation aux eduqus
    maxn_ma=int( 1./(alpha*phi0) )


    do j=1, 19                ! loop: for each category of wages and each person we compute the optimal choice


        pr_unwanted(2)=data_unwanted(j) ! function of probability to either control or not control fertility
        pr_unwanted(1)=1-pr_unwanted(2)
        wage_tp=exp(rho*(j-1))/normw  ! wage of male of same education category

        do i=1,nobseff          ! loop: for each category of wages and each person we compute the optimal choice

            wage_husband=exp(rho*(cat_husband(j,i)-1))/normw !  matching. she is of type j, he is of type cat_husband

            theta=(data_wages_female(j))/(data_wages_female(j)+wage_husband )+ &
                theta_inf*( 1./2.-(data_wages_female(j))/(data_wages_female(j)+wage_husband ))   ! bargaining power of women

            call utility_couple_ok(data_mortal(j),data_wages_female(j),omega(i),wage_husband,omega_husband(i),util_f,&
                util_m,fert_couple,birth_couple,cf_couple)
            call utility_couple_sterile(data_wages_female(j),omega(i),wage_husband,omega_husband(i),util_f_sterile,util_m_sterile)

            call utility_couple_uncontrolled(data_mortal(j),data_wages_female(j),omega(i),wage_husband,omega_husband(i),&
                util_f_un,util_m_un,fert_couple_un,birth_couple_un,cf_couple_un)

            if ((1-deltam)*wage_husband+omega_husband(i)>mu) then
                util_m_sing=cies((1-deltam)*wage_husband+omega_husband(i)-mu,0,nu)  !log((1-deltam)*wage_husband+omega_husband(i)-mu)+log(num)
            else
                util_m_sing=-10E33
            end if


            call utility_single_ok(data_mortal(j),data_wages_female(j),omega(i),util_f_sing,fert_single,   &
                birth_single,cf_single,cf1_single)
            call utility_single_sterile(data_wages_female(j),omega(i),util_f_sing_sterile)



            ! assumption: single perfectly control their fertility
            fert_single_un=fert_single
            util_f_sing_un=util_f_sing
            ! instead here, they do not control fertility:
            !        call utility_single_uncontrolled(data_mortal(j),data_wages_female(j),omega(i),util_f_sing_un,fert_single_un)

            ! as a function of wages and omega
            if (macho==0) then
                disagree=((util_m_sing-lovem>(1-steril_ma)*(pr_unwanted(1)*util_m+pr_unwanted(2)*util_m_un)+steril_ma*util_m_sterile) .or. &
                    ((1-sterilf)*(pr_unwanted(1)*util_f_sing+pr_unwanted(2)*util_f_sing_un)+sterilf*util_f_sing_sterile>&
                    (1-steril_ma)*(pr_unwanted(1)*util_f+pr_unwanted(2)*util_f_un)+steril_ma*util_f_sterile))
            else
                disagree=(util_m_sing-lovem>(1-steril_ma)*(pr_unwanted(1)*util_m+pr_unwanted(2)*util_m_un)+steril_ma*util_m_sterile)
            endif

            if (disagree) then
                if (i<pr_unwanted(2)*nobseff) then
                    simul_fert(i)=fert_single_un
                else
                    simul_fert(i)=fert_single
                endif
                marriage(j,i)=0
            else
                if (i<pr_unwanted(2)*nobseff) then
                    simul_fert(i)=fert_couple_un
                    cf_cpl=cf_couple_un
                    birth_cpl=birth_couple_un
                else
                    simul_fert(i)=fert_couple
                    cf_cpl=cf_couple
                    birth_cpl=birth_couple
                endif
                marriage(j,i)=1 ! this is the ith woman of category j
                distf(j,simul_fert(i)+1)=distf(j,simul_fert(i)+1)+1  ! we count the number of persons having sim_fert children among the married
            end if

            if (simul_sh_male(cat_husband(j,i))>0) then
                simul_marriage_male(cat_husband(j,i))=simul_marriage_male(cat_husband(j,i))    &
                    +data_sh_female(j)*marriage(j,i)/nobseff/simul_sh_male(cat_husband(j,i))
            endif

            if (prt==3 .or. prt==2 .or. prt==1) then
                regime(j,i)=0     ! if mistake
                !singles
                if (marriage(j,i)==0 .and. simul_fert(i)==0 .and. cf1_single<=cmin) then
                    regime(j,i)=2           ! social sterility
                endif
                if (marriage(j,i)==0 .and. birth_single>0 .and. fert_single==0  ) then
                    regime(j,i)=3            ! childlessness bec mortality
                endif
                if (marriage(j,i)==0 .and. simul_fert(i)>0 .and. cf_single>cmin .and. simul_fert(i)<maxn_sg) then
                    regime(j,i)=1             ! interior
                endif
                if (marriage(j,i)==0 .and. birth_single==0 .and. cf1_single>cmin ) then
                    regime(j,i)=4           ! voluntary childlessness
                endif
                if (marriage(j,i)==0 .and. simul_fert(i)==maxn_sg .and. cf_single>=cmin) then
                    regime(j,i)=5             ! maximum fertility
                endif

                ! couples
                if (marriage(j,i)==1 .and. birth_couple==0 .and. cf_cpl<cmin) then
                    regime(j,i)=7    ! involuntary childlessness
                endif
                if (marriage(j,i)==1 .and. birth_couple==0 .and. cf_cpl>=cmin ) then
                    regime(j,i)=9           ! voluntary childlessness
                endif
                if (marriage(j,i)==1 .and. birth_couple>0 .and. fert_couple==0 ) then
                    regime(j,i)=12           !  childlessness bec of mortality
                endif

                if (marriage(j,i)==1 .and. simul_fert(i)>0 .and. cf_cpl==cmin .and. simul_fert(i)<maxn_ma) then
                    regime(j,i)=8            ! eat more and procreate
                endif
                if (marriage(j,i)==1 .and. simul_fert(i)>0 .and. cf_cpl>cmin .and. simul_fert(i)<maxn_ma) then
                    regime(j,i)=6             ! interior
                endif
                if (marriage(j,i)==1 .and. simul_fert(i)==maxn_ma .and. cf_cpl==cmin) then
                    regime(j,i)=10             ! eat and procreate a maximum
                endif
                if (marriage(j,i)==1 .and. simul_fert(i)==maxn_ma .and. cf_cpl>cmin) then
                    regime(j,i)=11             ! maximum fertility
                endif
            endif
        enddo

        simul_marriage_female(j)=1.*(nobseff-count(marriage(j,1:nobseff)==0))/nobseff         ! compute the marriage rate for that category
        if (simul_marriage_female(j)<1.) then
            simul_childless_sg(j)=1.*count((marriage(j,1:nobseff)==0).and. &
                (simul_fert(1:nobseff)==0))/nobseff/(1-simul_marriage_female(j))  ! compute the childlessness rate for that category
        else
            simul_childless_sg(j)=0.
        end if
        if (simul_marriage_female(j)>0.) then
            simul_childless_ma(j)=1.*count((marriage(j,1:nobseff)==1).and.(simul_fert(1:nobseff)==0))/nobseff/simul_marriage_female(j)
        else
            simul_childless_ma(j)=0.
        end if


        simul_n_sg(j)=1.*sum(simul_fert(1:nobseff)*(1-marriage(j,1:nobseff)) )
        simul_n_ma(j)=1.*sum(simul_fert(1:nobseff)*marriage(j,1:nobseff))

        if ((simul_marriage_female(j)<1.).and.(simul_childless_sg(j)<1.)) then
            simul_n_sg(j)=simul_n_sg(j)/(nobseff*(1-simul_marriage_female(j))*(1-simul_childless_sg(j)) )
        else
            simul_n_sg(j)=0.
        end if
        if ((simul_marriage_female(j)>0.).and.(simul_childless_ma(j)<1.)) then
            simul_n_ma(j)=simul_n_ma(j)/(nobseff*simul_marriage_female(j)*(1-simul_childless_ma(j)) )
        else
            simul_n_ma(j)=0.
        end if
    end do



    moment=0
    dweight=0

    do k=1,19 !mom_dim
        moment(k)=(data_childless_ma(k)-simul_childless_ma(k)-steril_ma)
        if (data_childless_ma(k)>0) then
            dweight(k)=data_childless_ma_p(k)/data_childless_ma(k)**2
        else
            dweight(k)=0
        endif
        moment(19+k)=(data_n_ma(k)-simul_n_ma(k))
        if (data_n_ma(k)>0) then
            dweight(19+k)=data_n_ma_p(k)/data_n_ma(k)**2
        else
            dweight(19+k)=0
        endif
        moment(38+k)=(data_childless_sg(k)-simul_childless_sg(k)-sterilf)
        if (data_childless_sg(k)>0 .and. data_marriage_female_tot<.999) then
            dweight(38+k)=data_childless_sg_p(k)/data_childless_sg(k)**2
        else
            dweight(38+k)=0
        endif
        moment(57+k)=(data_n_sg(k)-simul_n_sg(k))
        if ( data_n_sg(k)>0 .and. data_marriage_female_tot<.999) then
            dweight(57+k)=data_n_sg_p(k)/ data_n_sg(k)**2
        else
            dweight(57+k)=0
        endif
        moment(76+k)=(data_marriage_female(k)-simul_marriage_female(k))
        if (data_marriage_female(k)>0 .and. data_marriage_female_tot<.999) then
            dweight(76+k)=data_marriage_female_p(k)/data_marriage_female(k)**2
        else
            dweight(76+k)=0
        endif
        !     here we remove the marriage of males from the moments to match
        !   moment(95+k)=0
        moment(95+k)=(data_marriage_male(k)-simul_marriage_male(k))
        if (data_marriage_male(k)>0) then
            dweight(95+k)=data_marriage_male_p(k)/data_marriage_male(k)**2
        else
            dweight(95+k)=0
        endif
    enddo

    simul_marriage_female_tot=1.*sum(simul_marriage_female*data_sh_female)
    simul_marriage_male_tot=1.*sum(simul_marriage_male*simul_sh_male)
    simul_n_sg_tot=1.*sum(simul_n_sg*data_sh_female*(1-simul_childless_sg)*(1-simul_marriage_female))
    simul_n_ma_tot=1.*sum(simul_n_ma*data_sh_female*(1-simul_childless_ma)*simul_marriage_female)
    simul_childless_sg_tot=1.*sum(simul_childless_sg*data_sh_female*(1-simul_marriage_female)   )
    simul_childless_ma_tot=1.*sum(simul_childless_ma*data_sh_female*simul_marriage_female)
    simul_childless_sg_tot=simul_childless_sg_tot/(1-simul_marriage_female_tot)
    simul_childless_ma_tot=simul_childless_ma_tot/simul_marriage_female_tot
    if (simul_childless_sg_tot<1.) then
        simul_n_sg_tot=simul_n_sg_tot/(1-simul_childless_sg_tot)/(1-simul_marriage_female_tot)
    else
        simul_n_sg_tot=0.
    endif
    if (simul_childless_ma_tot<1.) then
        simul_n_ma_tot=simul_n_ma_tot/(1-simul_childless_ma_tot)/simul_marriage_female_tot
    else
        simul_n_ma_tot=0.
    endif


    fn_val=sum(moment**2*dweight)
    fn_val=fn_val+1000*(data_marriage_female_tot-simul_marriage_female_tot)**2  !  we impose to fit aggregate marriage rates perfectly
    fn_val_prt=sum(moment**2*dweight)

    if (prt==1 .or. prt==2 .or. prt==3) then

        !    open(unit=19,file="education-pairs.txt",status='replace') ! dump of the file
        !    do j=1,19
        !        do i=1,int(data_sh_female(j)*nobseff)
        !            write(19,'(i3,i3,i3)') j,cat_husband(j,i),marriage(j,i)
        !        enddo
        !    enddo
        !    close(unit=19)

        !    open(unit=19,file="af.txt",status='replace') ! dump of the file
        !    do i=1,nobseff
        !        write(19,'(f8.4)') omega(i)
        !    enddo
        !    close(unit=19)

        !     open(unit=19,file="am.txt",status='replace') ! dump of the file
        !    do i=1,nobseff
        !        write(19,'(f8.4)') omega_husband(i)
        !    enddo
        !    close(unit=19)

        distf_tot=0
        do enfant=1,10
            do j=1,19
                distf_tot(enfant)=distf_tot(enfant)+distf(j,enfant)*data_sh_female(j)*simul_marriage_female(j)
            enddo
        enddo

        print '(a35)',' '
        print '(a20)','Equilibrium with'
        print *,'nobs ',nobseff,' lambda=',lambda
        print *,'country : ',fname
        print '(a35)',' '
        print '(a10,f7.3)','mean_exp ',mean_exp
        print '(a10,f7.3)','var_log ',var_log
        print '(a10,f7.3)','nu ',nu
        print '(a10,f7.3)','cmin ',cmin
        print '(a10,f7.3)','alpha ',alpha
        print '(a10,f7.3)','mu ',mu
        print '(a10,f7.3)','phi0 ',phi0
        print '(a10,f7.3)','delta m ',deltam
        print '(a10,f7.3)','delta f ',deltaf
        print '(a10,f7.3)','theta inf ',theta_inf
        print '(a10,f7.3)','lovem ',lovem
        print '(a10,f7.3)','rho ',rho
        print '(a10,f7.3)','gamma ',gamma
        print '(a35)',' '
        print '(a30,2f13.3)','The score for this simulation is ',fn_val_prt
        print '(a35)',' '
        print '(3a26)','marriage rate fe','childlessness rate (cpls)','fertility of mothers (cpls)'
        print '(6a13)','data','simul','data','simul','data','simul'
        do j=1, 19
            print '(6f13.3)',data_marriage_female(j),simul_marriage_female(j),data_childless_ma(j), &
                simul_childless_ma(j)+steril_ma,data_n_ma(j),simul_n_ma(j)
        end do
        print '(a35)',' '
        print '(6f13.3)',data_marriage_female_tot,simul_marriage_female_tot,data_childless_ma_tot, &
            simul_childless_ma_tot+steril_ma,data_n_ma_tot,simul_n_ma_tot
        print '(a35)',' '
        print '(3a26)','childlessness rate (sg)','fertility of mothers (sg)','marriage rate ma'
        print '(6a13)','data','simul','data','simul','data','simul'
        do j=1, 19
            print '(6f13.3)',data_childless_sg(j),simul_childless_sg(j)+sterilf,data_n_sg(j),simul_n_sg(j), &
                data_marriage_male(j),simul_marriage_male(j)
        end do
        print '(a35)',' '
        print '(6f13.3)',data_childless_sg_tot,simul_childless_sg_tot+sterilf,data_n_sg_tot,simul_n_sg_tot, &
            data_marriage_male_tot,simul_marriage_male_tot
        print '(a35)',' '
        print *,' '

        write(3,'(a35)') ' '
        write(3,'(a20)') 'Equilibrium with'
        write(3,'(a10,i7.3,a10,f7.3)') 'nobs ',nobseff,'lambda= ',lambda
        write(3,'(a10,a20)') 'country : ',fname
        write(3,'(a35)') ' '
        write(3,'(a10,f7.3)') 'mean_exp',mean_exp
        write(3,'(a10,f7.3)') 'var_log ',var_log
        write(3,'(a10,f7.3)') 'nu ',nu
        write(3,'(a10,f7.3)') 'cmin ',cmin
        write(3,'(a10,f7.3)') 'alpha ',alpha
        write(3,'(a10,f7.3)') 'mu ',mu
        write(3,'(a10,f7.3)') 'phi0 ',phi0
        write(3,'(a10,f7.3)') 'delta m ',deltam
        write(3,'(a10,f7.3)') 'delta f ',deltaf
        write(3,'(a10,f7.3)') 'theta_inf ',theta_inf
        write(3,'(a10,f7.3)') 'lovem ',lovem
        write(3,'(a10,f7.3)') 'rho ',rho
        write(3,'(a10,f7.3)') 'gamma ',gamma
        write(3,'(a35)') ' '
        write(3,'(a30,f13.3)') 'The score for this simulation is ',fn_val_prt
        write(3,'(a35)') ' '
        write(3,'(3a26)') 'marriage rate fe','childlessness rate (cpls)','fertility of mothers (cpls)'
        write(3,'(6a13)') 'data','simul','data','simul','data','simul'
        do j=1, 19
            write(3,'(6f13.3)') data_marriage_female(j),simul_marriage_female(j),data_childless_ma(j), &
                simul_childless_ma(j)+steril_ma, data_n_ma(j),simul_n_ma(j)
        end do
        write(3,'(a35)') ' '
        write(3,'(6f13.3)') data_marriage_female_tot,simul_marriage_female_tot,data_childless_ma_tot, &
            simul_childless_ma_tot+steril_ma, data_n_ma_tot,simul_n_ma_tot
        write(3,'(a35)') ' '
        write(3,'(3a26)') 'childlessness rate (sg)','fertility of mothers (sg)','marriage rate ma'
        write(3,'(6a13)') 'data','simul','data','simul','data','simul'
        do j=1, 19
            write(3,'(6f13.3)') data_childless_sg(j),simul_childless_sg(j)+sterilf,data_n_sg(j),simul_n_sg(j), &
                data_marriage_male(j),simul_marriage_male(j)
        end do
        write(3,'(a35)')' '
        write(3,'(6f13.3)') data_childless_sg_tot,simul_childless_sg_tot+sterilf,data_n_sg_tot,simul_n_sg_tot, &
            data_marriage_male_tot,simul_marriage_male_tot
        write(3, *)' '
        write(3,*)'Regimes:'
        write(3, *)' '
        write(3,'(13a6)')'I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','O'
        tpsteriltot=0.
        tpsteril=0.
        tptot=0.
        tp=0.
        do j=1, 19
            do hh=0, 5          ! regimes single
                tp(j,hh+1)=100*(count(regime(j,1:nobseff)==hh)*1./nobseff)*(1-sterilf)
                tptot(hh+1)=tptot(hh+1)+tp(j,hh+1)*data_sh_female(j)
            enddo
            do hh=6, 12          ! regimes married
                tp(j,hh+1)=100*(count(regime(j,1:nobseff)==hh)*1./nobseff)*(1-steril_ma)
                tptot(hh+1)=tptot(hh+1)+tp(j,hh+1)*data_sh_female(j)
            enddo

            !counting the share of sterile persons
            tpsteril(j)=100*(sterilf*(1-simul_marriage_female(j))+steril_ma*simul_marriage_female(j))
            tpsteriltot=tpsteriltot+tpsteril(j)*data_sh_female(j)
        enddo
        do j=1, 19
            write (3, '(13f6.2)') tp(j,2:13),tpsteril(j)
        enddo
        if (data_marriage_female_tot>.999 .or. sum(data_childless_sg_p)==0.) then     ! correct when no singles
            tptot(5)=0.
            tptot(4)=0.
            tptot(3)=0.
            tpsteriltot= 100*steril_ma
        endif
        write (3,'(13f6.2)')tptot(2:13),tpsteriltot

        chi_D=100*(data_childless_ma_tot*data_marriage_female_tot+(1-data_marriage_female_tot)*data_childless_sg_tot)
        chi_T=tptot(5)+tptot(10)+tptot(3)+tptot(8)+tptot(4)+tptot(13)+tpsteriltot
        chi_V=tptot(5)+tptot(10)
        chi_S=tptot(3)+tptot(8)
        chi_M=tptot(4)+tptot(13)
        chi_N=tpsteriltot


        write(3, *)' '
        write(3, *)'Decomposition of sources of childlessness '
        write(3, *)' '
        write(3, *)'Voluntary = ',chi_V
        write(3, *)'Involuntary (mortality) = ',chi_M
        write(3, *)'Involuntary (social) = ',chi_S
        write(3, *)'Involuntary (natural) = ',chi_N
        write(3, *)' '
        write(3, *)'Total = ',chi_T
        write(3, *)'In the data = ', chi_D
        write(3, *)' '
        write(3, *)'Total Fertility= ',simul_marriage_female_tot*(1-simul_childless_ma_tot)*simul_n_ma_tot &
            + (1-simul_marriage_female_tot)*(1-simul_childless_sg_tot)*simul_n_sg_tot
        write(3, *)' '
        if (prt==2) then
            do j=1,npar
                write(8,'(f9.5)') p(j)
            enddo
        endif
        write(3, *)' '
        write(3, *)'distribution of fertility:'

        do enfant=1,10
            write(3, *) enfant-1,' ',distf_tot(enfant)
        enddo
        if (prt==3) then
            write(9,'(a12,14f7.3)') fname,lambda,mean_exp,var_log,nu,cmin,alpha,mu,phi0,deltam,deltaf,theta_inf,lovem,fn_val_prt
        endif
    end if

    simul_childless_sg=simul_childless_sg+sterilf ! so that the main program gets the right childlessness rates
    simul_childless_ma=simul_childless_ma+steril_ma


    !======================================================================================================

    contains
    !======================================================================================================


    subroutine utility_couple_ok(survival,wf,omegaf,wm,omegam,utilf,utilm,nopt,npos,cfopt)

    implicit none

    real, intent(in) ::survival
    real, intent(in) :: wf,wm  ! female and male wage
    ! le qualificatif intent(in) precise qu'il s agit d'un argument d'entree
    real, intent(in) :: omegaf,omegam ! endowments
    real, intent(out) ::utilf,utilm  ! utility reached
    integer, intent(out):: nopt,npos  ! optimal fertility
    real,intent(out)::cfopt

    real:: utilc,utilcas   ! utility of the couple
    real :: cf,cm  !  consumption of the female for different levels of n
    real :: x  !  utility of the female for different levels of n
    real::income  ! income of the couple
    integer::neff
    real::vec1(maxn_ma+1)
    integer::vec2(maxn_ma+1)
    integer::i2
    integer:: nmax_eff ! nmax effective, taking budget constraint into account. To improve speed

    utilc= -10E33
    nopt=0

    nmax_eff=min( maxn_ma, int((wf+wm+omegaf+omegam- mu-cmin)/phi0/(alpha*wf+(1-alpha)*wm) ))
    do npos=0, nmax_eff        ! for each level of n, we compute consumption and utility
        utilcas=0.
        do neff=0,npos
            income=(1.-alpha*phi0*neff)*wf +(1.-(1.-alpha)*phi0*neff)*wm+ omegaf+omegam- mu   !income of the household
            cf=theta*income
            cm=(1.-theta)*income
            if (income<0) then   ! income is negative
                x=-10E33
            else  ! income is positive
                if (neff==0) then  !we consider the case of zero children - no constraint on consumption
                    x=theta*cies(cf,0,nu)+(1-theta)*cies(cm,0,nu) ! utility
                else     !we consider the case of a positive number of children
                    if (theta*income<cmin) then    !when bargaining leads to too low consumption for the wife
                        if (income>cmin) then    !income is large enough to pay cmin to the wife
                            cf=cmin           ! consumptions need to be redefined
                            cm=income-cmin
                            x=theta*cies(cf,neff,nu)+(1-theta)*cies(cm,neff,nu)
                        else           !income is not enough to pay cmin to the wife - this is not a solution
                            x=-10E33
                        endif
                    else        !interior solution
                        x=theta*cies(cf,neff,nu)+(1-theta)*cies(cm,neff,nu)
                    endif
                endif
            endif
            utilcas=utilcas+bin(neff,npos,survival)*x
        enddo
        if (utilcas>utilc) then                       ! if this is the highest utility, store the result
            utilc=utilcas
            nopt=npos ! we put the best number of birth into nopt
        endif
    enddo

    npos=nopt ! optimal number of birth

    !compute the results for npos
    utilf=0 ! we need here to compute the expected utility of both partners
    utilm=0
    do neff=0,npos
        income=(1-alpha*phi0*neff)*wf+(1-(1-alpha)*phi0*neff)*wm + omegaf+omegam- mu
        if ((npos>0) .and. (theta*income<cmin)) then    !when bargaining leads to too low consumption for the wife
            cf=cmin           ! consumptions need to be redefined
            cm=income-cmin
        else
            cf=theta*income
            cm=(1-theta)*income
        endif
        utilf=utilf+bin(neff,npos,survival)*cies(cf,neff,nu)
        utilm=utilm+bin(neff,npos,survival)*cies(cm,neff,nu)
    enddo

    vec1=0
    vec2=0
    do i2=1,npos+1
        vec2(i2)=i2-1
        vec1(i2)=bin(i2-1,npos,survival)
    enddo
    nopt=RandomChoicePrg2(npos+1,vec1,vec2) ! here we draw the effective number of surviving children

    if (npos>0) then
        income=((1-alpha*phi0*nopt)*wf +(1-(1-alpha)*phi0*nopt)*wm + omegaf+omegam- mu) ! corresponding income of couple
        cfopt=theta*income
        if  (cfopt<cmin) then    !when bargaining leads to too low consumption for the wife
            cfopt=cmin           ! consumptions need to be redefined
        endif
    else
        cfopt=theta*((1-alpha*phi0)*wf +(1-(1-alpha)*phi0)*wm + omegaf+omegam- mu) ! consumption if one child to test for involuntary
    endif

    end subroutine utility_couple_ok

    !======================================================================================================

    subroutine utility_couple_sterile(wf,omegaf,wm,omegam,utilf,utilm)

    implicit none

    real, intent(in) :: wf,wm  ! female and male wage
    ! le qualificatif intent(in) precise qu'il s agit d'un argument d'entree
    real, intent(in) :: omegaf,omegam ! endowments
    real, intent(out) ::utilf,utilm  ! utility reached

    real:: utilc   ! utility of the couple
    real :: cf,cm  !  consumption of the female for different levels of n
    real :: x  !  utility of the female for different levels of n
    real::income  ! income of the couple


    utilc= -10E33
    utilf= -10E33
    utilm= -10E33
    income=(wf +wm)+ omegaf+omegam- mu  !income of the household
    cf=theta*income
    cm=(1-theta)*income
    if (income<0) then   ! income is negative
        x=-10E33
    else  ! income is positive
        x=theta*cies(cf,0,nu)+(1-theta)*cies(cm,0,nu) ! utility
    endif
    if (x>utilc) then                       ! if this is the highest utility, store the result
        utilc=x
        utilf=cies(cf,0,nu)
        utilm=cies(cm,0,nu)
    end if

    end subroutine utility_couple_sterile

    !======================================================================================================

    subroutine utility_couple_uncontrolled(survival,wf,omegaf,wm,omegam,utilf_un,utilm_un,nopt,npos,cfopt)

    implicit none

    real, intent(in) ::survival
    real, intent(in) :: wf,wm  ! female and male wage
    real, intent(in) :: omegaf,omegam ! endowments
    real, intent(out) ::utilf_un,utilm_un  ! utility reached
    real,intent(out)::cfopt
    integer,intent(out):: nopt,npos

    real :: cf,cm  !  consumption of the female for different levels of n
    real :: x  !  utility of the female for different levels of n
    real::income  ! income of the couple
    integer::neff
    real::vec1(maxn_ma+1)
    integer::vec2(maxn_ma+1)
    integer::i2

    npos=min(maxn_ma,max(0,int((wf+wm+omegaf+omegam-cmin-mu)/(phi0*(alpha*wf+(1-alpha)*wm))))  )

    utilf_un=0
    utilm_un=0
    do neff=0,npos
        income=(1-alpha*phi0*neff)*wf +(1-(1-alpha)*phi0*neff)*wm + omegaf+omegam- mu
        if ((npos>0) .and. (theta*income<cmin)) then    !when bargaining leads to too low consumption for the wife
            cf=cmin           ! consumptions need to be redefined
            cm=income-cmin
        else
            cf=theta*income
            cm=(1-theta)*income
        endif
        utilf_un=utilf_un+bin(neff,npos,survival)*cies(cf,neff,nu)
        utilm_un=utilm_un+bin(neff,npos,survival)*cies(cm,neff,nu)
    enddo

    vec1=0
    vec2=0
    do i2=1,npos+1
        vec2(i2)=i2-1
        vec1(i2)=bin(i2-1,npos,survival)
    enddo
    nopt=RandomChoicePrg2(npos+1,vec1,vec2) ! here we draw the effective number of surviving children

    if (npos>0) then
        income=((1-alpha*phi0*nopt)*wf +(1-(1-alpha)*phi0*nopt)*wm + omegaf+omegam- mu) ! corresponding income of couple
        cfopt=theta*income
        if  (cfopt<cmin) then    !when bargaining leads to too low consumption for the wife
            cfopt=cmin           ! consumptions need to be redefined
        endif
    else
        cfopt=theta*((1-alpha*phi0)*wf +(1-(1-alpha)*phi0)*wm + omegaf+omegam- mu) ! consumption if one child to test for involuntary
    endif


    end subroutine utility_couple_uncontrolled

    ! ====================================================================================================

    real function factoriel(x3)

    integer, intent(in) ::x3
    integer:: i3,tp3

    if (x3==0) then
        factoriel=1
    else
        tp3=1
        do i3=1,x3
            tp3=tp3*i3
        enddo
        factoriel=tp3
    endif

    end function factoriel

    ! ====================================================================================================

    real function bin(neff,npos,survival)

    integer, intent(in) ::neff
    integer, intent(in) ::npos
    real, intent(in) ::survival

    bin=factoriel(npos) / factoriel(neff) / factoriel(npos-neff) * survival**neff * (1-survival)**(npos-neff)

    end function bin

    ! ====================================================================================================

    subroutine utility_single_uncontrolled(survival,wf,omega,util_s_un,nopt)  ! this procedure computes the expected utility
    ! of a single woman who does not control her fertility
    implicit none

    real, intent(in) :: survival
    real, intent(in) :: wf  ! female wage
    ! le qualificatif intent(in) precise qu'il s agit d'un argument d'entree
    real, intent(in) :: omega ! endowment
    real, intent(out) ::util_s_un  ! utility reached
    integer, intent(out):: nopt  ! optimal fertility
    real :: cf  !  consumption of the female for different levels of n
    real :: x  !  utility of the female for different levels of n
    integer::neff
    real:: utilcas
    integer:: nmax_u
    real::vec1(maxn_sg+1)
    integer::vec2(maxn_sg+1)
    integer::i2

    nmax_u=min(maxn_sg,max(0,int(((1-deltaf)*wf+omega-mu-cmin)/(phi0*wf))) )
    nopt=nmax_u

    do neff=0,nmax_u
        cf=(1-deltaf-phi0*nmax_u)*wf + omega- mu      ! consumption for when having neff children for npos birth
        x=cies(cf,neff,nu)          !  utility
        utilcas=utilcas+bin(neff,nmax_u,survival)*x
    enddo
    util_s_un=utilcas
    end subroutine utility_single_uncontrolled

    ! ====================================================================================================


    subroutine utility_single_ok(survival,wf,omega,util,nopt,npos,cfopt,cf1)  ! this procedure computes the utility and optimal fertility
    ! of a women with wage wf and non wage income omega conditionally of being single
    implicit none

    real, intent(in) :: survival
    real, intent(in) :: wf  ! female wage
    ! le qualificatif intent(in) precise qu'il s agit d'un argument d'entree
    real, intent(in) :: omega ! endowment
    real, intent(out) ::util  ! utility reached
    integer, intent(out):: nopt,npos  ! optimal fertility
    real,intent(out)::cfopt
    real :: cf  !  consumption of the female for different levels of n
    real,intent(out) :: cf1  !  consumption of the female for 1 kid
    real :: x  !  utility of the female for different levels of n
    integer::neff
    real:: utilcas
    real::vec1(maxn_sg+1)
    integer::vec2(maxn_sg+1)
    integer::i2
    integer::nmax_eff ! effective max number of children

    util= -10E33
    nopt=0

    cf1=(1-deltaf-phi0)*wf + omega- mu      ! consumption for when having 1 surviving child
    if ((1-deltaf)*wf +omega -mu<0) then
        npos=0
        nopt=0
        goto 222
    endif

    nmax_eff=min(maxn_sg,int(((1-deltaf)*wf+omega- mu-cmin)/phi0/(wf) ))

    do npos=0, nmax_eff        ! for each level of n, we compute consumption and utility

        utilcas=0

        do neff=0,npos
            cf=(1-deltaf-phi0*neff)*wf + omega- mu    ! consumption for when having neff children for npos birth
            if ((cf<cmin).and.(npos>0)) then
                x=-10E33
            else if (cf>0) then
                x=cies(cf,neff,nu)          !  utility
            else
                x=-10E33
            end if
            utilcas=utilcas+bin(neff,npos,survival)*x
        enddo

        if (utilcas>util) then                       ! if this is the highest utility, store the result
            util=utilcas                             ! expected utility of the best case
            nopt=npos ! memorize choice
        end if
    end do

    npos=nopt

    vec1=0
    vec2=0
    do i2=1,npos+1
        vec2(i2)=i2-1
        vec1(i2)=bin(i2-1,npos,survival)
    enddo
    nopt=RandomChoicePrg2(npos+1,vec1,vec2)
222 cfopt=(1-deltaf-phi0*nopt)*wf + omega- mu    ! average consumption

    end subroutine utility_single_ok
    ! ====================================================================================================

    subroutine utility_single_sterile(wf,omega,util)  ! this procedure computes the utility and optimal fertility
    ! of a women with wage wf and non wage income omega conditionally of being single

    real, intent(in) :: wf  ! female wage
    ! le qualificatif intent(in) precise qu'il s agit d'un argument d'entree
    real, intent(in) :: omega ! endowment
    real, intent(out) ::util  ! utility reached
    real :: cf  !  consumption of the female for different levels of n
    real :: x  !  utility of the female for different levels of n


    util= -10E33

    cf=(1-deltaf)*wf + omega- mu     ! consumption for when having n children
    if (cf>0) then
        x=cies(cf,0,nu)          !  utility
    else
        x=-10E33
    end if
    if (x>util) then                       ! if this is the highest utility, store the result
        util=x
    end if

    end subroutine utility_single_sterile

    ! ====================================================================================================

    real function cies(conso,nchild,ncst)  ! this procedure computes the utility

    real, intent(in) :: conso,ncst
    integer, intent(in):: nchild
    real:: ex

    if (conso<0.) then
        cies=-10E33
    else
        cies=log(conso)+log(ncst+nchild)
    endif

    end function cies
    ! ====================================================================================================


    integer function RandomChoicePrg2(nvalues,prob,values)
    !same as RandomChoicePrg but for integer values

    implicit none

    integer,intent(in)    :: nvalues                          ! nvalues is the number of categories
    real, dimension(nvalues),intent(in) :: prob        ! prob is a vector of probabilities
    integer, dimension(nvalues),intent(in) ::values          ! values is a vector of values
    integer :: ii, jj
    real::norm,xx

    real, dimension(nvalues+1)::seuil
    ! seuil is a vector of transformed proba so that we get thresholds in the interval [0,1]

    ! here we compute the thresholds
    norm=sum(prob)
    seuil(1)=0

    do ii=2,nvalues+1
        seuil(ii)=0
        do jj=1,ii-1
            seuil(ii)=seuil(ii)+prob(jj)/norm
        end do
    end do

    !  we draw a random number in [0,1] and then check in which
    !  interval defined by the threholds it falls.

    call random_number(xx)
    do jj=1, nvalues
        if ( (xx .lt. seuil(jj+1)) .and. (xx .ge. seuil(jj)) ) then
            RandomChoicePrg2=values(jj)
        end if
    end do

    end function RandomChoicePrg2

    end subroutine comp_mom